//  .----------------.  .----------------.  .----------------.  .----------------.  .-----------------.
// | .--------------. || .--------------. || .--------------. || .--------------. || .--------------. |
// | |   _____      | || |   ______     | || | ____    ____ | || |      __      | || | ____  _____  | |
// | |  |_   _|     | || |  |_   _ \    | || ||_   \  /   _|| || |     /  \     | || ||_   \|_   _| | |
// | |    | |       | || |    | |_) |   | || |  |   \/   |  | || |    / /\ \    | || |  |   \ | |   | |
// | |    | |   _   | || |    |  __'.   | || |  | |\  /| |  | || |   / ____ \   | || |  | |\ \| |   | |
// | |   _| |__/ |  | || |   _| |__) |  | || | _| |_\/_| |_ | || | _/ /    \ \_ | || | _| |_\   |_  | |
// | |  |________|  | || |  |_______/   | || ||_____||_____|| || ||____|  |____|| || ||_____|\____| | |
// | |              | || |              | || |              | || |              | || |              | |
// | '--------------' || '--------------' || '--------------' || '--------------' || '--------------' |
//  '----------------'  '----------------'  '----------------'  '----------------'  '----------------' 
//
// https://mmg-gitlab.fjfi.cvut.cz/gitlab/lbm/lbman
//
// Written by Radek Fučík (fucik@fjfi.cvut.cz) FNSPE CTU in Prague, Czech Republic
// with a great help of Robert Straka (straka@metal.agh.edu.pl) AGH Krakow, Poland
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// https://www.gnu.org/licenses/

#pragma once

#include "defs.h"
#include "lbm.h"
#include "taylor.h"
#include "cs.h"
#include "pde_struct2.h"

enum { NSE, ADE };

struct CORE : DEF
{
	LBM lbm;

	int ORDER=2;		// Taylor order
	int collision_model=0;
	int verbosity=0;
	bool constvel=false;
	bool constmatrixA=true;
	bool numericmatrixA=false;
	bool EFDEprint;
	bool use_tay_vels=true;

	bool use_symbolic_matrices=!true;
	bool use_NSE_strategies=!true; // do not use, the "zeros generator" is much faster!!!
	bool use_NSE_omegas = false;	// set omega_0 = omega_1 = ... = omega_d+1 = 1 for NSE to speed up everything
	int pde_strategy = NSE;

	// internal
	bool matrix_A_numeric=false;
	ex matrix_A_coef;



        vct f_mu, f_mueq;
        ex f_rho, f_vel[3];
        lst sub_vels;

	char target_dir[1024];
	char model_desc[1024];

	matrix I, E[MQ], A, C, M, Mi, e[MQ];
	symbol *symB;
	matrix *As;
	vector<CS> vecCS;
	int CC[MDIM][MQ];
	bool allocated=false;
	bool allocated_tay=false;

	symbol **sym_vels;
	TAYLOR **tay_vels;
	symbol sym_vel[MDIM];

	ex sym_cs;
	symbol x=symbol("x"),y=symbol("y"),z=symbol("z"),t=symbol("t");
	symbol dl=symbol("dl","\\dl"), dt=symbol("dt","\\dt");

	// optimization coefs
	symbol symA[MDIM][MQ];
	lst optimal_pars[MQ];
	bool optimal_pars_loaded=false;
	void load_optimal_pars();
	void find_optimal_pars(int pde, int lfrom, int lto);
	void set_optimal_pars(int pde, lst &list);

	void taylor_list_vels(lst &list);
	void taylor_list_vels_functions(lst &list);
	void taylor_list_vels_moments(lst &list);

	void prepare_EFDE_load_table(symtab &table);

	void desc2str(char*str, int k, int j1, int j2, int j3);
	void desc2str2(char*str, int k, int j1, int j2, int j3);
	void number2str(char*str, const char*s1, const char*s2, int number);
	void coord2str(char*str, char*str_tex, int t, int x, int y, int z);
	void desc2str(char *s_space_coeff, char* s_time_coeff, char* s_space_formula, char* s_time_formula, int l, int i1, int i2, int i3);

	inline void simplify_matrix(matrix &A) { simplify_matrix(A,Q,Q); }
	inline void simplify_matrix(matrix &A, int Nrows, int Ncols) { for (int i=0;i<Nrows;i++) for (int j=0;j<Ncols;j++) A(i,j) = expand(A(i,j)); }
	inline void empty_matrix(matrix &A) { empty_matrix(A,Q,Q); }
	inline void empty_matrix(matrix &A, int Nrows, int Ncols) { for (int i=0;i<Nrows;i++) for (int j=0;j<Ncols;j++) A(i,j)=0;; }

	inline int indO(int a, int b=0, int c=0) { return a + b*(ORDER+1) + c*(ORDER+1)*(ORDER+1); }

	void define_matrix_A_worker(matrix &_A, ex u, ex v, ex w);
	void define_matrix_A();

	void allocate();
	void allocate_Taylor();

	void log(const char* fmt, ... );

	////////////////////////////////////////////////////////////////////////////////////////////
	//
	// STEP 01: compute LBMAN matrices
	//
	////////////////////////////////////////////////////////////////////////////////////////////

	void load_CP(matrix &P, int l, int i1, int i2, int i3, const char *desc, parser &reader);
	void save_matrix(matrix &P, int l, int i1, int i2, int i3, const char *name, const char *desc);
	ex load_trP(int l, int i1, int i2, int i3, const char *desc, parser &reader);
	void save_trP(ex trP, int l, int i1, int i2, int i3, const char *desc);


	void find_optimal_coefficients(matrix &I_A, ex *pars, int i);
	void compute_LBMAN_B(matrix &B, matrix &X, matrix &BB, matrix *dBMi, matrix *Xs);

	void compute_LBMAN_matrices_SRT_save();

	bool compute_LBMAN_matrices_symbolic_save_check();
	void compute_LBMAN_matrices_symbolic_save_parallel();
	void compute_LBMAN_matrices_symbolic_load(matrix*CP, ex*trP);
	void compute_LBMAN_matrices_symbolic_load_simplify_and_save();
	bool compute_LBMAN_matrices_symbolic_load_simplify_and_save_check();
	void compute_LBMAN_matrices_load(matrix*CP, ex*trP);

	////////////////////////////////////////////////////////////////////////////////////////////
	//
	// STEP 02: compute EFDE coefficients
	//
	////////////////////////////////////////////////////////////////////////////////////////////
	void compute_EFDE_coefs_save();

	void compute_EFDE_parts_save_parallel();

	bool compute_EFDE_parts_save_parallel_check();

	void compute_EFDE_coefs_save_parallel();
	bool compute_EFDE_coefs_save_parallel_check();

	void compute_EFDE_coefs_symbolic_save();

	void load_EFDE_coefs(matrix &coefs, int pde, int l, int i1, int i2, int i3, const char *desc, parser &reader, bool do_subs);
	void load_EPDE_coefs(matrix &coefs, int i0, int i1, int i2, int i3, const char *desc, parser &reader);

	void export_EFDE_coefs(int pde);

	////////////////////////////////////////////////////////////////////////////////////////////
	//
	// STEP 03: compute EPDE coefficients
	//
	////////////////////////////////////////////////////////////////////////////////////////////
	void compute_EPDE_coefs_and_save();
	bool compute_EPDE_coefs_and_save_check();

	void export_EPDE_coefs(int pde);

	CORE(int iDIM, int iQ); // : DEF(iDIM, iQ) { sprintf(target_dir,"default"); };
	~CORE();
};
