#include "defs.h"

bool BASE::file_exists(const std::string &s)
{
	struct stat buffer;
	return (stat (s.c_str(), &buffer) == 0);
}

bool BASE::mark_exists(const char* fmt, ... )
{
	char txt[1024];
	va_list args;
	va_start(args, fmt);
	vsprintf(txt, fmt, args);
	va_end(args);

	return (file_exists(txt)) ? true : false;
}

bool BASE::mark(const char* fmt, ... )
{
	char txt[1024];
	va_list args;
	va_start(args, fmt);
	vsprintf(txt, fmt, args);
	va_end(args);

	if (file_exists(txt))
	{
		return true;
	} else
	{
		ofstream fout(txt);
		fout << getpid() <<endl;
		fout.close();
		usleep(5000);
		int ipid;
		ifstream fin(txt);
		fin >> ipid;
		fin.close();
		if (ipid!=getpid()) return true;
		return false;
	}
	return false;
}

void BASE::create_mark(const char* fmt, ... )
{
	char txt[1024];
	va_list args;
	va_start(args, fmt);
	vsprintf(txt, fmt, args);
	va_end(args);
	if (!file_exists(txt))
	{
		int fd = open(txt, O_CREAT | O_RDWR, 0666);
		close(fd);
	}
}

std::string BASE::ReplaceString(std::string subject, const std::string& search, const std::string& replace)
{
	size_t pos = 0;
	while ((pos = subject.find(search, pos)) != std::string::npos)
	{
		subject.replace(pos, search.length(), replace);
		pos += replace.length();
	}
	return subject;
}

string BASE::factorize_latex(ex co)
{
	ostringstream s;
	s << latex;
	// prepare co
	ex co1 = subs( co, subs_vels );
	ex r = GiNaC::normal( co1 );
	ex nr = numer(r);
	nr = GiNaC::subs( nr, subs_vels);
	nr = GiNaC::normal( nr );
	ex nrex = GiNaC::expand( nr , factor_options::all );
	ex dr = denom(r);
	nr=GiNaC::expand(nr);
	if (dr!=1)
	{
		if (nr.nops()>2)
			s << "\\frac{1}{"<<dr<<"} \\Big[ "<<nr<<" \\Big]";
		else
			s << "\\frac{"<<nr<<"}{"<<dr<<"}";
	} else
	{
		if (nrex.nops()>1)
			s << "\\Big[ "<<nr<<" \\Big]";
		else
			s <<nr;
	}
	string str1 = s.str();
	str1 = ReplaceString( str1, "(t,x,y,z)", "");
	return str1;
}

void DEF::allocate_index()
{
	_index = new int*[L+1];
	for (int l=0;l<=L;l++)
	{
		_index[l] = new int[(2*R+1)*(2*R+1)*(2*R+1)];
		for (int i=0;i<(2*R+1)*(2*R+1)*(2*R+1);i++) _index[l][i]=-1;
	}
}

// test whether we are within the range
bool DEF::isin(int l, int a, int b, int c)
{
	return (_index[l][ind(a,b,c)]>=0);
}

int DEF::index(int l, int a, int b, int c)
{
	if (_index[l][ind(a,b,c)] < 0)
	{
		printf("indexing error: out of range: %d of %d,  %d of %d,  %d of %d,  %d vs. %d\n", l,L,R+a,2*R,R+b,2*R,R+c,2*R);
		return 0;
	}
	// adjust for dimension
	if (DIM==1) return _index[l][ind(a,0,0)];
	if (DIM==2) return _index[l][ind(a,b,0)];
	return _index[l][ind(a,b,c)];
}

void DEF::fill_index()
{
	int count=0;
	int ra=1; // plus one!
	for (int l=0;l<=L;l++)
	{
		if (l<=L/2)
			ra++;
		else
			ra--;
		for (int z=-ra;(z<=ra && DIM==3) || (z<=-ra);z++)
		for (int y=-ra;(y<=ra && DIM>=2) || (y<=-ra);y++)
		for (int x=-ra;x<=ra;x++)
		{
			int xx=x;
			int yy=(DIM==1) ? 0 : y;
			int zz=(DIM<=2) ? 0 : z;
			_index[l][ind(xx,yy,zz)]=count;
			count++;
		}
	}
	Nindex=count;
}


int DEF::ind(int a, int b, int c)
{
	switch (DIM)
	{
		case 1: return (a+R);
		case 2: return (a+R)+(b+R)*(2*R+1);
		case 3: return (a+R)+(b+R)*(2*R+1)+(c+R)*(2*R+1)*(2*R+1);
	}
	return 0;
}


// constructor
DEF::DEF(int iDIM, int iQ)
{
	DIM=iDIM;
	Q=iQ;
	L=iQ+1;
	R=L+1;
	// allocated indexes
	allocate_index();
	fill_index();
	internal_index=true;
}

// constructor
DEF::DEF(int iDIM, int iQ, int**external_index, int external_Nindex)
{
	DIM=iDIM;
	Q=iQ;
	L=iQ+1;
	R=L+1;
	_index=external_index;
	Nindex = external_Nindex;
	internal_index=false;
}
DEF::~DEF()
{
	// delete index
	if (internal_index)
	{
		for (int l=0;l<=L;l++) delete [] _index[l];
		delete [] _index;
	}
}



int Nrows=0, Ncols=0;
INT **d=0;
bool allocated=false;

void INT_MATRIX::allocate(int iNrows, int iNcols)
{
	if (Nrows<0 || Ncols<0) return;
	Nrows=iNrows;
	Ncols=iNcols;
	d=new INT*[Nrows];
	for (int i=0;i<Nrows;i++) d[i]=new INT[Ncols];
	allocated=true;
}

void INT_MATRIX::empty()
{
	if (!allocated) return;
	for (int i=0;i<Nrows;i++)
	for (int j=0;j<Ncols;j++)
		d[i][j]=0;
}

INT_MATRIX::~INT_MATRIX()
{
	if (allocated)
	{
		for (int i=0;i<Nrows;i++) delete []  d[i];
		delete [] d;
	}
}

INT INT_MATRIX::copy_ginac_matrix(matrix &M)
{
	// first analyze denominators in the matrix
	INT lowest=1;
	bool work=true;
	while (work)
	{
		work=false;
		for (int i=0;i<Nrows;i++)
		for (int j=0;j<Ncols;j++)
		{
			INT den = ex_to<numeric>(denom((ex)lowest*M(i,j))).to_int();

			if (den != 1)
			{
				lowest = (lowest<den)?den:lowest;
				work=true;
			}
		}
	}
	cout << "lowest denominator found to be "<<lowest<<endl;

	for (int i=0;i<Nrows;i++)
	for (int j=0;j<Ncols;j++)
		d[i][j]=ex_to<numeric>(lowest*M(i,j)).to_int();
	return lowest;
}

void INT_MATRIX::copy(INT_MATRIX &A)
{
	if (Ncols!=A.Ncols || Nrows!=A.Nrows) { cout << "INT_MATRIX::copy: incorrect dimensions"<<endl; return; }
	for (int i=0;i<Nrows;i++)
	for (int j=0;j<Ncols;j++)
		d[i][j] = A.d[i][j];
}

// this = A.B
void INT_MATRIX::mul(INT_MATRIX &A, INT_MATRIX &B)
{
	if (A.Ncols!=B.Nrows) { cout << "INT_MATRIX::mul: incorrect dimensions A.Ncols!=B.Nrows"<<endl; return; }
	if (Nrows != A.Nrows) { cout << "INT_MATRIX::mul: incorrect dimensions Nrows != A.Nrows"<<endl; return; }
	if (Ncols != B.Ncols) { cout << "INT_MATRIX::mul: incorrect dimensions Ncols != B.Ncols"<<endl; return; }
	for (int i=0;i<Nrows;i++)
	for (int j=0;j<Ncols;j++)
	{
		d[i][j]=0;
		for (int k=0;k<A.Ncols;k++)
			d[i][j] += A.d[i][k]*B.d[k][j];
	}
}

// this = this . A
void INT_MATRIX::mul(INT_MATRIX &A)
{
	INT_MATRIX THIS;
	THIS.allocate(Ncols,Nrows);
	THIS.copy(*this);
	mul(THIS, A);
}

void INT_MATRIX::add(INT_MATRIX &A)
{
	if (Ncols!=A.Ncols || Nrows!=A.Nrows) { cout << "INT_MATRIX::add: incorrect dimensions"<<endl; return; }
	for (int i=0;i<Nrows;i++)
	for (int j=0;j<Ncols;j++)
		d[i][j] += A.d[i][j];
}

bool INT_MATRIX::is_zero_matrix()
{
	for (int i=0;i<Nrows;i++)
	for (int j=0;j<Ncols;j++)
		if (d[i][j]!=0) return false;
	return true;
}

INT INT_MATRIX::trace()
{
	INT ret=0;
	for (int i=0;i<Nrows;i++) ret += d[i][i];
	return ret;
}

void INT_MATRIX::mul_scalar(INT fctr)
{
	for (int i=0;i<Nrows;i++)
	for (int j=0;j<Ncols;j++)
		d[i][j]*=fctr;
}

void INT_MATRIX::sub_diagonal(INT expr)
{
	for (int j=0;j<Ncols;j++) d[j][j]-=expr;
}

