//  .----------------.  .----------------.  .----------------.  .----------------.  .-----------------.
// | .--------------. || .--------------. || .--------------. || .--------------. || .--------------. |
// | |   _____      | || |   ______     | || | ____    ____ | || |      __      | || | ____  _____  | |
// | |  |_   _|     | || |  |_   _ \    | || ||_   \  /   _|| || |     /  \     | || ||_   \|_   _| | |
// | |    | |       | || |    | |_) |   | || |  |   \/   |  | || |    / /\ \    | || |  |   \ | |   | |
// | |    | |   _   | || |    |  __'.   | || |  | |\  /| |  | || |   / ____ \   | || |  | |\ \| |   | |
// | |   _| |__/ |  | || |   _| |__) |  | || | _| |_\/_| |_ | || | _/ /    \ \_ | || | _| |_\   |_  | |
// | |  |________|  | || |  |_______/   | || ||_____||_____|| || ||____|  |____|| || ||_____|\____| | |
// | |              | || |              | || |              | || |              | || |              | |
// | '--------------' || '--------------' || '--------------' || '--------------' || '--------------' |
//  '----------------'  '----------------'  '----------------'  '----------------'  '----------------'
//
// https://mmg-gitlab.fjfi.cvut.cz/gitlab/lbm/lbman
//
// Written by Radek Fučík (fucik@fjfi.cvut.cz) FNSPE CTU in Prague, Czech Republic
// with a great help of Robert Straka (straka@metal.agh.edu.pl) AGH Krakow, Poland
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// https://www.gnu.org/licenses/

#pragma once

#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <iostream>
#include <sstream>
#include <fstream>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <dirent.h>
#include <sys/file.h>
#include <ginac/ginac.h>

const int MQ=27;// maximal Q ... lazy memory allocation
const int MDIM=3; // maximal DIM .. lazy memory allocation

using namespace std;
using namespace GiNaC;

using vct=std::vector<ex>;
using vcti=std::vector<int>;

enum {
	SRT,		// 0
	MRT1,		// 1
	MRT2,		// 2
	MRT1_OG,	// 3
	MRT2_OG,	// 4
	CLBM1,		// 5
	CLBM2,		// 6
	CLBM1_OG,	// 7
	CLBM2_OG,	// 8
	GENERAL		// 9
};


struct BASE
{
	bool factor_coefs = true;
	bool file_exists(const std::string &s);
	bool mark_exists(const char* fmt, ... );
	bool mark(const char* fmt, ... );
	void create_mark(const char* fmt, ... );

	string factorize_latex(ex co);
	lst subs_vels; // for factorize_latex

	std::string ReplaceString(std::string subject, const std::string& search, const std::string& replace);
};


struct DEF : BASE
{
	bool internal_index=true;
	int R;
	int Q;
	int DIM;
	int L;

	int Nindex=0;  // total number of entries needed for the algorithm
	int **_index; // main index function (replaces ind)


	void allocate_index();
	bool isin(int l, int a, int b, int c);
	int index(int l, int a, int b, int c);
	void fill_index();
	int ind(int a, int b, int c);


	DEF(int iDIM, int iQ);
	DEF(int iDIM, int iQ, int**external_index, int external_Nindex);
	~DEF();
};

using INT=long long int;

struct INT_MATRIX
{
	int Nrows=0, Ncols=0;
	INT **d=0;
	bool allocated=false;

	void allocate(int iNrows, int iNcols);
	void empty();

	INT copy_ginac_matrix(matrix &M);
	void copy(INT_MATRIX &A);
	void mul(INT_MATRIX &A, INT_MATRIX &B);
	void mul(INT_MATRIX &A);
	void add(INT_MATRIX &A);
	bool is_zero_matrix();
	INT trace();
	void mul_scalar(INT fctr);
	void sub_diagonal(INT expr);

	~INT_MATRIX();
};

struct SUBS_STRUCT
{
	symbol s; // symbolic representation
	ex e; // expression
};
using subs_vct = std::vector<SUBS_STRUCT>;
