//  .----------------.  .----------------.  .----------------.  .----------------.  .-----------------.
// | .--------------. || .--------------. || .--------------. || .--------------. || .--------------. |
// | |   _____      | || |   ______     | || | ____    ____ | || |      __      | || | ____  _____  | |
// | |  |_   _|     | || |  |_   _ \    | || ||_   \  /   _|| || |     /  \     | || ||_   \|_   _| | |
// | |    | |       | || |    | |_) |   | || |  |   \/   |  | || |    / /\ \    | || |  |   \ | |   | |
// | |    | |   _   | || |    |  __'.   | || |  | |\  /| |  | || |   / ____ \   | || |  | |\ \| |   | |
// | |   _| |__/ |  | || |   _| |__) |  | || | _| |_\/_| |_ | || | _/ /    \ \_ | || | _| |_\   |_  | |
// | |  |________|  | || |  |_______/   | || ||_____||_____|| || ||____|  |____|| || ||_____|\____| | |
// | |              | || |              | || |              | || |              | || |              | |
// | '--------------' || '--------------' || '--------------' || '--------------' || '--------------' |
//  '----------------'  '----------------'  '----------------'  '----------------'  '----------------' 
//
// https://mmg-gitlab.fjfi.cvut.cz/gitlab/lbm/lbman
//
// Written by Radek Fučík (fucik@fjfi.cvut.cz) FNSPE CTU in Prague, Czech Republic
// with a great help of Robert Straka (straka@metal.agh.edu.pl) AGH Krakow, Poland
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// https://www.gnu.org/licenses/

#pragma once

#include "defs.h"

struct PDE_STRUCT2 : BASE
{
	int max_factors_common_order=2;

	vct bas;
	int MO;	// MAX_ORDER
	int MOindex; // maximal index possible for one set of basis functions (with derivatives up to order MO)
	int Nbas;

	// duplicit allocation for easy derivation
	// MO x MO
	ex **data2D0=0;
	ex **data2D1=0;
	ex **data2D=data2D0;
	ex **prev2D=data2D1;
	// MO
	ex *data1D0=0;
	ex *data1D1=0;
	ex *data1D=data1D0;
	ex *prev1D=data1D1;

	ex data0D=0;

	ex *basders=0;
	symbol x,y,z,t;

	lst zero_ders;

	void allocate(vct &ibas, int iMAX_ORDER, symbol &it, symbol &ix, symbol &iy, symbol &iz);

	inline int ind(int bindex, int a, int b, int c, int d) { return a + b*(MO+1) + c*(MO+1)*(MO+1) + d*(MO+1)*(MO+1)*(MO+1) + bindex*(MO+1)*(MO+1)*(MO+1)*(MO+1); }

	bool is_empty();
	bool nonzero_coeff(int bi, int j0, int j1, int j2, int j3);
	bool replace_der_by_expr(int bi, int kt, int kx, int ky, int kz, PDE_STRUCT2 &subsexpr);


	// data writting with ordering in place
	inline ex& data_w(int bi, int i0, int i1, int i2, int i3, int bj, int j0, int j1, int j2, int j3)
	{
		if (ind(bi,i0,i1,i2,i3)<=ind(bj,j0,j1,j2,j3))
			return data2D[ind(bi,i0,i1,i2,i3)][ind(bj,j0,j1,j2,j3)];
		else
			return data2D[ind(bj,j0,j1,j2,j3)][ind(bi,i0,i1,i2,i3)];
	}

	inline ex& data(int bi, int i0, int i1, int i2, int i3, int bj, int j0, int j1, int j2, int j3) { return data2D[ind(bi,i0,i1,i2,i3)][ind(bj,j0,j1,j2,j3)]; }
	inline ex& prev(int bi, int i0, int i1, int i2, int i3, int bj, int j0, int j1, int j2, int j3) { return prev2D[ind(bi,i0,i1,i2,i3)][ind(bj,j0,j1,j2,j3)]; }
	inline ex& data(int bi, int i0, int i1, int i2, int i3) { return data1D[ind(bi,i0,i1,i2,i3)]; }
	inline ex& prev(int bi, int i0, int i1, int i2, int i3) { return prev1D[ind(bi,i0,i1,i2,i3)]; }

	bool get_der_index(ex der, int &bi, int &i0, int &i1, int &i2, int &i3);

	// decompose expression into the structure
	void eat(ex expression);
	void print() { ostringstream sout; print(sout); cout << sout.str(); }
	void print(ostringstream &sout);
	void empty(ex **d);
	void empty(ex *d);
	inline void empty() { empty(data2D0); empty(data2D1); empty(data1D0); empty(data1D1); data0D=0; }

	void mul(ex e);
	void add(PDE_STRUCT2 &A);
	bool compare(PDE_STRUCT2 &A);

	void mul(PDE_STRUCT2 &A);

	ex raw();
	void expand();
	void normal();

	void diff_worker(int i);
	void diff(int i, int order);

	string derivative_tex(int bj, int j0, int j1, int j2, int j3);
	void save(const char*dirname, const char*subdirname, const char*fmt, ...);
	void save_tex(const char*dirname, const char*subdirname, const char*fmt, ...);

	void copy(PDE_STRUCT2 &A);

	void swap();
	void free();

	PDE_STRUCT2() {}

	~PDE_STRUCT2() { free(); }
};
