#include "spatial_pde.h"


bool SPATIAL_PDE::find_lowest_time_derivative(PDE_STRUCT &expression, int basindex, int order, int &kt, int &kx, int &ky, int &kz)
{
	int k=order-1;
	for (int j0=1;j0<=k;j0++)
	for (int j3=0;(j3<=k && DIM==3) || (j3<=0);j3++)
	for (int j2=0;(j2<=k && DIM>=2) || (j2<=0);j2++)
	for (int j1=0;j1<=k;j1++)
	if (j0+j1+j2+j3<=order)
	if (expression.nonzero_coeff(basindex, j0,j1,j2,j3))
	{
		kt=j0;
		kx=j1;
		ky=j2;
		kz=j3;
		return true;
	}
	return false;
}

ex SPATIAL_PDE::trunc_orders(ex& expression, ex fnct, int order_from, int order_to)
{
	lst list;
	// print orders
	for (int k=order_from;k<=order_to;k++)
	for (int j=0;j<=k;j++)
	for (int j3=0;(j3<=k && DIM==3) || (j3<=0);j3++)
	for (int j2=0;(j2<=k && DIM>=2) || (j2<=0);j2++)
	for (int j1=0;j1<=k;j1++)
	if (j1+j2+j3==j)
		list.append( diff(diff(diff(diff(fnct,t,k-j),x,j1),y,j2),z,j3) == 0 );
	return expand(loc_normal(subs(expression, list)));
}

void SPATIAL_PDE::subs_time_derivatives_up_to_order(PDE_STRUCT &expr, int basindex, PDE_STRUCT *ders, int max_order)
{
	PDE_STRUCT subsexpr;
	// substitute spatial derivatives of ders[i], i=0..order into expr
	for (int o=0;o<max_order;o++)
	{
		bool difference=true;
		while (difference)
		{
			int kt=0,kx=0,ky=0,kz=0;
			if (find_lowest_time_derivative(expr,basindex,ORDER+1,kt,kx,ky,kz))
			{
				if (kt==o+1)
				{
					// differentiate step-by-step with truncation is much faster than the version commented below
					subsexpr.copy(ders[o]);
					subsexpr.diff(1,kx);
					if (DIM>=2) subsexpr.diff(2,ky);
					if (DIM==3) subsexpr.diff(3,kz);
					difference = expr.replace_der_by_expr(basindex, kt, kx, ky, kz, subsexpr);
				} else break;
			} else break;
		}
	}
	expr.expand();
}

void SPATIAL_PDE::eliminate_all_spatial_orders_for_given_time_order(int basindex, int time_order, PDE_STRUCT *ders)
{
	PDE_STRUCT subsexpr;
	bool difference=true;
	int iter=0;
	while(difference)
	{
		int kt=0,kx=0,ky=0,kz=0;
		if (!find_lowest_time_derivative(ders[time_order],basindex,ORDER+1,kt,kx,ky,kz)) break;
		cout << "iter"<<iter++<<": lowest order with time derivative found: basindex=" << basindex <<" kt="<<kt<<" kx="<<kx<<" ky="<<ky<<" kz="<<kz<< " time_order="<<time_order<<endl;
		fflush(stdout);

		if (kt <= time_order)
		{
			cout << "ERROR: lowest time derivative " <<kt<< "  is lower or equal than the time_order " << time_order << endl;
			cout << "DEBUG:"<<endl;
			ders[time_order].print();
			break;
		}
		if (kt > time_order+1 || kt > ORDER) break;

		subsexpr.copy(ders[time_order]);
		subsexpr.diff(1,kx);
		if (DIM>=2) subsexpr.diff(2,ky);
		if (DIM==3) subsexpr.diff(3,kz);
		difference = ders[time_order].replace_der_by_expr(basindex, kt, kx, ky, kz, subsexpr);
	}
	ders[time_order].expand();
}


void SPATIAL_PDE::eliminate_all_spatial_orders_for_given_time_order(PDE_STRUCT &target, int basindex, int time_order, PDE_STRUCT &source)
{
	PDE_STRUCT subsexpr;
	bool difference=true;
	int iter=0;
	while(difference)
	{
		int kt=0,kx=0,ky=0,kz=0;
		if (!find_lowest_time_derivative(target,basindex,ORDER+1,kt,kx,ky,kz)) break;
		cout << "iter"<<iter++<<": lowest order with time derivative found: basindex=" << basindex <<" kt="<<kt<<" kx="<<kx<<" ky="<<ky<<" kz="<<kz<< " time_order="<<time_order<<endl;
		fflush(stdout);

		if (kt <= time_order)
		{
			cout << "ERROR: lowest time derivative " <<kt<< "  is lower or equal than the time_order " << time_order << endl;
			cout << "DEBUG:"<<endl;
			target.print();
			break;
		}
		if (kt > time_order+1 || kt > ORDER) break;

		subsexpr.copy(source);
		subsexpr.diff(1,kx);
		if (DIM>=2) subsexpr.diff(2,ky);
		if (DIM==3) subsexpr.diff(3,kz);
		difference = target.replace_der_by_expr(basindex, kt, kx, ky, kz, subsexpr);
	}
	target.expand();
}



// take ders[bi] and eliminate all time derivatives of bas[bj] of given order in this equation
void SPATIAL_PDE::replace_time_derivative(PDE_STRUCT &target, int bj, int time_order, PDE_STRUCT &derivative)
{
	PDE_STRUCT subsexpr;
	subsexpr.copy(derivative);
	target.replace_der_by_expr(bj, time_order, 0, 0, 0, subsexpr);
}

void SPATIAL_PDE::eliminate_same_time_derivative(int basindex, int time_order, PDE_STRUCT *ders)
{
	// check coef
	PDE_STRUCT subsexpr;
	bool difference=true;
	int iter=0;
	while(difference)
	{
		int kt=0,kx=0,ky=0,kz=0;
		if (!find_lowest_time_derivative(ders[time_order],basindex,ORDER+1,kt,kx,ky,kz)) break;
		cout << "iter"<<iter++<<": same_time_derivative with time_order="<<time_order<<": lowest order with time derivative found: basindex=" << basindex <<" kt="<<kt<<" kx="<<kx<<" ky="<<ky<<" kz="<<kz<<endl;
		fflush(stdout);
		if (kt!=time_order+1)  break;
		if (kx!=0 || ky!=0 || kz!=0) break;
		// else replace the same time order in the same expression
		subsexpr.copy(ders[time_order]);
		difference = ders[time_order].replace_der_by_expr(basindex, kt, kx, ky, kz, subsexpr);
	}
	ders[time_order].expand();
}


void SPATIAL_PDE::pde_assembly(const int N, PDE_STRUCT *pde, bool loc_use_moments=true, bool loc_use_moments_eq=false)
{
	symtab table;
	prepare_EFDE_load_table(table);
	parser reader(table);

	matrix epde_coefs=matrix(Q,Q), epde_coefs_eq=matrix(Q,Q);
	// prepare subs lists
	lst list_vels, list_vels_functions;
	if (use_tay_vels)
	{
		allocate_Taylor();
		taylor_list_vels(list_vels);
		if (use_moments)
			taylor_list_vels_moments(list_vels_functions);
		else
			taylor_list_vels_functions(list_vels_functions);
	}

	lst subs_vels;
	for (int d=0;d<DIM;d++)
	if (use_moments)
		subs_vels.append( sym_vel[d]==f_mu[d+1]/f_mu[0] );
	else
		subs_vels.append( sym_vel[d]==f_vel[d] );

	for (int pi=0;pi<N;pi++)
	{
		for (int k=0;k<=ORDER;k++)
		for (int j3=0;(j3<=k && DIM==3) || (j3<=0);j3++)
		for (int j2=0;(j2<=k && DIM>=2) || (j2<=0);j2++)
		for (int j1=0;j1<=k;j1++)
		for (int j0=0;j0<=k;j0++)
		if (j0+j1+j2+j3==k)
		{
			load_EPDE_coefs(epde_coefs, j0, j1, j2, j3, "epde_coefs", reader);
			load_EPDE_coefs(epde_coefs_eq, j0, j1, j2, j3, "epde_coefs_eq", reader);
			for (int q=0;q<Q;q++) // moment order
			{
				// moments
				ex expr = epde_coefs(pi,q);
				if (expr!=0)
				{
					if (use_tay_vels)
					{
						expr = expand(subs(expr, list_vels));
						expr = expand(subs(expr, list_vels_functions));
					}
					expr = subs( expr, subs_vels );
					ex der=0;

					if (loc_use_moments)
					{
						der = (q==0) ? f_rho : f_mu[q];
					} else
					{
						if (q==0) der = f_rho;
						if (pde_strategy==NSE)
						{
							switch (DIM)
							{
								case 3:	if (q==3) der = f_vel[2]*f_rho;
								case 2:	if (q==2) der = f_vel[1]*f_rho;
								case 1:	if (q==1) der = f_vel[0]*f_rho;
							}
						} else
						{
							if (q>0 && q<=DIM) der = f_mu[q];
						}
						if (q>=DIM+1) der = f_mu[q];
					}
					if (der==0 && expr!=0) cout << "k="<<k<<" warning der is zero but expr:"<<expr<<endl;
					pde[pi].eat(expand(expr*diff(diff(diff(diff(der,t,j0),x,j1),y,j2),z,j3)));
				}
				// eqilibrium
				expr = epde_coefs_eq(pi,q);
				if (expr!=0)
				{
					if (use_tay_vels)
					{
						expr = expand(subs(expr, list_vels));
						expr = expand(subs(expr, list_vels_functions));
					}
					expr = subs( expr, subs_vels );
					ex der=0;
					if (loc_use_moments_eq)
					{
						cout << "ERROR: loc_use_moments_eq not implemented"<<endl;
/*
						if (q==0) der = f_mu[0];
						switch (DIM)
						{
							case 1:
								if (q==1) der = f_mu[1];
								if (q==2) der = f_mu[0]*(f_mu[1]*f_mu[1]/f_mu[0]/f_mu[0] + sym_cs*sym_cs);
								break;
							case 2:
								if (q==1) der = f_mu[1];
								if (q==2) der = f_mu[2];
								if (q==3) der = f_mu[0]*(f_mu[1]*f_mu[1]/f_mu[0]/f_mu[0] + sym_cs*sym_cs);
								if (q==4) der = f_mu[0]*(f_mu[2]*f_mu[2]/f_mu[0]/f_mu[0] + sym_cs*sym_cs);
								// equilibrium Asinari_PRE_2008bR.pdf P7 Eq 16
								// beware of correct order of moments - Asinari has different 4th, 5th and 6th row!
								if (q==5) der = f_mu[1]*f_mu[2]/f_mu[0];
								if (q==6) der = f_mu[2]*(sym_cs*sym_cs + f_mu[1]*f_mu[1]/f_mu[0]/f_mu[0]);
								if (q==7) der = f_mu[1]*(sym_cs*sym_cs + f_mu[2]*f_mu[2]/f_mu[0]/f_mu[0]);
								if (q==8) der = f_mu[0]*(sym_cs*sym_cs*sym_cs*sym_cs + sym_cs*sym_cs*(f_mu[1]*f_mu[1] + f_mu[2]*f_mu[2])/f_mu[0]/f_mu[0] + f_mu[1]*f_mu[1]*f_mu[2]*f_mu[2]/f_mu[0]/f_mu[0]/f_mu[0]/f_mu[0]);
								break;
							case 3: cout << "ERROR: loc_use_moments_eq for DIM"<<DIM<<" NOT IMPLEMENTED"<<endl;
								break;
						}
*/
					} else
					{
						if (q==0) der = f_rho;
						switch (Q)
						{
						case 3:
							if (q==1) der = f_rho*f_vel[0];
							if (q==2) der = f_rho*(f_vel[0]*f_vel[0] + sym_cs*sym_cs);
							break;
						case 5:
							if (q==1) der = f_rho*f_vel[0];
							if (q==2) der = f_rho*f_vel[1];
							if (q==3) der = f_rho*(f_vel[0]*f_vel[0] + sym_cs*sym_cs);
							if (q==4) der = f_rho*(f_vel[1]*f_vel[1] + sym_cs*sym_cs);
							break;
						case 7:
							if (q==1) der = f_rho*f_vel[0];
							if (q==2) der = f_rho*f_vel[1];
							if (q==3) der = f_rho*f_vel[2];
							if (q==4) der = f_rho*(f_vel[0]*f_vel[0] + sym_cs*sym_cs);
							if (q==5) der = f_rho*(f_vel[1]*f_vel[1] + sym_cs*sym_cs);
							if (q==6) der = f_rho*(f_vel[2]*f_vel[2] + sym_cs*sym_cs);
							break;
						case 9: // equilibrium Asinari_PRE_2008bR.pdf P7 Eq 16
							// beware of correct order of moments - Asinari has different 4th, 5th and 6th row!
							if (q==1) der = f_rho*f_vel[0];
							if (q==2) der = f_rho*f_vel[1];
							if (q==3) der = f_rho*(f_vel[0]*f_vel[1]);
							if (q==4) der = f_rho*(f_vel[0]*f_vel[0] + sym_cs*sym_cs);
							if (q==5) der = f_rho*(f_vel[1]*f_vel[1] + sym_cs*sym_cs);
							if (q==6) der = f_rho*(f_vel[1]*sym_cs*sym_cs + f_vel[0]*f_vel[0]*f_vel[1]);
							if (q==7) der = f_rho*(f_vel[0]*sym_cs*sym_cs + f_vel[0]*f_vel[1]*f_vel[1]);
							if (q==8) der = f_rho*(sym_cs*sym_cs*sym_cs*sym_cs + sym_cs*sym_cs*(f_vel[0]*f_vel[0] + f_vel[1]*f_vel[1]) + f_vel[0]*f_vel[0]*f_vel[1]*f_vel[1]);
							break;
						case 27: // using c_s^2 = 1/3
							if (q==1) der = f_rho*f_vel[0];
							if (q==2) der = f_rho*f_vel[1];
							if (q==3) der = f_rho*f_vel[2];
							if (q==4) der = f_rho*(f_vel[0]*f_vel[1]);
							if (q==5) der = f_rho*(f_vel[0]*f_vel[2]);
							if (q==6) der = f_rho*(f_vel[1]*f_vel[2]);
							if (q==7) der = 0;
							if (q==8)  der = f_rho*f_vel[0]*f_vel[0] + f_rho/3;
							if (q==9)  der = f_rho*f_vel[1]*f_vel[1] + f_rho/3;
							if (q==10) der = f_rho*f_vel[2]*f_vel[2] + f_rho/3;
							if (q==11) der = f_rho*f_vel[1]/3;
							if (q==12) der = f_rho*f_vel[2]/3;
							if (q==13) der = f_rho*f_vel[1]*f_vel[2]/3;
							if (q==14) der = f_rho*f_vel[0]/3;
							if (q==15) der = f_rho*f_vel[2]/3;
							if (q==16) der = f_rho*f_vel[0]*f_vel[2]/3;
							if (q==17) der = f_rho*f_vel[0]/3;
							if (q==18) der = f_rho*f_vel[1]/3;
							if (q==19) der = f_rho*f_vel[0]*f_vel[1]/3;
							if (q==20) der = f_rho*(f_vel[1]*f_vel[1] + f_vel[0]*f_vel[0])/3 +f_rho/9;
							if (q==21) der = f_rho*f_vel[2]*f_vel[2]/3 + f_rho*f_vel[0]*f_vel[0]/3 + f_rho/9;
							if (q==22) der = f_rho*f_vel[2]*f_vel[2]/3 + f_rho*f_vel[1]*f_vel[1]/3 + f_rho/9;
							if (q==23) der = f_rho*f_vel[2]/9;
							if (q==24) der = f_rho*f_vel[1]/9;
							if (q==25) der = f_rho*f_vel[0]/9;
							if (q==26) der = f_rho*f_vel[2]*f_vel[2]/9 + f_rho*f_vel[1]*f_vel[1]/9 + f_rho*f_vel[0]*f_vel[0]/9 + f_rho/27;
							break;
						}
					}
					if (der==0 && expr!=0) cout << "k="<<k<<" warning der is zero: "<<der<<" but expr:"<<expr<<endl;
					pde[pi].eat(expand(expr*diff(diff(diff(diff(der,t,j0),x,j1),y,j2),z,j3)));
				}
			}
		}
	}
}

void SPATIAL_PDE::spatial_EPDE_NSE()
{
	log("spatial_EPDE_NSE\n");
	const int N=DIM+1; // number of equations to deal with here
	int step=0;
	ex _pde[N];
	lst list[N];
	subs_vct sv[N];

	vct bas;
	bas.push_back( f_rho );
	bas.push_back( f_vel[0] );
	if (DIM>=2) bas.push_back( f_vel[1] );
	if (DIM>=3) bas.push_back( f_vel[2] );

	// copy pde into PS
	PDE_STRUCT pde[N];
	for (int pi=0;pi<N;pi++)
	{
		pde[pi].allocate(bas, ORDER, t, x, y, z); // possible derivatives
		pde[pi].empty();
	}

	if (verbosity) 	cout << "[STEP "<<step++<<": assemble initial PDE]"<<endl; fflush(stdout);
	pde_assembly(N, pde, false, false);

	PDE_STRUCT der[N][MAXO]; // store derivatives of mu0 wrt to t only --> we can easily get all spatial derivatives from that
	for (int k=0;k<MAXO;k++)
	for (int pi=0;pi<N;pi++)
	{
		der[pi][k].allocate(bas, ORDER, t,x,y,z);
		der[pi][k].empty();
	}

	if (verbosity) cout << "[STEP "<<step++<<": process pde0]"<<endl; fflush(stdout);
	// get dbas[0]/dt coefficient
	ex c0m0 = pde[0].data(0,1,0,0,0);
	cout << "coef c0m0:" << endl << c0m0 << endl;

	der[0][0].copy(pde[0]);
	der[0][0].mul(-1/c0m0);
	der[0][0].eat( diff(bas[0],t) );
	der[0][0].normal();

	// eliminate first order time derivatives
	eliminate_same_time_derivative(0, 0, der[0]);
	eliminate_all_spatial_orders_for_given_time_order(0, 0, der[0]); 

	for (int k=1;k<MAXO;k++)
	{
		der[0][k].copy(der[0][k-1]);
		der[0][k].diff(0,1);
		subs_time_derivatives_up_to_order(der[0][k], 0, der[0], k);
		eliminate_same_time_derivative(0, k, der[0]);
		eliminate_all_spatial_orders_for_given_time_order(0, k, der[0]); 
		for (int j=0;j<k;j++)
			subs_time_derivatives_up_to_order(der[0][j], 0, der[0], k+1);
	}

	if (verbosity) cout << "[STEP "<<step++<<": process pde1]"<<endl; fflush(stdout);

	subs_time_derivatives_up_to_order(pde[1], 0, der[0], MAXO);

	ex c1m1 = pde[1].data(1,1,0,0,0);
	cout << " c1m1: "<<c1m1<<endl;

	der[1][0].copy(pde[1]);
	der[1][0].mul(-1/c1m1);
	der[1][0].eat( diff(bas[1],t) );
	der[1][0].normal();

	// eliminate first order time derivatives
	eliminate_same_time_derivative(1, 0, der[1]);
	eliminate_all_spatial_orders_for_given_time_order(1, 0, der[1]); 

	for (int k=1;k<MAXO;k++)
	{
		der[1][k].copy(der[1][k-1]);
		der[1][k].diff(0,1);

		subs_time_derivatives_up_to_order(der[1][k], 0, der[0], MAXO);
		subs_time_derivatives_up_to_order(der[1][k], 1, der[1], k);

		eliminate_same_time_derivative(1, k, der[1]);
		eliminate_all_spatial_orders_for_given_time_order(1, k, der[1]); 
		for (int j=0;j<k;j++)
		{
			subs_time_derivatives_up_to_order(der[1][j], 1, der[1], k+1);
			subs_time_derivatives_up_to_order(der[0][j], 1, der[1], k+1);
		}
	}

	if (DIM>=2)
	{
		if (verbosity) 	cout << "[STEP "<<step++<<": process pde2]"<<endl<<endl; fflush(stdout);

		subs_time_derivatives_up_to_order(pde[2], 0, der[0], MAXO);
		subs_time_derivatives_up_to_order(pde[2], 1, der[1], MAXO);

		ex c2m2 = pde[2].data(2,1,0,0,0);
		for (int pi=0;pi<bas.size();pi++) c2m2 = trunc_orders(c2m2, bas[pi], 1, ORDER+1);
		cout << " c2m2: "<<c2m2<<endl;

		der[2][0].copy(pde[2]);
		der[2][0].mul(-1/c2m2);
		der[2][0].eat( diff(bas[2],t) );
		der[2][0].normal();

		// eliminate first order time derivatives
		eliminate_same_time_derivative(2, 0, der[2]);
		eliminate_all_spatial_orders_for_given_time_order(2, 0, der[2]); 

		for (int k=1;k<MAXO;k++)
		{
			der[2][k].copy(der[2][k-1]);
			der[2][k].diff(0,1);
			subs_time_derivatives_up_to_order(der[2][k], 0, der[0], MAXO);
			subs_time_derivatives_up_to_order(der[2][k], 1, der[1], MAXO);
			subs_time_derivatives_up_to_order(der[2][k], 2, der[2], k);
			eliminate_same_time_derivative(2, k, der[2]);
			eliminate_all_spatial_orders_for_given_time_order(2, k, der[2]); 
			for (int j=0;j<k;j++)
			{
				subs_time_derivatives_up_to_order(der[2][j], 2, der[2], k+1);
				subs_time_derivatives_up_to_order(der[1][j], 2, der[2], k+1);
				subs_time_derivatives_up_to_order(der[0][j], 2, der[2], k+1);
			}
		}
	}

	if (DIM>=3)
	{
		if (verbosity) 	cout << "[STEP "<<step++<<": process pde3]"<<endl<<endl; fflush(stdout);

		subs_time_derivatives_up_to_order(pde[3], 0, der[0], MAXO);
		subs_time_derivatives_up_to_order(pde[3], 1, der[1], MAXO);
		subs_time_derivatives_up_to_order(pde[3], 2, der[2], MAXO);

		ex c3m3 = pde[3].data(3,1,0,0,0);
		for (int pi=0;pi<bas.size();pi++) c3m3 = trunc_orders(c3m3, bas[pi], 1, ORDER+1);
		cout << " c3m3: "<<c3m3<<endl;

		der[3][0].copy(pde[3]);
		der[3][0].mul(-1/c3m3);
		der[3][0].eat( diff(bas[3],t) );
		der[3][0].normal();

		// eliminate first order time derivatives
		eliminate_all_spatial_orders_for_given_time_order(3, 0, der[3]); 

		for (int k=1;k<MAXO;k++)
		{
			der[3][k].copy(der[3][k-1]);
			der[3][k].diff(0,1);
			subs_time_derivatives_up_to_order(der[3][k], 0, der[0], MAXO);
			subs_time_derivatives_up_to_order(der[3][k], 1, der[1], MAXO);
			subs_time_derivatives_up_to_order(der[3][k], 2, der[2], MAXO);
			subs_time_derivatives_up_to_order(der[3][k], 3, der[3], k);
			eliminate_all_spatial_orders_for_given_time_order(3, k, der[3]); 
			for (int j=0;j<k;j++)
			{
				subs_time_derivatives_up_to_order(der[3][j], 3, der[3], k+1);
				subs_time_derivatives_up_to_order(der[2][j], 3, der[3], k+1);
				subs_time_derivatives_up_to_order(der[1][j], 3, der[3], k+1);
				subs_time_derivatives_up_to_order(der[0][j], 3, der[3], k+1);
			}
		}
	}

	if (verbosity) cout << "[STEP "<<step++<<": output and save]"<<endl<<endl; fflush(stdout);

	for (int pi=0;pi<N;pi++) der[pi][0].normal();
	PDE_STRUCT pde_phys[N], aux;
	// mass conservation
	pde_phys[0].copy(der[0][0]);
	pde_phys[0].mul(-(ex)1);
	pde_phys[0].eat(diff(bas[0],t));

	for (int d=0;d<DIM;d++)
	{
		// momentum 0: pde_phys[1] = expand(loc_normal( bas[0]*(diff(bas[1],t) - der[1][0]) + bas[1]*(diff(bas[0],t) - der[0][0]))  );
		aux.copy(der[0][0]);
		aux.mul(-bas[d+1]);
		pde_phys[d+1].copy(der[d+1][0]);
		pde_phys[d+1].mul(-bas[0]);
		pde_phys[d+1].add(aux);
		pde_phys[d+1].eat( bas[0]*diff(bas[d+1],t)  + bas[d+1]*diff(bas[0],t) );
	}

	for (int pi=0;pi<N;pi++)
	{
		pde_phys[pi].normal();
		pde_phys[pi].save(target_dir, "spatial_EPDE","pde_%02d", pi);
		pde_phys[pi].subs_vels = subs_vels;
		pde_phys[pi].save_tex(target_dir, "spatial_EPDE","pde_%02d", pi);
	}
}

//
void SPATIAL_PDE::spatial_EPDE_NSE_matrix()
{
	log("spatial_EPDE_NSE_matrix\n");
	const int N=DIM+1; // number of equations to deal with here
	int step=0;
	ex _pde[N];
	lst list[N];
	subs_vct sv[N];

	vct bas;
	bas.push_back( f_rho );
	bas.push_back( f_vel[0] );
	if (DIM>=2) bas.push_back( f_vel[1] );
	if (DIM>=3) bas.push_back( f_vel[2] );

	// copy pde into PS
	PDE_STRUCT pde[N];
	for (int pi=0;pi<N;pi++)
	{
		pde[pi].allocate(bas, ORDER, t, x, y, z); // possible derivatives
		pde[pi].empty();
	}

	if (verbosity) 	cout << "[STEP "<<step++<<": assemble initial PDE]"<<endl; fflush(stdout);
	pde_assembly(N, pde, false, false);

	PDE_STRUCT der[N][MAXO]; // store derivatives of mu0 wrt to t only --> we can easily get all spatial derivatives from that
	for (int k=0;k<MAXO;k++)
	for (int pi=0;pi<N;pi++)
	{
		der[pi][k].allocate(bas, ORDER, t,x,y,z);
		der[pi][k].empty();
	}

	if (verbosity) cout << "[STEP "<<step++<<": create initial matrix]"<<endl; fflush(stdout);

	matrix W = matrix(N, N);
	for (int i=0;i<N;i++)
	for (int j=0;j<N;j++)
	{
		W(i,j) = pde[i].data(j,1,0,0,0);
		pde[i].data(j,1,0,0,0)=0; // clear it!
	}
	matrix Winv = inverse(W);
	cout << "matrix W="<<W<<endl;
	cout << "matrix Winv="<<Winv<<endl;
	// multiply -Winv*pde
	PDE_STRUCT a;
	for (int i=0;i<N;i++)
	{
		for (int j=0;j<N;j++)
		{
			a.copy(pde[j]);
			a.mul(-Winv(i,j));
			der[i][0].add(a);
		}
		der[i][0].normal();
	}

// two versions of implementation
#ifndef matrix_spatial

	int ord=0;

	if (verbosity) cout << "[STEP "<<step++<<": eliminate same time derivatives up to first order]"<<endl; fflush(stdout);

	for (int i=0;i<N;i++)
	{
		// diagonal first
		replace_time_derivative(der[i][ord], i, ord+1, der[i][ord]);
		eliminate_all_spatial_orders_for_given_time_order(der[i][ord], i, ord, der[i][ord]);
		for (int j=0;j<N;j++) if (i!=j)
		{
			replace_time_derivative(der[j][ord], i, ord+1, der[i][ord]);
			eliminate_all_spatial_orders_for_given_time_order(der[j][ord], i, ord, der[i][ord]);
		}
		der[i][ord].normal();
	}

	bool work=true;
	while (work)
	{
		ord++;
		if (verbosity) cout << "[STEP "<<step++<<": starting order="<<ord<<"]"<<endl; fflush(stdout);
		for (int i=0;i<N;i++)
		{
			// time diff
			der[i][ord].copy(der[i][ord-1]);
			der[i][ord].diff(0,1);
		}

		for (int i=0;i<N;i++)
		for (int j=0;j<N;j++) 
			subs_time_derivatives_up_to_order(der[i][ord], j, der[j], ord);

		if (verbosity) cout << "[STEP "<<step++<<": eliminate same time derivatives up to order="<<ord+1<<"]"<<endl; fflush(stdout);
		for (int i=0;i<N;i++)
		{
			// diagonal first
			replace_time_derivative(der[i][ord], i, ord+1, der[i][ord]);
			eliminate_all_spatial_orders_for_given_time_order(der[i][ord], i, ord, der[i][ord]);

			for (int j=0;j<N;j++) if (i!=j)
			{
				replace_time_derivative(der[j][ord], i, ord+1, der[i][ord]);
				eliminate_all_spatial_orders_for_given_time_order(der[j][ord], i, ord, der[i][ord]);
			}
		}

		if (verbosity) cout << "[STEP "<<step++<<": backward substitution]"<<endl; fflush(stdout);
		for (int k=0;k<ord;k++)
		for (int i=0;i<N;i++)
		for (int j=0;j<N;j++)
			subs_time_derivatives_up_to_order(der[i][k], j, der[j], ord+1);

		for (int i=0;i<N;i++) der[i][ord].normal();

		work=false;
		int kt,kx,ky,kz;
		for (int i=0;i<N;i++)
		for (int j=0;j<N;j++)
			if (find_lowest_time_derivative(der[i][0], j, ORDER+1,kt,kx,ky,kz)) work = true;
	}

#else
	if (verbosity) cout << "[STEP "<<step++<<": process pde0]"<<endl; fflush(stdout);
	// eliminate first order time derivatives
	eliminate_same_time_derivative(0, 0, der[0]);
	eliminate_all_spatial_orders_for_given_time_order(0, 0, der[0]);

	for (int k=1;k<MAXO;k++)
	{
		der[0][k].copy(der[0][k-1]);
		der[0][k].diff(0,1);
		subs_time_derivatives_up_to_order(der[0][k], 0, der[0], k);
		eliminate_same_time_derivative(0, k, der[0]);
		eliminate_all_spatial_orders_for_given_time_order(0, k, der[0]);
		for (int j=0;j<k;j++)
			subs_time_derivatives_up_to_order(der[0][j], 0, der[0], k+1);
	}

	if (verbosity) cout << "[STEP "<<step++<<": process pde1]"<<endl; fflush(stdout);

	subs_time_derivatives_up_to_order(der[1][0], 0, der[0], MAXO);

	// eliminate first order time derivatives
	eliminate_same_time_derivative(1, 0, der[1]);
	eliminate_all_spatial_orders_for_given_time_order(1, 0, der[1]);

	for (int k=1;k<MAXO;k++)
	{
		der[1][k].copy(der[1][k-1]);
		der[1][k].diff(0,1);

		subs_time_derivatives_up_to_order(der[1][k], 0, der[0], MAXO);
		subs_time_derivatives_up_to_order(der[1][k], 1, der[1], k);

		eliminate_same_time_derivative(1, k, der[1]);
		eliminate_all_spatial_orders_for_given_time_order(1, k, der[1]);
		for (int j=0;j<k;j++)
		{
			subs_time_derivatives_up_to_order(der[1][j], 1, der[1], k+1);
			subs_time_derivatives_up_to_order(der[0][j], 1, der[1], k+1);
		}
	}

	if (DIM>=2)
	{
		if (verbosity) 	cout << "[STEP "<<step++<<": process pde2]"<<endl<<endl; fflush(stdout);

		subs_time_derivatives_up_to_order(der[2][0], 0, der[0], MAXO);
		subs_time_derivatives_up_to_order(der[2][0], 1, der[1], MAXO);

		// eliminate first order time derivatives
		eliminate_same_time_derivative(2, 0, der[2]);
		eliminate_all_spatial_orders_for_given_time_order(2, 0, der[2]);

		for (int k=1;k<MAXO;k++)
		{
			der[2][k].copy(der[2][k-1]);
			der[2][k].diff(0,1);
			subs_time_derivatives_up_to_order(der[2][k], 0, der[0], MAXO);
			subs_time_derivatives_up_to_order(der[2][k], 1, der[1], MAXO);
			subs_time_derivatives_up_to_order(der[2][k], 2, der[2], k);
			eliminate_same_time_derivative(2, k, der[2]);
			eliminate_all_spatial_orders_for_given_time_order(2, k, der[2]);
			for (int j=0;j<k;j++)
			{
				subs_time_derivatives_up_to_order(der[2][j], 2, der[2], k+1);
				subs_time_derivatives_up_to_order(der[1][j], 2, der[2], k+1);
				subs_time_derivatives_up_to_order(der[0][j], 2, der[2], k+1);
			}
		}
	}

	if (DIM>=3)
	{
		if (verbosity) 	cout << "[STEP "<<step++<<": process pde3]"<<endl<<endl; fflush(stdout);

		subs_time_derivatives_up_to_order(der[3][0], 0, der[0], MAXO);
		subs_time_derivatives_up_to_order(der[3][0], 1, der[1], MAXO);
		subs_time_derivatives_up_to_order(der[3][0], 2, der[2], MAXO);
		// eliminate first order time derivatives
		eliminate_all_spatial_orders_for_given_time_order(3, 0, der[3]);

		for (int k=1;k<MAXO;k++)
		{
			der[3][k].copy(der[3][k-1]);
			der[3][k].diff(0,1);
			subs_time_derivatives_up_to_order(der[3][k], 0, der[0], MAXO);
			subs_time_derivatives_up_to_order(der[3][k], 1, der[1], MAXO);
			subs_time_derivatives_up_to_order(der[3][k], 2, der[2], MAXO);
			subs_time_derivatives_up_to_order(der[3][k], 3, der[3], k);
			eliminate_all_spatial_orders_for_given_time_order(3, k, der[3]);
			for (int j=0;j<k;j++)
			{
				subs_time_derivatives_up_to_order(der[3][j], 3, der[3], k+1);
				subs_time_derivatives_up_to_order(der[2][j], 3, der[3], k+1);
				subs_time_derivatives_up_to_order(der[1][j], 3, der[3], k+1);
				subs_time_derivatives_up_to_order(der[0][j], 3, der[3], k+1);
			}
		}
	}
#endif
	if (verbosity) cout << "[STEP "<<step++<<": output and save]"<<endl<<endl; fflush(stdout);

	for (int pi=0;pi<N;pi++) der[pi][0].normal();
	PDE_STRUCT pde_phys[N], aux;
	// mass conservation
	pde_phys[0].copy(der[0][0]);
	pde_phys[0].mul(-(ex)1);
	pde_phys[0].eat(diff(bas[0],t));

	for (int d=0;d<DIM;d++)
	{
		// momentum 0: pde_phys[1] = expand(loc_normal( bas[0]*(diff(bas[1],t) - der[1][0]) + bas[1]*(diff(bas[0],t) - der[0][0]))  );
		aux.copy(der[0][0]);
		aux.mul(-bas[d+1]);
		pde_phys[d+1].copy(der[d+1][0]);
		pde_phys[d+1].mul(-bas[0]);
		pde_phys[d+1].add(aux);
		pde_phys[d+1].eat( bas[0]*diff(bas[d+1],t)  + bas[d+1]*diff(bas[0],t) );
	}

	for (int pi=0;pi<N;pi++)
	{
		pde_phys[pi].normal();
		pde_phys[pi].save(target_dir, "spatial_EPDE","pde_%02d_order_%d", pi, ORDER);
		pde_phys[pi].subs_vels = subs_vels;
		pde_phys[pi].save_tex(target_dir, "spatial_EPDE","pde_%02d_order_%d", pi, ORDER);
	}
}


// todo: expand to general dimension
void SPATIAL_PDE::spatial_EPDE_ADE()
{
	log("spatial_EPDE_ADE starting\n");
	const int N=1; // number of equations to deal with here
	int step=0;

	vct bas;
	bas.push_back( f_rho );
	bas.push_back( f_vel[0] );
	if (DIM>=2) bas.push_back( f_vel[1] );
	if (DIM>=3) bas.push_back( f_vel[2] );
	bas.push_back( f_mu[1] );
	if (DIM>=2) bas.push_back( f_mu[2] );
	if (DIM>=3) bas.push_back( f_mu[3] );

	// copy pde into PS
	PDE_STRUCT pde[N];
	pde[0].allocate(bas, ORDER, t, x, y, z); // possible derivatives

	if (verbosity) 	cout << "[STEP "<<step++<<": assemble initial PDE]"<<endl<<endl; fflush(stdout);
	pde_assembly(N, pde, true, false);

	PDE_STRUCT der[N][MAXO]; // store derivatives of mu0 wrt to t only --> we can easily get all spatial derivatives from that
	for (int k=0;k<MAXO;k++) der[0][k].allocate(bas, ORDER, t,x,y,z);

	if (verbosity) cout << "[STEP "<<step++<<": process pde0]"<<endl<<endl; fflush(stdout);
	// get dbas[0]/dt coefficient
	ex c0m0 = pde[0].data(0,1,0,0,0);
	cout << "coef c0m0:" << endl << c0m0 << endl;

	der[0][0].copy(pde[0]);
	der[0][0].mul(-(ex)1/c0m0);
	der[0][0].eat( diff(bas[0],t) );
	der[0][0].normal();

	// eliminate first order time derivatives
	eliminate_same_time_derivative(0, 0, der[0]);
	eliminate_all_spatial_orders_for_given_time_order(0, 0, der[0]);

	for (int k=1;k<MAXO;k++)
	{
		der[0][k].copy(der[0][k-1]);
		der[0][k].diff(0,1);
		subs_time_derivatives_up_to_order(der[0][k], 0, der[0], k);
		eliminate_same_time_derivative(0, k, der[0]);
		eliminate_all_spatial_orders_for_given_time_order(0, k, der[0]);
		for (int j=0;j<k;j++)
			subs_time_derivatives_up_to_order(der[0][j], 0, der[0], k+1);
	}

	if (verbosity) cout << "[STEP "<<step++<<": output]"<<endl<<endl; fflush(stdout);
	log("spatial_EPDE_ADE step %02d\n", step);
	PDE_STRUCT pde_out[N];
	int pi=0;
	pde_out[pi].copy(der[pi][0]);
	pde_out[pi].mul(-(ex)1);
	pde_out[pi].eat(diff(bas[pi],t));
	pde_out[pi].normal();

	pde_out[pi].subs_vels = subs_vels;
	pde_out[pi].save(target_dir, "spatial_EPDE", "pde_order_%d",ORDER);
	pde_out[pi].save_tex(target_dir, "spatial_EPDE", "pde_order_%d",ORDER);
	log("spatial_EPDE finished.\n");
}
