//  .----------------.  .----------------.  .----------------.  .----------------.  .-----------------.
// | .--------------. || .--------------. || .--------------. || .--------------. || .--------------. |
// | |   _____      | || |   ______     | || | ____    ____ | || |      __      | || | ____  _____  | |
// | |  |_   _|     | || |  |_   _ \    | || ||_   \  /   _|| || |     /  \     | || ||_   \|_   _| | |
// | |    | |       | || |    | |_) |   | || |  |   \/   |  | || |    / /\ \    | || |  |   \ | |   | |
// | |    | |   _   | || |    |  __'.   | || |  | |\  /| |  | || |   / ____ \   | || |  | |\ \| |   | |
// | |   _| |__/ |  | || |   _| |__) |  | || | _| |_\/_| |_ | || | _/ /    \ \_ | || | _| |_\   |_  | |
// | |  |________|  | || |  |_______/   | || ||_____||_____|| || ||____|  |____|| || ||_____|\____| | |
// | |              | || |              | || |              | || |              | || |              | |
// | '--------------' || '--------------' || '--------------' || '--------------' || '--------------' |
//  '----------------'  '----------------'  '----------------'  '----------------'  '----------------' 
//
// https://mmg-gitlab.fjfi.cvut.cz/gitlab/lbm/lbman
//
// Written by Radek Fučík (fucik@fjfi.cvut.cz) FNSPE CTU in Prague, Czech Republic
// with a great help of Robert Straka (straka@metal.agh.edu.pl) AGH Krakow, Poland
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// https://www.gnu.org/licenses/

#pragma once

#include "defs.h"
#include "core.h"
#include "pde_struct2.h"

using PDE_STRUCT=PDE_STRUCT2;

struct SPATIAL_PDE : CORE
{
	const int MAXO=5; // maximal time order to deal with

	// config:
	bool use_moments=!true; 	// use MOMENTS for EPDE -- use the default choice, it is much faster than the other choice!!!
	bool use_moments_eq=!true; 	// use MOMENTS for EPDE -- use the default choice, it is much faster than the other choice!!!
	bool use_normal=true;  		// use the normal function when simplifying?
	ex loc_normal(ex e) { if (use_normal) return normal(e); else return e; }



	void find_optimal_coefficients(matrix &I_A, ex *pars, int i, int moment_index);
	bool find_lowest_time_derivative(PDE_STRUCT &expression, int basindex, int order, int& kt, int& kx, int& ky, int&kz);
	ex trunc_orders(ex &expression, ex fnct, int order_from, int order_to);

	void subs_time_derivatives_up_to_order(PDE_STRUCT &expr, int basindex, PDE_STRUCT *ders, int max_order);
	void eliminate_all_spatial_orders_for_given_time_order(int basindex, int time_order, PDE_STRUCT *ders);
	void eliminate_all_spatial_orders_for_given_time_order(PDE_STRUCT &target, int basindex, int time_order, PDE_STRUCT &source);
	void replace_time_derivative(PDE_STRUCT &target, int bj, int time_order, PDE_STRUCT &derivative);
	void eliminate_same_time_derivative(int basindex, int time_order, PDE_STRUCT *ders);

	void pde_assembly(const int N, PDE_STRUCT *pde, bool loc_use_moments, bool loc_use_moments_eq);

	void spatial_EPDE_NSE();
	void spatial_EPDE_NSE_matrix();
	void spatial_EPDE_ADE();

	void spatial_EPDE()
	{
//		if (pde_strategy==NSE) spatial_EPDE_NSE();
		if (pde_strategy==NSE) spatial_EPDE_NSE_matrix();
		if (pde_strategy==ADE) spatial_EPDE_ADE();
	};


	SPATIAL_PDE(int iDIM, int iQ) : CORE(iDIM, iQ) {};
	~SPATIAL_PDE() {} ;
};
