#include "lbm.h"

void LBM::setA_GENERAL(matrix &A, int Q, ex u, ex v, ex w)
{
	ex SA = symbolic_matrix(Q,Q,"a");
	for (int i=0;i<Q;i++)
	for (int j=0;j<Q;j++)
		A(i,j) = SA[j+i*Q]; // symbolic
}


void LBM::setA_SRT(matrix &A, int Q, ex u, ex v, ex w)
{
	// SRT: identity * omega
	for (int i=0;i<Q;i++)
		A(i,i)=sym_o;
}

void LBM::setA_MRT_using_M(matrix &A, int Q,ex u,ex v,ex w)
{
	matrix M = matrix(Q,Q);
	setM(M,Q);
	// prepare S matrix
	matrix S = matrix(Q,Q);
	for (int i=0;i<Q;i++)
	for (int j=0;j<Q;j++)
	{
		if (i==j)
			S(i,j) = sym_os[i];
		else
			S(i,j) = 0;
	}
	matrix Mi = inverse(M);
	A = Mi.mul(S).mul(M);
	for (int i=0;i<Q;i++)
	for (int j=0;j<Q;j++)
		A(i,j) = expand(A(i,j));
}

// non-orthogonal A = Mi S M
void LBM::setA_MRT1(matrix &A, int Q, ex u, ex v, ex w)
{
	switch (Q)
	{
	case 3:
		A(0,0) = sym_os[0];
		A(0,1) = sym_os[0]-sym_os[2];
		A(0,2) = sym_os[0]-sym_os[2];
		A(1,0) = 0;
		A(1,1) = n1o2*sym_os[2]+n1o2*sym_os[1];
		A(1,2) = n1o2*sym_os[2]-n1o2*sym_os[1];
		A(2,0) = 0;
		A(2,1) = n1o2*sym_os[2]-n1o2*sym_os[1];
		A(2,2) = n1o2*sym_os[2]+n1o2*sym_os[1];
		break;
	case 5:
		A(0,0) = sym_os[0];
		A(0,1) = sym_os[0]-sym_os[3];
		A(0,2) = sym_os[0]-sym_os[4];
		A(0,3) = sym_os[0]-sym_os[3];
		A(0,4) = sym_os[0]-sym_os[4];
		A(1,0) = 0;
		A(1,1) = n1o2*sym_os[3]+n1o2*sym_os[1];
		A(1,2) = 0;
		A(1,3) = n1o2*sym_os[3]-n1o2*sym_os[1];
		A(1,4) = 0;
		A(2,0) = 0;
		A(2,1) = 0;
		A(2,2) = n1o2*sym_os[4]+n1o2*sym_os[2];
		A(2,3) = 0;
		A(2,4) = n1o2*sym_os[4]-n1o2*sym_os[2];
		A(3,0) = 0;
		A(3,1) = n1o2*sym_os[3]-n1o2*sym_os[1];
		A(3,2) = 0;
		A(3,3) = n1o2*sym_os[3]+n1o2*sym_os[1];
		A(3,4) = 0;
		A(4,0) = 0;
		A(4,1) = 0;
		A(4,2) = n1o2*sym_os[4]-n1o2*sym_os[2];
		A(4,3) = 0;
		A(4,4) = n1o2*sym_os[4]+n1o2*sym_os[2];
		break;
	case 7:
		A(0,0) = sym_os[0];
		A(0,1) = sym_os[0]-sym_os[4];
		A(0,2) = sym_os[0]-sym_os[5];
		A(0,3) = sym_os[0]-sym_os[6];
		A(0,4) = sym_os[0]-sym_os[4];
		A(0,5) = sym_os[0]-sym_os[5];
		A(0,6) = sym_os[0]-sym_os[6];
		A(1,0) = 0;
		A(1,1) = n1o2*sym_os[4]+n1o2*sym_os[1];
		A(1,2) = 0;
		A(1,3) = 0;
		A(1,4) = n1o2*sym_os[4]-n1o2*sym_os[1];
		A(1,5) = 0;
		A(1,6) = 0;
		A(2,0) = 0;
		A(2,1) = 0;
		A(2,2) = n1o2*sym_os[5]+n1o2*sym_os[2];
		A(2,3) = 0;
		A(2,4) = 0;
		A(2,5) = n1o2*sym_os[5]-n1o2*sym_os[2];
		A(2,6) = 0;
		A(3,0) = 0;
		A(3,1) = 0;
		A(3,2) = 0;
		A(3,3) = n1o2*sym_os[6]+n1o2*sym_os[3];
		A(3,4) = 0;
		A(3,5) = 0;
		A(3,6) = n1o2*sym_os[6]-n1o2*sym_os[3];
		A(4,0) = 0;
		A(4,1) = n1o2*sym_os[4]-n1o2*sym_os[1];
		A(4,2) = 0;
		A(4,3) = 0;
		A(4,4) = n1o2*sym_os[4]+n1o2*sym_os[1];
		A(4,5) = 0;
		A(4,6) = 0;
		A(5,0) = 0;
		A(5,1) = 0;
		A(5,2) = n1o2*sym_os[5]-n1o2*sym_os[2];
		A(5,3) = 0;
		A(5,4) = 0;
		A(5,5) = n1o2*sym_os[5]+n1o2*sym_os[2];
		A(5,6) = 0;
		A(6,0) = 0;
		A(6,1) = 0;
		A(6,2) = 0;
		A(6,3) = n1o2*sym_os[6]-n1o2*sym_os[3];
		A(6,4) = 0;
		A(6,5) = 0;
		A(6,6) = n1o2*sym_os[6]+n1o2*sym_os[3];
		break;
	case 9:
		A(0,0)=sym_os[0];
		A(0,1)=sym_os[0]-sym_os[4];
		A(0,2)=sym_os[0]-sym_os[5];
		A(0,3)=sym_os[0]-sym_os[4];
		A(0,4)=sym_os[0]-sym_os[5];
		A(0,5)=sym_os[8]-sym_os[5]-sym_os[4]+sym_os[0];
		A(0,6)=sym_os[8]-sym_os[5]-sym_os[4]+sym_os[0];
		A(0,7)=sym_os[8]-sym_os[5]-sym_os[4]+sym_os[0];
		A(0,8)=sym_os[8]-sym_os[5]-sym_os[4]+sym_os[0];
		A(1,0)=0;
		A(1,1)=sym_os[4]/2+sym_os[1]/2;
		A(1,2)=0;
		A(1,3)=sym_os[4]/2-sym_os[1]/2;
		A(1,4)=0;
		A(1,5)=(-sym_os[8]/2)-sym_os[7]/2+sym_os[4]/2+sym_os[1]/2;
		A(1,6)=(-sym_os[8]/2)+sym_os[7]/2+sym_os[4]/2-sym_os[1]/2;
		A(1,7)=(-sym_os[8]/2)+sym_os[7]/2+sym_os[4]/2-sym_os[1]/2;
		A(1,8)=(-sym_os[8]/2)-sym_os[7]/2+sym_os[4]/2+sym_os[1]/2;
		A(2,0)=0;
		A(2,1)=0;
		A(2,2)=sym_os[5]/2+sym_os[2]/2;
		A(2,3)=0;
		A(2,4)=sym_os[5]/2-sym_os[2]/2;
		A(2,5)=(-sym_os[8]/2)-sym_os[6]/2+sym_os[5]/2+sym_os[2]/2;
		A(2,6)=(-sym_os[8]/2)-sym_os[6]/2+sym_os[5]/2+sym_os[2]/2;
		A(2,7)=(-sym_os[8]/2)+sym_os[6]/2+sym_os[5]/2-sym_os[2]/2;
		A(2,8)=(-sym_os[8]/2)+sym_os[6]/2+sym_os[5]/2-sym_os[2]/2;
		A(3,0)=0;
		A(3,1)=sym_os[4]/2-sym_os[1]/2;
		A(3,2)=0;
		A(3,3)=sym_os[4]/2+sym_os[1]/2;
		A(3,4)=0;
		A(3,5)=(-sym_os[8]/2)+sym_os[7]/2+sym_os[4]/2-sym_os[1]/2;
		A(3,6)=(-sym_os[8]/2)-sym_os[7]/2+sym_os[4]/2+sym_os[1]/2;
		A(3,7)=(-sym_os[8]/2)-sym_os[7]/2+sym_os[4]/2+sym_os[1]/2;
		A(3,8)=(-sym_os[8]/2)+sym_os[7]/2+sym_os[4]/2-sym_os[1]/2;
		A(4,0)=0;
		A(4,1)=0;
		A(4,2)=sym_os[5]/2-sym_os[2]/2;
		A(4,3)=0;
		A(4,4)=sym_os[5]/2+sym_os[2]/2;
		A(4,5)=(-sym_os[8]/2)+sym_os[6]/2+sym_os[5]/2-sym_os[2]/2;
		A(4,6)=(-sym_os[8]/2)+sym_os[6]/2+sym_os[5]/2-sym_os[2]/2;
		A(4,7)=(-sym_os[8]/2)-sym_os[6]/2+sym_os[5]/2+sym_os[2]/2;
		A(4,8)=(-sym_os[8]/2)-sym_os[6]/2+sym_os[5]/2+sym_os[2]/2;
		A(5,0)=0;
		A(5,1)=0;
		A(5,2)=0;
		A(5,3)=0;
		A(5,4)=0;
		A(5,5)=sym_os[8]/4+sym_os[7]/4+sym_os[6]/4+sym_os[3]/4;
		A(5,6)=sym_os[8]/4-sym_os[7]/4+sym_os[6]/4-sym_os[3]/4;
		A(5,7)=sym_os[8]/4-sym_os[7]/4-sym_os[6]/4+sym_os[3]/4;
		A(5,8)=sym_os[8]/4+sym_os[7]/4-sym_os[6]/4-sym_os[3]/4;
		A(6,0)=0;
		A(6,1)=0;
		A(6,2)=0;
		A(6,3)=0;
		A(6,4)=0;
		A(6,5)=sym_os[8]/4-sym_os[7]/4+sym_os[6]/4-sym_os[3]/4;
		A(6,6)=sym_os[8]/4+sym_os[7]/4+sym_os[6]/4+sym_os[3]/4;
		A(6,7)=sym_os[8]/4+sym_os[7]/4-sym_os[6]/4-sym_os[3]/4;
		A(6,8)=sym_os[8]/4-sym_os[7]/4-sym_os[6]/4+sym_os[3]/4;
		A(7,0)=0;
		A(7,1)=0;
		A(7,2)=0;
		A(7,3)=0;
		A(7,4)=0;
		A(7,5)=sym_os[8]/4-sym_os[7]/4-sym_os[6]/4+sym_os[3]/4;
		A(7,6)=sym_os[8]/4+sym_os[7]/4-sym_os[6]/4-sym_os[3]/4;
		A(7,7)=sym_os[8]/4+sym_os[7]/4+sym_os[6]/4+sym_os[3]/4;
		A(7,8)=sym_os[8]/4-sym_os[7]/4+sym_os[6]/4-sym_os[3]/4;
		A(8,0)=0;
		A(8,1)=0;
		A(8,2)=0;
		A(8,3)=0;
		A(8,4)=0;
		A(8,5)=sym_os[8]/4+sym_os[7]/4-sym_os[6]/4-sym_os[3]/4;
		A(8,6)=sym_os[8]/4-sym_os[7]/4-sym_os[6]/4+sym_os[3]/4;
		A(8,7)=sym_os[8]/4-sym_os[7]/4+sym_os[6]/4-sym_os[3]/4;
		A(8,8)=sym_os[8]/4+sym_os[7]/4+sym_os[6]/4+sym_os[3]/4;
		break;
	case 15:
	case 27:
		setA_MRT_using_M(A,Q,u,v,w);
		break;
	default:
		printf("Error: matrix A not implemented for Q%d and MRT1\n",Q);
	}
}

void LBM::setA_MRT2(matrix &A, int Q, ex u, ex v, ex w)
{
	switch (Q)
	{
	case 5:
		A(0,0) = sym_os[0];
		A(0,1) = sym_os[0]-sym_os[3];
		A(0,2) = sym_os[0]-sym_os[3];
		A(0,3) = sym_os[0]-sym_os[3];
		A(0,4) = sym_os[0]-sym_os[3];
		A(1,0) = 0;
		A(1,1) = n1o4*sym_os[4]+n1o4*sym_os[3]+n1o2*sym_os[1];
		A(1,2) = n1o4*sym_os[3]-n1o4*sym_os[4];
		A(1,3) = n1o4*sym_os[4]+n1o4*sym_os[3]-n1o2*sym_os[1];
		A(1,4) = n1o4*sym_os[3]-n1o4*sym_os[4];
		A(2,0) = 0;
		A(2,1) = n1o4*sym_os[3]-n1o4*sym_os[4];
		A(2,2) = n1o4*sym_os[4]+n1o4*sym_os[3]+n1o2*sym_os[2];
		A(2,3) = n1o4*sym_os[3]-n1o4*sym_os[4];
		A(2,4) = n1o4*sym_os[4]+n1o4*sym_os[3]-n1o2*sym_os[2];
		A(3,0) = 0;
		A(3,1) = n1o4*sym_os[4]+n1o4*sym_os[3]-n1o2*sym_os[1];
		A(3,2) = n1o4*sym_os[3]-n1o4*sym_os[4];
		A(3,3) = n1o4*sym_os[4]+n1o4*sym_os[3]+n1o2*sym_os[1];
		A(3,4) = n1o4*sym_os[3]-n1o4*sym_os[4];
		A(4,0) = 0;
		A(4,1) = n1o4*sym_os[3]-n1o4*sym_os[4];
		A(4,2) = n1o4*sym_os[4]+n1o4*sym_os[3]-n1o2*sym_os[2];
		A(4,3) = n1o4*sym_os[4]+n1o4*sym_os[3];
		A(4,4) = n1o4*sym_os[4]+n1o4*sym_os[3]+n1o2*sym_os[2];
		break;
	case 7:
		A(0,0) = sym_os[0];
		A(0,1) = sym_os[0]-sym_os[4];
		A(0,2) = sym_os[0]-sym_os[4];
		A(0,3) = sym_os[0]-sym_os[4];
		A(0,4) = sym_os[0]-sym_os[4];
		A(0,5) = sym_os[0]-sym_os[4];
		A(0,6) = sym_os[0]-sym_os[4];

		A(1,0) = 0;
		A(1,1) = sym_os[6]/6+sym_os[5]/6+sym_os[4]/6+sym_os[1]/2;
		A(1,2) = sym_os[4]/6-sym_os[5]/6;
		A(1,3) = sym_os[4]/6-sym_os[6]/6;
		A(1,4) = sym_os[6]/6+sym_os[5]/6+sym_os[4]/6-sym_os[1]/2;
		A(1,5) = sym_os[4]/6-sym_os[5]/6;
		A(1,6) = sym_os[4]/6-sym_os[6]/6;

		A(2,0) = 0;
		A(2,1) = sym_os[6]/6-sym_os[5]/3+sym_os[4]/6;
		A(2,2) = sym_os[5]/3+sym_os[4]/6+sym_os[2]/2;
		A(2,3) = sym_os[4]/6-sym_os[6]/6;
		A(2,4) = sym_os[6]/6-sym_os[5]/3+sym_os[4]/6;
		A(2,5) = sym_os[5]/3+sym_os[4]/6-sym_os[2]/2;
		A(2,6) = sym_os[4]/6-sym_os[6]/6;

		A(3,0) = 0;
		A(3,1) = -sym_os[6]/3+sym_os[5]/6+sym_os[4]/6;
		A(3,2) = sym_os[4]/6-sym_os[5]/6;
		A(3,3) = sym_os[6]/3+sym_os[4]/6+sym_os[3]/2;
		A(3,4) = -sym_os[6]/3+sym_os[5]/6+sym_os[4]/6;
		A(3,5) = sym_os[4]/6-sym_os[5]/6;
		A(3,6) = sym_os[6]/3+sym_os[4]/6-sym_os[3]/2;

		A(4,0) = 0;
		A(4,1) = sym_os[6]/6+sym_os[5]/6+sym_os[4]/6-sym_os[1]/2;
		A(4,2) = sym_os[4]/6-sym_os[5]/6;
		A(4,3) = sym_os[4]/6-sym_os[6]/6;
		A(4,4) = sym_os[6]/6+sym_os[5]/6+sym_os[4]/6+sym_os[1]/2;
		A(4,5) = sym_os[4]/6-sym_os[5]/6;
		A(4,6) = sym_os[4]/6-sym_os[6]/6;

		A(5,0) = 0;
		A(5,1) = sym_os[6]/6-sym_os[5]/3+sym_os[4]/6;
		A(5,2) = sym_os[5]/3+sym_os[4]/6-sym_os[2]/2;
		A(5,3) = sym_os[4]/6-sym_os[6]/6;
		A(5,4) = sym_os[6]/6-sym_os[5]/3+sym_os[4]/6;
		A(5,5) = sym_os[5]/3+sym_os[4]/6+sym_os[2]/2;
		A(5,6) = sym_os[4]/6-sym_os[6]/6;

		A(6,0) = 0;
		A(6,1) = -sym_os[6]/3+sym_os[5]/6+sym_os[4]/6;
		A(6,2) = sym_os[4]/6-sym_os[5]/6;
		A(6,3) = sym_os[6]/3+sym_os[4]/6-sym_os[3]/2;
		A(6,4) = -sym_os[6]/3+sym_os[5]/6+sym_os[4]/6;
		A(6,5) = sym_os[4]/6-sym_os[5]/6;
		A(6,6) = sym_os[6]/3+sym_os[4]/6+sym_os[3]/2;
		break;
	default:
		printf("Error: matrix A not implemented for Q%d and MRT2\n",Q);
	}
}

void LBM::setA_CLBM1(matrix &A, int Q, ex u, ex v, ex w)
{
	ex one=1;
	switch (Q)
	{
	case 3:
		A(0,0) = (-sym_os[2]+2*sym_os[1]-sym_os[0])*u*u+sym_os[0];
		A(0,1) = (-sym_os[2]+2*sym_os[1]-sym_os[0])*u*u+(2*sym_os[2]-2*sym_os[1])*u-sym_os[2]+sym_os[0];
		A(0,2) = (-sym_os[2]+2*sym_os[1]-sym_os[0])*u*u+(2*sym_os[1]-2*sym_os[2])*u-sym_os[2]+sym_os[0];
		A(1,0) = n1o2*(sym_os[2]-2*sym_os[1]+sym_os[0])*u*u+n1o2*(sym_os[0]-sym_os[1])*u;
		A(1,1) = n1o2*(sym_os[2]-2*sym_os[1]+sym_os[0])*u*u+n1o2*(-2*sym_os[2]+sym_os[1]+sym_os[0])*u+n1o2*sym_os[2]+n1o2*sym_os[1];
		A(1,2) = n1o2*(sym_os[2]-2*sym_os[1]+sym_os[0])*u*u+n1o2*(2*sym_os[2]-3*sym_os[1]+sym_os[0])*u+n1o2*sym_os[2]-n1o2*sym_os[1];
		A(2,0) = n1o2*(sym_os[2]-2*sym_os[1]+sym_os[0])*u*u+n1o2*(sym_os[1]-sym_os[0])*u;
		A(2,1) = n1o2*(sym_os[2]-2*sym_os[1]+sym_os[0])*u*u+n1o2*(-2*sym_os[2]+3*sym_os[1]-sym_os[0])*u+n1o2*sym_os[2]-n1o2*sym_os[1];
		A(2,2) = n1o2*(sym_os[2]-2*sym_os[1]+sym_os[0])*u*u+n1o2*(2*sym_os[2]-sym_os[1]-sym_os[0])*u+n1o2*sym_os[2]+n1o2*sym_os[1];
		break;
	case 5:
		A(0,0) = (-sym_os[4]+2*sym_os[2]-sym_os[0])*v*v+(-sym_os[3]+2*sym_os[1]-sym_os[0])*u*u+sym_os[0];
		A(0,1) = (-sym_os[4]+2*sym_os[2]-sym_os[0])*v*v+(-sym_os[3]+2*sym_os[1]-sym_os[0])*u*u+(2*sym_os[3]-2*sym_os[1])*u-sym_os[3]+sym_os[0];
		A(0,2) = (-sym_os[4]+2*sym_os[2]-sym_os[0])*v*v+(2*sym_os[4]-2*sym_os[2])*v+(-sym_os[3]+2*sym_os[1]-sym_os[0])*u*u-sym_os[4]+sym_os[0];
		A(0,3) = (-sym_os[4]+2*sym_os[2]-sym_os[0])*v*v+(-sym_os[3]+2*sym_os[1]-sym_os[0])*u*u+(2*sym_os[1]-2*sym_os[3])*u-sym_os[3]+sym_os[0];
		A(0,4) = (-sym_os[4]+2*sym_os[2]-sym_os[0])*v*v+(2*sym_os[2]-2*sym_os[4])*v+(-sym_os[3]+2*sym_os[1]-sym_os[0])*u*u-sym_os[4]+sym_os[0];
		A(1,0) = n1o2*(sym_os[3]-2*sym_os[1]+sym_os[0])*u*u+n1o2*(sym_os[0]-sym_os[1])*u;
		A(1,1) = n1o2*(sym_os[3]-2*sym_os[1]+sym_os[0])*u*u+n1o2*(-2*sym_os[3]+sym_os[1]+sym_os[0])*u+n1o2*sym_os[3]+n1o2*sym_os[1];
		A(1,2) = n1o2*(sym_os[3]-2*sym_os[1]+sym_os[0])*u*u+n1o2*(sym_os[0]-sym_os[1])*u;
		A(1,3) = n1o2*(sym_os[3]-2*sym_os[1]+sym_os[0])*u*u+n1o2*(2*sym_os[3]-3*sym_os[1]+sym_os[0])*u+n1o2*sym_os[3]-n1o2*sym_os[1];
		A(1,4) = n1o2*(sym_os[3]-2*sym_os[1]+sym_os[0])*u*u+n1o2*(sym_os[0]-sym_os[1])*u;
		A(2,0) = n1o2*(sym_os[4]-2*sym_os[2]+sym_os[0])*v*v+n1o2*(sym_os[0]-sym_os[2])*v;
		A(2,1) = n1o2*(sym_os[4]-2*sym_os[2]+sym_os[0])*v*v+n1o2*(sym_os[0]-sym_os[2])*v;
		A(2,2) = n1o2*(sym_os[4]-2*sym_os[2]+sym_os[0])*v*v+n1o2*(-2*sym_os[4]+sym_os[2]+sym_os[0])*v+n1o2*sym_os[4]+n1o2*sym_os[2];
		A(2,3) = n1o2*(sym_os[4]-2*sym_os[2]+sym_os[0])*v*v+n1o2*(sym_os[0]-sym_os[2])*v;
		A(2,4) = n1o2*(sym_os[4]-2*sym_os[2]+sym_os[0])*v*v+n1o2*(2*sym_os[4]-3*sym_os[2]+sym_os[0])*v+n1o2*sym_os[4]-n1o2*sym_os[2];
		A(3,0) = n1o2*(sym_os[3]-2*sym_os[1]+sym_os[0])*u*u+n1o2*(sym_os[1]-sym_os[0])*u;
		A(3,1) = n1o2*(sym_os[3]-2*sym_os[1]+sym_os[0])*u*u+n1o2*(-2*sym_os[3]+3*sym_os[1]-sym_os[0])*u+n1o2*sym_os[3]-n1o2*sym_os[1];
		A(3,2) = n1o2*(sym_os[3]-2*sym_os[1]+sym_os[0])*u*u+n1o2*(sym_os[1]-sym_os[0])*u;
		A(3,3) = n1o2*(sym_os[3]-2*sym_os[1]+sym_os[0])*u*u+n1o2*(2*sym_os[3]-sym_os[1]-sym_os[0])*u+n1o2*sym_os[3]+n1o2*sym_os[1];
		A(3,4) = n1o2*(sym_os[3]-2*sym_os[1]+sym_os[0])*u*u+n1o2*(sym_os[1]-sym_os[0])*u;
		A(4,0) = n1o2*(sym_os[4]-2*sym_os[2]+sym_os[0])*v*v+n1o2*(sym_os[2]-sym_os[0])*v;
		A(4,1) = n1o2*(sym_os[4]-2*sym_os[2]+sym_os[0])*v*v+n1o2*(sym_os[2]-sym_os[0])*v;
		A(4,2) = n1o2*(sym_os[4]-2*sym_os[2]+sym_os[0])*v*v+n1o2*(-2*sym_os[4]+3*sym_os[2]-sym_os[0])*v+n1o2*sym_os[4]-n1o2*sym_os[2];
		A(4,3) = n1o2*(sym_os[4]-2*sym_os[2]+sym_os[0])*v*v+n1o2*(sym_os[2]-sym_os[0])*v;
		A(4,4) = n1o2*(sym_os[4]-2*sym_os[2]+sym_os[0])*v*v+n1o2*(2*sym_os[4]-sym_os[2]-sym_os[0])*v+n1o2*sym_os[4]+n1o2*sym_os[2];
		break;
	case 7:
		A(0,0) = (-sym_os[6]+2*sym_os[3]-sym_os[0])*w*w+(-sym_os[5]+2*sym_os[2]-sym_os[0])*v*v+(-sym_os[4]+2*sym_os[1]-sym_os[0])*u*u+sym_os[0];
		A(0,1) = (-sym_os[6]+2*sym_os[3]-sym_os[0])*w*w+(-sym_os[5]+2*sym_os[2]-sym_os[0])*v*v+(-sym_os[4]+2*sym_os[1]-sym_os[0])*u*u+(2*sym_os[4]-2*sym_os[1])*u-sym_os[4]+sym_os[0];
		A(0,2) = (-sym_os[6]+2*sym_os[3]-sym_os[0])*w*w+(-sym_os[5]+2*sym_os[2]-sym_os[0])*v*v+(2*sym_os[5]-2*sym_os[2])*v+(-sym_os[4]+2*sym_os[1]-sym_os[0])*u*u-sym_os[5]+sym_os[0];
		A(0,3) = (-sym_os[6]+2*sym_os[3]-sym_os[0])*w*w+(2*sym_os[6]-2*sym_os[3])*w+(-sym_os[5]+2*sym_os[2]-sym_os[0])*v*v+(-sym_os[4]+2*sym_os[1]-sym_os[0])*u*u-sym_os[6]+sym_os[0];
		A(0,4) = (-sym_os[6]+2*sym_os[3]-sym_os[0])*w*w+(-sym_os[5]+2*sym_os[2]-sym_os[0])*v*v+(-sym_os[4]+2*sym_os[1]-sym_os[0])*u*u+(2*sym_os[1]-2*sym_os[4])*u-sym_os[4]+sym_os[0];
		A(0,5) = (-sym_os[6]+2*sym_os[3]-sym_os[0])*w*w+(-sym_os[5]+2*sym_os[2]-sym_os[0])*v*v+(2*sym_os[2]-2*sym_os[5])*v+(-sym_os[4]+2*sym_os[1]-sym_os[0])*u*u-sym_os[5]+sym_os[0];
		A(0,6) = (-sym_os[6]+2*sym_os[3]-sym_os[0])*w*w+(2*sym_os[3]-2*sym_os[6])*w+(-sym_os[5]+2*sym_os[2]-sym_os[0])*v*v+(-sym_os[4]+2*sym_os[1]-sym_os[0])*u*u-sym_os[6]+sym_os[0];

		A(1,0) = ((sym_os[4]-2*sym_os[1]+sym_os[0])*u*u+(sym_os[0]-sym_os[1])*u)*n1o2;
		A(1,1) = ((sym_os[4]-2*sym_os[1]+sym_os[0])*u*u+(-2*sym_os[4]+sym_os[1]+sym_os[0])*u+sym_os[4]+sym_os[1])*n1o2;
		A(1,2) = ((sym_os[4]-2*sym_os[1]+sym_os[0])*u*u+(sym_os[0]-sym_os[1])*u)*n1o2;
		A(1,3) = ((sym_os[4]-2*sym_os[1]+sym_os[0])*u*u+(sym_os[0]-sym_os[1])*u)*n1o2;
		A(1,4) = ((sym_os[4]-2*sym_os[1]+sym_os[0])*u*u+(2*sym_os[4]-3*sym_os[1]+sym_os[0])*u+sym_os[4]-sym_os[1])*n1o2;
		A(1,5) = ((sym_os[4]-2*sym_os[1]+sym_os[0])*u*u+(sym_os[0]-sym_os[1])*u)*n1o2;
		A(1,6) = ((sym_os[4]-2*sym_os[1]+sym_os[0])*u*u+(sym_os[0]-sym_os[1])*u)*n1o2;

		A(2,0) = ((sym_os[5]-2*sym_os[2]+sym_os[0])*v*v+(sym_os[0]-sym_os[2])*v)*n1o2;
		A(2,1) = ((sym_os[5]-2*sym_os[2]+sym_os[0])*v*v+(sym_os[0]-sym_os[2])*v)*n1o2;
		A(2,2) = ((sym_os[5]-2*sym_os[2]+sym_os[0])*v*v+(-2*sym_os[5]+sym_os[2]+sym_os[0])*v+sym_os[5]+sym_os[2])*n1o2;
		A(2,3) = ((sym_os[5]-2*sym_os[2]+sym_os[0])*v*v+(sym_os[0]-sym_os[2])*v)*n1o2;
		A(2,4) = ((sym_os[5]-2*sym_os[2]+sym_os[0])*v*v+(sym_os[0]-sym_os[2])*v)*n1o2;
		A(2,5) = ((sym_os[5]-2*sym_os[2]+sym_os[0])*v*v+(2*sym_os[5]-3*sym_os[2]+sym_os[0])*v+sym_os[5]-sym_os[2])*n1o2;
		A(2,6) = ((sym_os[5]-2*sym_os[2]+sym_os[0])*v*v+(sym_os[0]-sym_os[2])*v)*n1o2;

		A(3,0) = ((sym_os[6]-2*sym_os[3]+sym_os[0])*w*w+(sym_os[0]-sym_os[3])*w)*n1o2;
		A(3,1) = ((sym_os[6]-2*sym_os[3]+sym_os[0])*w*w+(sym_os[0]-sym_os[3])*w)*n1o2;
		A(3,2) = ((sym_os[6]-2*sym_os[3]+sym_os[0])*w*w+(sym_os[0]-sym_os[3])*w)*n1o2;
		A(3,3) = ((sym_os[6]-2*sym_os[3]+sym_os[0])*w*w+(-2*sym_os[6]+sym_os[3]+sym_os[0])*w+sym_os[6]+sym_os[3])*n1o2;
		A(3,4) = ((sym_os[6]-2*sym_os[3]+sym_os[0])*w*w+(sym_os[0]-sym_os[3])*w)*n1o2;
		A(3,5) = ((sym_os[6]-2*sym_os[3]+sym_os[0])*w*w+(sym_os[0]-sym_os[3])*w)*n1o2;
		A(3,6) = ((sym_os[6]-2*sym_os[3]+sym_os[0])*w*w+(2*sym_os[6]-3*sym_os[3]+sym_os[0])*w+sym_os[6]-sym_os[3])*n1o2;

		A(4,0) = ((sym_os[4]-2*sym_os[1]+sym_os[0])*u*u+(sym_os[1]-sym_os[0])*u)*n1o2;
		A(4,1) = ((sym_os[4]-2*sym_os[1]+sym_os[0])*u*u+(-2*sym_os[4]+3*sym_os[1]-sym_os[0])*u+sym_os[4]-sym_os[1])*n1o2;
		A(4,2) = ((sym_os[4]-2*sym_os[1]+sym_os[0])*u*u+(sym_os[1]-sym_os[0])*u)*n1o2;
		A(4,3) = ((sym_os[4]-2*sym_os[1]+sym_os[0])*u*u+(sym_os[1]-sym_os[0])*u)*n1o2;
		A(4,4) = ((sym_os[4]-2*sym_os[1]+sym_os[0])*u*u+(2*sym_os[4]-sym_os[1]-sym_os[0])*u+sym_os[4]+sym_os[1])*n1o2;
		A(4,5) = ((sym_os[4]-2*sym_os[1]+sym_os[0])*u*u+(sym_os[1]-sym_os[0])*u)*n1o2;
		A(4,6) = ((sym_os[4]-2*sym_os[1]+sym_os[0])*u*u+(sym_os[1]-sym_os[0])*u)*n1o2;

		A(5,0) = ((sym_os[5]-2*sym_os[2]+sym_os[0])*v*v+(sym_os[2]-sym_os[0])*v)*n1o2;
		A(5,1) = ((sym_os[5]-2*sym_os[2]+sym_os[0])*v*v+(sym_os[2]-sym_os[0])*v)*n1o2;
		A(5,2) = ((sym_os[5]-2*sym_os[2]+sym_os[0])*v*v+(-2*sym_os[5]+3*sym_os[2]-sym_os[0])*v+sym_os[5]-sym_os[2])*n1o2;
		A(5,3) = ((sym_os[5]-2*sym_os[2]+sym_os[0])*v*v+(sym_os[2]-sym_os[0])*v)*n1o2;
		A(5,4) = ((sym_os[5]-2*sym_os[2]+sym_os[0])*v*v+(sym_os[2]-sym_os[0])*v)*n1o2;
		A(5,5) = ((sym_os[5]-2*sym_os[2]+sym_os[0])*v*v+(2*sym_os[5]-sym_os[2]-sym_os[0])*v+sym_os[5]+sym_os[2])*n1o2;
		A(5,6) = ((sym_os[5]-2*sym_os[2]+sym_os[0])*v*v+(sym_os[2]-sym_os[0])*v)*n1o2;

		A(6,0) = ((sym_os[6]-2*sym_os[3]+sym_os[0])*w*w+(sym_os[3]-sym_os[0])*w)*n1o2;
		A(6,1) = ((sym_os[6]-2*sym_os[3]+sym_os[0])*w*w+(sym_os[3]-sym_os[0])*w)*n1o2;
		A(6,2) = ((sym_os[6]-2*sym_os[3]+sym_os[0])*w*w+(sym_os[3]-sym_os[0])*w)*n1o2;
		A(6,3) = ((sym_os[6]-2*sym_os[3]+sym_os[0])*w*w+(-2*sym_os[6]+3*sym_os[3]-sym_os[0])*w+sym_os[6]-sym_os[3])*n1o2;
		A(6,4) = ((sym_os[6]-2*sym_os[3]+sym_os[0])*w*w+(sym_os[3]-sym_os[0])*w)*n1o2;
		A(6,5) = ((sym_os[6]-2*sym_os[3]+sym_os[0])*w*w+(sym_os[3]-sym_os[0])*w)*n1o2;
		A(6,6) = ((sym_os[6]-2*sym_os[3]+sym_os[0])*w*w+(2*sym_os[6]-sym_os[3]-sym_os[0])*w+sym_os[6]+sym_os[3])*n1o2;
		break;
	case 9:
		A(0,0)=-((sym_os[5]-sym_os[4])*v*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+2*sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u-2*sym_os[0])/2;
		A(0,1)=-((sym_os[5]-sym_os[4])*v*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[5]+2*sym_os[4]+8*sym_os[3]-4*sym_os[1])*u-2*sym_os[8]+4*sym_os[6]-sym_os[5]+sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[4]-2*sym_os[5])*u*u*u+(sym_os[5]+sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[1]-4*sym_os[4])*u+2*sym_os[4]-2*sym_os[0])/2;
		A(0,2)=-((sym_os[5]-sym_os[4])*v*v*v*v+(2*sym_os[4]-2*sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+sym_os[5]+sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+((4*sym_os[8]-8*sym_os[7]-4*sym_os[6]+2*sym_os[5]+2*sym_os[4]+8*sym_os[3]-4*sym_os[2])*u*u-4*sym_os[4]+4*sym_os[2])*v+(sym_os[5]-sym_os[4])*u*u*u*u+((-2*sym_os[8])+4*sym_os[7]-sym_os[5]+sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u+2*sym_os[4]-2*sym_os[0])/2;
		A(0,3)=-((sym_os[5]-sym_os[4])*v*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[1])*u-2*sym_os[8]+4*sym_os[6]-sym_os[5]+sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[5]-2*sym_os[4])*u*u*u+(sym_os[5]+sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[4]-4*sym_os[1])*u+2*sym_os[4]-2*sym_os[0])/2;
		A(0,4)=-((sym_os[5]-sym_os[4])*v*v*v*v+(2*sym_os[5]-2*sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+sym_os[5]+sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+(((-4*sym_os[8])+8*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2])*u*u+4*sym_os[4]-4*sym_os[2])*v+(sym_os[5]-sym_os[4])*u*u*u*u+((-2*sym_os[8])+4*sym_os[7]-sym_os[5]+sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u+2*sym_os[4]-2*sym_os[0])/2;
		A(0,5)=-((sym_os[5]-sym_os[4])*v*v*v*v+(2*sym_os[4]-2*sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[5]+2*sym_os[4]+8*sym_os[3]-4*sym_os[1])*u-2*sym_os[8]+4*sym_os[6]-4*sym_os[2]+2*sym_os[0])*v*v+((4*sym_os[8]-8*sym_os[7]-4*sym_os[6]+2*sym_os[5]+2*sym_os[4]+8*sym_os[3]-4*sym_os[2])*u*u+((-8*sym_os[8])+8*sym_os[7]+8*sym_os[6]-8*sym_os[3])*u+4*sym_os[8]-4*sym_os[6]-4*sym_os[4]+4*sym_os[2])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[4]-2*sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-4*sym_os[4]+4*sym_os[1])*u-2*sym_os[8]+4*sym_os[4]-2*sym_os[0])/2;
		A(0,6)=-((sym_os[5]-sym_os[4])*v*v*v*v+(2*sym_os[4]-2*sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[1])*u-2*sym_os[8]+4*sym_os[6]-4*sym_os[2]+2*sym_os[0])*v*v+((4*sym_os[8]-8*sym_os[7]-4*sym_os[6]+2*sym_os[5]+2*sym_os[4]+8*sym_os[3]-4*sym_os[2])*u*u+(8*sym_os[8]-8*sym_os[7]-8*sym_os[6]+8*sym_os[3])*u+4*sym_os[8]-4*sym_os[6]-4*sym_os[4]+4*sym_os[2])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[5]-2*sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+4*sym_os[4]-4*sym_os[1])*u-2*sym_os[8]+4*sym_os[4]-2*sym_os[0])/2;
		A(0,7)=-((sym_os[5]-sym_os[4])*v*v*v*v+(2*sym_os[5]-2*sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[1])*u-2*sym_os[8]+4*sym_os[6]-4*sym_os[2]+2*sym_os[0])*v*v+(((-4*sym_os[8])+8*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2])*u*u+((-8*sym_os[8])+8*sym_os[7]+8*sym_os[6]-8*sym_os[3])*u-4*sym_os[8]+4*sym_os[6]+4*sym_os[4]-4*sym_os[2])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[5]-2*sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+4*sym_os[4]-4*sym_os[1])*u-2*sym_os[8]+4*sym_os[4]-2*sym_os[0])/2;
		A(0,8)=-((sym_os[5]-sym_os[4])*v*v*v*v+(2*sym_os[5]-2*sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[5]+2*sym_os[4]+8*sym_os[3]-4*sym_os[1])*u-2*sym_os[8]+4*sym_os[6]-4*sym_os[2]+2*sym_os[0])*v*v+(((-4*sym_os[8])+8*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2])*u*u+(8*sym_os[8]-8*sym_os[7]-8*sym_os[6]+8*sym_os[3])*u-4*sym_os[8]+4*sym_os[6]+4*sym_os[4]-4*sym_os[2])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[4]-2*sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-4*sym_os[4]+4*sym_os[1])*u-2*sym_os[8]+4*sym_os[4]-2*sym_os[0])/2;
		A(1,0)=((sym_os[5]-sym_os[4])*v*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(2*sym_os[7]-sym_os[5]-sym_os[4]-4*sym_os[3]+4*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u-sym_os[5]+sym_os[4])*v*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[5]-sym_os[4])*u*u*u+(sym_os[5]+sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u+(2*sym_os[0]-2*sym_os[1])*u)/4;
		A(1,1)=((sym_os[5]-sym_os[4])*v*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-8*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[3]+4*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+4*sym_os[6]-2*sym_os[5]+4*sym_os[3]-2*sym_os[1])*v*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u*u+(2*sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u+((-sym_os[5])-3*sym_os[4]+2*sym_os[1]+2*sym_os[0])*u+sym_os[5]+sym_os[4]+2*sym_os[1])/4;
		A(1,2)=((sym_os[5]-sym_os[4])*v*v*v*v+(2*sym_os[4]-2*sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(2*sym_os[7]-sym_os[5]-sym_os[4]-4*sym_os[3]+4*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u)*v*v+((4*sym_os[8]-8*sym_os[7]-4*sym_os[6]+2*sym_os[5]+2*sym_os[4]+8*sym_os[3]-4*sym_os[2])*u*u+((-4*sym_os[7])+2*sym_os[5]+2*sym_os[4]+4*sym_os[3]-4*sym_os[2])*u+2*sym_os[5]-2*sym_os[4])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[5]-sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-4*sym_os[1]+2*sym_os[0])*u*u+(2*sym_os[7]-sym_os[5]-sym_os[4]-2*sym_os[1]+2*sym_os[0])*u-sym_os[5]+sym_os[4])/4;
		A(1,3)=((sym_os[5]-sym_os[4])*v*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+8*sym_os[6]-3*sym_os[5]-3*sym_os[4]-12*sym_os[3]+4*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+4*sym_os[6]-2*sym_os[5]-4*sym_os[3]+2*sym_os[1])*v*v+(sym_os[5]-sym_os[4])*u*u*u*u+(3*sym_os[5]-3*sym_os[4])*u*u*u+(4*sym_os[5]-2*sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u+(3*sym_os[5]+sym_os[4]-6*sym_os[1]+2*sym_os[0])*u+sym_os[5]+sym_os[4]-2*sym_os[1])/4;
		A(1,4)=((sym_os[5]-sym_os[4])*v*v*v*v+(2*sym_os[5]-2*sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(2*sym_os[7]-sym_os[5]-sym_os[4]-4*sym_os[3]+4*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u)*v*v+(((-4*sym_os[8])+8*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2])*u*u+(4*sym_os[7]-2*sym_os[5]-2*sym_os[4]-4*sym_os[3]+4*sym_os[2])*u-2*sym_os[5]+2*sym_os[4])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[5]-sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-4*sym_os[1]+2*sym_os[0])*u*u+(2*sym_os[7]-sym_os[5]-sym_os[4]-2*sym_os[1]+2*sym_os[0])*u-sym_os[5]+sym_os[4])/4;
		A(1,5)=((sym_os[5]-sym_os[4])*v*v*v*v+(2*sym_os[4]-2*sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-8*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[3]+4*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+4*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-4*sym_os[6]+2*sym_os[5]+2*sym_os[4]+8*sym_os[3]-4*sym_os[2])*u*u+((-8*sym_os[8])+4*sym_os[7]+8*sym_os[6]+2*sym_os[5]+2*sym_os[4]-4*sym_os[3]-4*sym_os[2])*u+4*sym_os[8]+4*sym_os[7]-4*sym_os[6]+2*sym_os[5]-2*sym_os[4]-4*sym_os[3])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-sym_os[5]+sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-2*sym_os[5]-4*sym_os[4]+2*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+2*sym_os[4]+2*sym_os[1])/4;
		A(1,6)=((sym_os[5]-sym_os[4])*v*v*v*v+(2*sym_os[4]-2*sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+8*sym_os[6]-3*sym_os[5]-3*sym_os[4]-12*sym_os[3]+4*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+4*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-4*sym_os[6]+2*sym_os[5]+2*sym_os[4]+8*sym_os[3]-4*sym_os[2])*u*u+(8*sym_os[8]-12*sym_os[7]-8*sym_os[6]+2*sym_os[5]+2*sym_os[4]+12*sym_os[3]-4*sym_os[2])*u+4*sym_os[8]-4*sym_os[7]-4*sym_os[6]+2*sym_os[5]-2*sym_os[4]+4*sym_os[3])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(3*sym_os[5]-3*sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+3*sym_os[5]-3*sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+2*sym_os[5]-6*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+2*sym_os[4]-2*sym_os[1])/4;
		A(1,7)=((sym_os[5]-sym_os[4])*v*v*v*v+(2*sym_os[5]-2*sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+8*sym_os[6]-3*sym_os[5]-3*sym_os[4]-12*sym_os[3]+4*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+4*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2])*u*u+((-8*sym_os[8])+12*sym_os[7]+8*sym_os[6]-2*sym_os[5]-2*sym_os[4]-12*sym_os[3]+4*sym_os[2])*u-4*sym_os[8]+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]+2*sym_os[4]-4*sym_os[3])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(3*sym_os[5]-3*sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+3*sym_os[5]-3*sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+2*sym_os[5]-6*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+2*sym_os[4]-2*sym_os[1])/4;
		A(1,8)=((sym_os[5]-sym_os[4])*v*v*v*v+(2*sym_os[5]-2*sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-8*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[3]+4*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+4*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2])*u*u+(8*sym_os[8]-4*sym_os[7]-8*sym_os[6]-2*sym_os[5]-2*sym_os[4]+4*sym_os[3]+4*sym_os[2])*u-4*sym_os[8]-4*sym_os[7]+4*sym_os[6]-2*sym_os[5]+2*sym_os[4]+4*sym_os[3])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-sym_os[5]+sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-2*sym_os[5]-4*sym_os[4]+2*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+2*sym_os[4]+2*sym_os[1])/4;
		A(2,0)=((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[5]-sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+sym_os[5]+sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+((2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u-2*sym_os[2]+2*sym_os[0])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u)/4;
		A(2,1)=((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[5]-sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[5]+2*sym_os[4]+8*sym_os[3]-4*sym_os[1])*u-2*sym_os[8]+4*sym_os[6]-4*sym_os[2]+2*sym_os[0])*v*v+((2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[6])+2*sym_os[5]+2*sym_os[4]+4*sym_os[3]-4*sym_os[1])*u+2*sym_os[6]-sym_os[5]-sym_os[4]-2*sym_os[2]+2*sym_os[0])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[4]-2*sym_os[5])*u*u*u+(2*sym_os[5]-2*sym_os[4])*u-sym_os[5]+sym_os[4])/4;
		A(2,2)=((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[4]-sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+2*sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+((4*sym_os[8]-8*sym_os[7]-2*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u-sym_os[5]-3*sym_os[4]+2*sym_os[2]+2*sym_os[0])*v+(sym_os[5]-sym_os[4])*u*u*u*u+((-2*sym_os[8])+4*sym_os[7]-2*sym_os[6]-2*sym_os[5]+4*sym_os[3]-2*sym_os[2])*u*u+sym_os[5]+sym_os[4]+2*sym_os[2])/4;
		A(2,3)=((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[5]-sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[1])*u-2*sym_os[8]+4*sym_os[6]-4*sym_os[2]+2*sym_os[0])*v*v+((2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-4*sym_os[3]+4*sym_os[1])*u+2*sym_os[6]-sym_os[5]-sym_os[4]-2*sym_os[2]+2*sym_os[0])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[5]-2*sym_os[4])*u*u*u+(2*sym_os[4]-2*sym_os[5])*u-sym_os[5]+sym_os[4])/4;
		A(2,4)=((sym_os[5]-sym_os[4])*v*v*v*v+(3*sym_os[5]-3*sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+4*sym_os[5]-2*sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+(((-4*sym_os[8])+8*sym_os[7]+6*sym_os[6]-3*sym_os[5]-3*sym_os[4]-12*sym_os[3]+6*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+3*sym_os[5]+sym_os[4]-6*sym_os[2]+2*sym_os[0])*v+(sym_os[5]-sym_os[4])*u*u*u*u+((-2*sym_os[8])+4*sym_os[7]+2*sym_os[6]-2*sym_os[5]-4*sym_os[3]+2*sym_os[2])*u*u+sym_os[5]+sym_os[4]-2*sym_os[2])/4;
		A(2,5)=((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[4]-sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[5]+2*sym_os[4]+8*sym_os[3]-4*sym_os[1])*u-2*sym_os[8]+4*sym_os[6]-sym_os[5]+sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+((4*sym_os[8]-8*sym_os[7]-2*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-8*sym_os[8])+8*sym_os[7]+4*sym_os[6]+2*sym_os[5]+2*sym_os[4]-4*sym_os[3]-4*sym_os[1])*u+4*sym_os[8]-2*sym_os[6]-2*sym_os[5]-4*sym_os[4]+2*sym_os[2]+2*sym_os[0])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[4]-2*sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-2*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+(4*sym_os[8]-4*sym_os[7]+4*sym_os[6]+2*sym_os[5]-2*sym_os[4]-4*sym_os[3])*u-2*sym_os[8]-2*sym_os[6]+2*sym_os[4]+2*sym_os[2])/4;
		A(2,6)=((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[4]-sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[1])*u-2*sym_os[8]+4*sym_os[6]-sym_os[5]+sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+((4*sym_os[8]-8*sym_os[7]-2*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(8*sym_os[8]-8*sym_os[7]-4*sym_os[6]-2*sym_os[5]-2*sym_os[4]+4*sym_os[3]+4*sym_os[1])*u+4*sym_os[8]-2*sym_os[6]-2*sym_os[5]-4*sym_os[4]+2*sym_os[2]+2*sym_os[0])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[5]-2*sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-2*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+((-4*sym_os[8])+4*sym_os[7]-4*sym_os[6]-2*sym_os[5]+2*sym_os[4]+4*sym_os[3])*u-2*sym_os[8]-2*sym_os[6]+2*sym_os[4]+2*sym_os[2])/4;
		A(2,7)=((sym_os[5]-sym_os[4])*v*v*v*v+(3*sym_os[5]-3*sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[1])*u-2*sym_os[8]+4*sym_os[6]+3*sym_os[5]-3*sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+(((-4*sym_os[8])+8*sym_os[7]+6*sym_os[6]-3*sym_os[5]-3*sym_os[4]-12*sym_os[3]+6*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-8*sym_os[8])+8*sym_os[7]+12*sym_os[6]-2*sym_os[5]-2*sym_os[4]-12*sym_os[3]+4*sym_os[1])*u-4*sym_os[8]+6*sym_os[6]+2*sym_os[5]-6*sym_os[2]+2*sym_os[0])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[5]-2*sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+((-4*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]+2*sym_os[4]-4*sym_os[3])*u-2*sym_os[8]+2*sym_os[6]+2*sym_os[4]-2*sym_os[2])/4;
		A(2,8)=((sym_os[5]-sym_os[4])*v*v*v*v+(3*sym_os[5]-3*sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[5]+2*sym_os[4]+8*sym_os[3]-4*sym_os[1])*u-2*sym_os[8]+4*sym_os[6]+3*sym_os[5]-3*sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+(((-4*sym_os[8])+8*sym_os[7]+6*sym_os[6]-3*sym_os[5]-3*sym_os[4]-12*sym_os[3]+6*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(8*sym_os[8]-8*sym_os[7]-12*sym_os[6]+2*sym_os[5]+2*sym_os[4]+12*sym_os[3]-4*sym_os[1])*u-4*sym_os[8]+6*sym_os[6]+2*sym_os[5]-6*sym_os[2]+2*sym_os[0])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[4]-2*sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+(4*sym_os[8]-4*sym_os[7]-4*sym_os[6]+2*sym_os[5]-2*sym_os[4]+4*sym_os[3])*u-2*sym_os[8]+2*sym_os[6]+2*sym_os[4]-2*sym_os[2])/4;
		A(3,0)=((sym_os[5]-sym_os[4])*v*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-2*sym_os[7])+sym_os[5]+sym_os[4]+4*sym_os[3]-4*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u-sym_os[5]+sym_os[4])*v*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u*u+(sym_os[5]+sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u+(2*sym_os[1]-2*sym_os[0])*u)/4;
		A(3,1)=((sym_os[5]-sym_os[4])*v*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-8*sym_os[6]+3*sym_os[5]+3*sym_os[4]+12*sym_os[3]-4*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+4*sym_os[6]-2*sym_os[5]-4*sym_os[3]+2*sym_os[1])*v*v+(sym_os[5]-sym_os[4])*u*u*u*u+(3*sym_os[4]-3*sym_os[5])*u*u*u+(4*sym_os[5]-2*sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u+((-3*sym_os[5])-sym_os[4]+6*sym_os[1]-2*sym_os[0])*u+sym_os[5]+sym_os[4]-2*sym_os[1])/4;
		A(3,2)=((sym_os[5]-sym_os[4])*v*v*v*v+(2*sym_os[4]-2*sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-2*sym_os[7])+sym_os[5]+sym_os[4]+4*sym_os[3]-4*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u)*v*v+((4*sym_os[8]-8*sym_os[7]-4*sym_os[6]+2*sym_os[5]+2*sym_os[4]+8*sym_os[3]-4*sym_os[2])*u*u+(4*sym_os[7]-2*sym_os[5]-2*sym_os[4]-4*sym_os[3]+4*sym_os[2])*u+2*sym_os[5]-2*sym_os[4])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-4*sym_os[1]+2*sym_os[0])*u*u+((-2*sym_os[7])+sym_os[5]+sym_os[4]+2*sym_os[1]-2*sym_os[0])*u-sym_os[5]+sym_os[4])/4;
		A(3,3)=((sym_os[5]-sym_os[4])*v*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+8*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]-4*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+4*sym_os[6]-2*sym_os[5]+4*sym_os[3]-2*sym_os[1])*v*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[5]-sym_os[4])*u*u*u+(2*sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u+(sym_os[5]+3*sym_os[4]-2*sym_os[1]-2*sym_os[0])*u+sym_os[5]+sym_os[4]+2*sym_os[1])/4;
		A(3,4)=((sym_os[5]-sym_os[4])*v*v*v*v+(2*sym_os[5]-2*sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-2*sym_os[7])+sym_os[5]+sym_os[4]+4*sym_os[3]-4*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u)*v*v+(((-4*sym_os[8])+8*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2])*u*u+((-4*sym_os[7])+2*sym_os[5]+2*sym_os[4]+4*sym_os[3]-4*sym_os[2])*u-2*sym_os[5]+2*sym_os[4])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-4*sym_os[1]+2*sym_os[0])*u*u+((-2*sym_os[7])+sym_os[5]+sym_os[4]+2*sym_os[1]-2*sym_os[0])*u-sym_os[5]+sym_os[4])/4;
		A(3,5)=((sym_os[5]-sym_os[4])*v*v*v*v+(2*sym_os[4]-2*sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-8*sym_os[6]+3*sym_os[5]+3*sym_os[4]+12*sym_os[3]-4*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+4*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-4*sym_os[6]+2*sym_os[5]+2*sym_os[4]+8*sym_os[3]-4*sym_os[2])*u*u+((-8*sym_os[8])+12*sym_os[7]+8*sym_os[6]-2*sym_os[5]-2*sym_os[4]-12*sym_os[3]+4*sym_os[2])*u+4*sym_os[8]-4*sym_os[7]-4*sym_os[6]+2*sym_os[5]-2*sym_os[4]+4*sym_os[3])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(3*sym_os[4]-3*sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+3*sym_os[5]-3*sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-2*sym_os[5]+6*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+2*sym_os[4]-2*sym_os[1])/4;
		A(3,6)=((sym_os[5]-sym_os[4])*v*v*v*v+(2*sym_os[4]-2*sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+8*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]-4*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+4*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-4*sym_os[6]+2*sym_os[5]+2*sym_os[4]+8*sym_os[3]-4*sym_os[2])*u*u+(8*sym_os[8]-4*sym_os[7]-8*sym_os[6]-2*sym_os[5]-2*sym_os[4]+4*sym_os[3]+4*sym_os[2])*u+4*sym_os[8]+4*sym_os[7]-4*sym_os[6]+2*sym_os[5]-2*sym_os[4]-4*sym_os[3])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[5]-sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-sym_os[5]+sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+2*sym_os[5]+4*sym_os[4]-2*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+2*sym_os[4]+2*sym_os[1])/4;
		A(3,7)=((sym_os[5]-sym_os[4])*v*v*v*v+(2*sym_os[5]-2*sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+8*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]-4*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+4*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2])*u*u+((-8*sym_os[8])+4*sym_os[7]+8*sym_os[6]+2*sym_os[5]+2*sym_os[4]-4*sym_os[3]-4*sym_os[2])*u-4*sym_os[8]-4*sym_os[7]+4*sym_os[6]-2*sym_os[5]+2*sym_os[4]+4*sym_os[3])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[5]-sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-sym_os[5]+sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+2*sym_os[5]+4*sym_os[4]-2*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+2*sym_os[4]+2*sym_os[1])/4;
		A(3,8)=((sym_os[5]-sym_os[4])*v*v*v*v+(2*sym_os[5]-2*sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-8*sym_os[6]+3*sym_os[5]+3*sym_os[4]+12*sym_os[3]-4*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+4*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2])*u*u+(8*sym_os[8]-12*sym_os[7]-8*sym_os[6]+2*sym_os[5]+2*sym_os[4]+12*sym_os[3]-4*sym_os[2])*u-4*sym_os[8]+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]+2*sym_os[4]-4*sym_os[3])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(3*sym_os[4]-3*sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+3*sym_os[5]-3*sym_os[4]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-2*sym_os[5]+6*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+2*sym_os[4]-2*sym_os[1])/4;
		A(4,0)=((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[4]-sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+sym_os[5]+sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+(((-2*sym_os[6])+sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+2*sym_os[2]-2*sym_os[0])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u)/4;
		A(4,1)=((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[4]-sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[5]+2*sym_os[4]+8*sym_os[3]-4*sym_os[1])*u-2*sym_os[8]+4*sym_os[6]-4*sym_os[2]+2*sym_os[0])*v*v+(((-2*sym_os[6])+sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-4*sym_os[3]+4*sym_os[1])*u-2*sym_os[6]+sym_os[5]+sym_os[4]+2*sym_os[2]-2*sym_os[0])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[4]-2*sym_os[5])*u*u*u+(2*sym_os[5]-2*sym_os[4])*u-sym_os[5]+sym_os[4])/4;
		A(4,2)=((sym_os[5]-sym_os[4])*v*v*v*v+(3*sym_os[4]-3*sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+4*sym_os[5]-2*sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+((4*sym_os[8]-8*sym_os[7]-6*sym_os[6]+3*sym_os[5]+3*sym_os[4]+12*sym_os[3]-6*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u-3*sym_os[5]-sym_os[4]+6*sym_os[2]-2*sym_os[0])*v+(sym_os[5]-sym_os[4])*u*u*u*u+((-2*sym_os[8])+4*sym_os[7]+2*sym_os[6]-2*sym_os[5]-4*sym_os[3]+2*sym_os[2])*u*u+sym_os[5]+sym_os[4]-2*sym_os[2])/4;
		A(4,3)=((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[4]-sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[1])*u-2*sym_os[8]+4*sym_os[6]-4*sym_os[2]+2*sym_os[0])*v*v+(((-2*sym_os[6])+sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[6])+2*sym_os[5]+2*sym_os[4]+4*sym_os[3]-4*sym_os[1])*u-2*sym_os[6]+sym_os[5]+sym_os[4]+2*sym_os[2]-2*sym_os[0])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[5]-2*sym_os[4])*u*u*u+(2*sym_os[4]-2*sym_os[5])*u-sym_os[5]+sym_os[4])/4;
		A(4,4)=((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[5]-sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+2*sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+(((-4*sym_os[8])+8*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+sym_os[5]+3*sym_os[4]-2*sym_os[2]-2*sym_os[0])*v+(sym_os[5]-sym_os[4])*u*u*u*u+((-2*sym_os[8])+4*sym_os[7]-2*sym_os[6]-2*sym_os[5]+4*sym_os[3]-2*sym_os[2])*u*u+sym_os[5]+sym_os[4]+2*sym_os[2])/4;
		A(4,5)=((sym_os[5]-sym_os[4])*v*v*v*v+(3*sym_os[4]-3*sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[5]+2*sym_os[4]+8*sym_os[3]-4*sym_os[1])*u-2*sym_os[8]+4*sym_os[6]+3*sym_os[5]-3*sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+((4*sym_os[8]-8*sym_os[7]-6*sym_os[6]+3*sym_os[5]+3*sym_os[4]+12*sym_os[3]-6*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-8*sym_os[8])+8*sym_os[7]+12*sym_os[6]-2*sym_os[5]-2*sym_os[4]-12*sym_os[3]+4*sym_os[1])*u+4*sym_os[8]-6*sym_os[6]-2*sym_os[5]+6*sym_os[2]-2*sym_os[0])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[4]-2*sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+(4*sym_os[8]-4*sym_os[7]-4*sym_os[6]+2*sym_os[5]-2*sym_os[4]+4*sym_os[3])*u-2*sym_os[8]+2*sym_os[6]+2*sym_os[4]-2*sym_os[2])/4;
		A(4,6)=((sym_os[5]-sym_os[4])*v*v*v*v+(3*sym_os[4]-3*sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[1])*u-2*sym_os[8]+4*sym_os[6]+3*sym_os[5]-3*sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+((4*sym_os[8]-8*sym_os[7]-6*sym_os[6]+3*sym_os[5]+3*sym_os[4]+12*sym_os[3]-6*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(8*sym_os[8]-8*sym_os[7]-12*sym_os[6]+2*sym_os[5]+2*sym_os[4]+12*sym_os[3]-4*sym_os[1])*u+4*sym_os[8]-6*sym_os[6]-2*sym_os[5]+6*sym_os[2]-2*sym_os[0])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[5]-2*sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+((-4*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]+2*sym_os[4]-4*sym_os[3])*u-2*sym_os[8]+2*sym_os[6]+2*sym_os[4]-2*sym_os[2])/4;
		A(4,7)=((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[5]-sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[1])*u-2*sym_os[8]+4*sym_os[6]-sym_os[5]+sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+(((-4*sym_os[8])+8*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-8*sym_os[8])+8*sym_os[7]+4*sym_os[6]+2*sym_os[5]+2*sym_os[4]-4*sym_os[3]-4*sym_os[1])*u-4*sym_os[8]+2*sym_os[6]+2*sym_os[5]+4*sym_os[4]-2*sym_os[2]-2*sym_os[0])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[5]-2*sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-2*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+((-4*sym_os[8])+4*sym_os[7]-4*sym_os[6]-2*sym_os[5]+2*sym_os[4]+4*sym_os[3])*u-2*sym_os[8]-2*sym_os[6]+2*sym_os[4]+2*sym_os[2])/4;
		A(4,8)=((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[5]-sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[5]+2*sym_os[4]+8*sym_os[3]-4*sym_os[1])*u-2*sym_os[8]+4*sym_os[6]-sym_os[5]+sym_os[4]-4*sym_os[2]+2*sym_os[0])*v*v+(((-4*sym_os[8])+8*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(8*sym_os[8]-8*sym_os[7]-4*sym_os[6]-2*sym_os[5]-2*sym_os[4]+4*sym_os[3]+4*sym_os[1])*u-4*sym_os[8]+2*sym_os[6]+2*sym_os[5]+4*sym_os[4]-2*sym_os[2]-2*sym_os[0])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(2*sym_os[4]-2*sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-2*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+(4*sym_os[8]-4*sym_os[7]+4*sym_os[6]+2*sym_os[5]-2*sym_os[4]-4*sym_os[3])*u-2*sym_os[8]-2*sym_os[6]+2*sym_os[4]+2*sym_os[2])/4;
		A(5,0)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[5]-sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(2*sym_os[7]-sym_os[5]-sym_os[4]-4*sym_os[3]+4*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u)*v*v+((2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-2*sym_os[3])+2*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u)*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[5]-sym_os[4])*u*u*u)/8;
		A(5,1)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[5]-sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-8*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[3]+4*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+4*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+((2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[6])+2*sym_os[5]+2*sym_os[4]+2*sym_os[3]+2*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u+2*sym_os[6]-sym_os[5]-sym_os[4]+2*sym_os[3]-2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u*u+(sym_os[4]-sym_os[5])*u*u+(sym_os[5]-sym_os[4])*u)/8;
		A(5,2)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[4]-sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(2*sym_os[7]-sym_os[5]-sym_os[4]-4*sym_os[3]+4*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u-sym_os[5]+sym_os[4])*v*v+((4*sym_os[8]-8*sym_os[7]-2*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[7])+2*sym_os[5]+2*sym_os[4]+2*sym_os[3]-2*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u+sym_os[5]-sym_os[4])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[5]-sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-2*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+(2*sym_os[7]-sym_os[5]-sym_os[4]+2*sym_os[3]-2*sym_os[2])*u)/8;
		A(5,3)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[5]-sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+8*sym_os[6]-3*sym_os[5]-3*sym_os[4]-12*sym_os[3]+4*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+4*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+((2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-6*sym_os[3]+2*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u+2*sym_os[6]-sym_os[5]-sym_os[4]-2*sym_os[3]+2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(3*sym_os[5]-3*sym_os[4])*u*u*u+(3*sym_os[5]-3*sym_os[4])*u*u+(sym_os[5]-sym_os[4])*u)/8;
		A(5,4)=-((sym_os[5]-sym_os[4])*v*v*v*v+(3*sym_os[5]-3*sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(2*sym_os[7]-sym_os[5]-sym_os[4]-4*sym_os[3]+4*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u+3*sym_os[5]-3*sym_os[4])*v*v+(((-4*sym_os[8])+8*sym_os[7]+6*sym_os[6]-3*sym_os[5]-3*sym_os[4]-12*sym_os[3]+6*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[7]-2*sym_os[5]-2*sym_os[4]-6*sym_os[3]+6*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u+sym_os[5]-sym_os[4])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[5]-sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+(2*sym_os[7]-sym_os[5]-sym_os[4]-2*sym_os[3]+2*sym_os[2])*u)/8;
		A(5,5)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[4]-sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-8*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[3]+4*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+4*sym_os[6]-2*sym_os[5]+4*sym_os[3]-2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-2*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-8*sym_os[8])+4*sym_os[7]+4*sym_os[6]+4*sym_os[5]+4*sym_os[4]-2*sym_os[3]-2*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u+4*sym_os[8]+4*sym_os[7]-2*sym_os[6]-2*sym_os[4]-2*sym_os[3]-2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-2*sym_os[6]-2*sym_os[5]+4*sym_os[3]-2*sym_os[2])*u*u+(4*sym_os[8]-2*sym_os[7]+4*sym_os[6]-2*sym_os[4]-2*sym_os[3]-2*sym_os[2])*u-2*sym_os[8]-2*sym_os[7]-2*sym_os[6]-2*sym_os[3])/8;
		A(5,6)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[4]-sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+8*sym_os[6]-3*sym_os[5]-3*sym_os[4]-12*sym_os[3]+4*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+4*sym_os[6]-2*sym_os[5]-4*sym_os[3]+2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-2*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(8*sym_os[8]-12*sym_os[7]-4*sym_os[6]+6*sym_os[3]-2*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u+4*sym_os[8]-4*sym_os[7]-2*sym_os[6]-2*sym_os[4]+2*sym_os[3]+2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(3*sym_os[5]-3*sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-2*sym_os[6]+2*sym_os[5]-4*sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+((-4*sym_os[8])+6*sym_os[7]-4*sym_os[6]-2*sym_os[4]+6*sym_os[3]-2*sym_os[2])*u-2*sym_os[8]+2*sym_os[7]-2*sym_os[6]+2*sym_os[3])/8;
		A(5,7)=-((sym_os[5]-sym_os[4])*v*v*v*v+(3*sym_os[5]-3*sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+8*sym_os[6]-3*sym_os[5]-3*sym_os[4]-12*sym_os[3]+4*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+4*sym_os[6]+2*sym_os[5]-4*sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+6*sym_os[6]-3*sym_os[5]-3*sym_os[4]-12*sym_os[3]+6*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-8*sym_os[8])+12*sym_os[7]+12*sym_os[6]-4*sym_os[5]-4*sym_os[4]-18*sym_os[3]+6*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u-4*sym_os[8]+4*sym_os[7]+6*sym_os[6]-2*sym_os[4]-6*sym_os[3]+2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(3*sym_os[5]-3*sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+2*sym_os[6]+2*sym_os[5]-4*sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+((-4*sym_os[8])+6*sym_os[7]+4*sym_os[6]-2*sym_os[4]-6*sym_os[3]+2*sym_os[2])*u-2*sym_os[8]+2*sym_os[7]+2*sym_os[6]-2*sym_os[3])/8;
		A(5,8)=-((sym_os[5]-sym_os[4])*v*v*v*v+(3*sym_os[5]-3*sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-8*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[3]+4*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+4*sym_os[6]+2*sym_os[5]-4*sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+6*sym_os[6]-3*sym_os[5]-3*sym_os[4]-12*sym_os[3]+6*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(8*sym_os[8]-4*sym_os[7]-12*sym_os[6]+6*sym_os[3]+6*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u-4*sym_os[8]-4*sym_os[7]+6*sym_os[6]-2*sym_os[4]+6*sym_os[3]-2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+2*sym_os[6]-2*sym_os[5]-4*sym_os[3]+2*sym_os[2])*u*u+(4*sym_os[8]-2*sym_os[7]-4*sym_os[6]-2*sym_os[4]+2*sym_os[3]+2*sym_os[2])*u-2*sym_os[8]-2*sym_os[7]+2*sym_os[6]+2*sym_os[3])/8;
		A(6,0)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[5]-sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-2*sym_os[7])+sym_os[5]+sym_os[4]+4*sym_os[3]-4*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u)*v*v+((2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(2*sym_os[3]-2*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u)*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u*u)/8;
		A(6,1)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[5]-sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-8*sym_os[6]+3*sym_os[5]+3*sym_os[4]+12*sym_os[3]-4*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+4*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+((2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[6])+2*sym_os[5]+2*sym_os[4]+6*sym_os[3]-2*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u+2*sym_os[6]-sym_os[5]-sym_os[4]-2*sym_os[3]+2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(3*sym_os[4]-3*sym_os[5])*u*u*u+(3*sym_os[5]-3*sym_os[4])*u*u+(sym_os[4]-sym_os[5])*u)/8;
		A(6,2)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[4]-sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-2*sym_os[7])+sym_os[5]+sym_os[4]+4*sym_os[3]-4*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u-sym_os[5]+sym_os[4])*v*v+((4*sym_os[8]-8*sym_os[7]-2*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[7]-2*sym_os[5]-2*sym_os[4]-2*sym_os[3]+2*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u+sym_os[5]-sym_os[4])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-2*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+((-2*sym_os[7])+sym_os[5]+sym_os[4]-2*sym_os[3]+2*sym_os[2])*u)/8;
		A(6,3)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[5]-sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+8*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]-4*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+4*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+((2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-2*sym_os[3]-2*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u+2*sym_os[6]-sym_os[5]-sym_os[4]+2*sym_os[3]-2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[5]-sym_os[4])*u*u*u+(sym_os[4]-sym_os[5])*u*u+(sym_os[4]-sym_os[5])*u)/8;
		A(6,4)=-((sym_os[5]-sym_os[4])*v*v*v*v+(3*sym_os[5]-3*sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-2*sym_os[7])+sym_os[5]+sym_os[4]+4*sym_os[3]-4*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u+3*sym_os[5]-3*sym_os[4])*v*v+(((-4*sym_os[8])+8*sym_os[7]+6*sym_os[6]-3*sym_os[5]-3*sym_os[4]-12*sym_os[3]+6*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[7])+2*sym_os[5]+2*sym_os[4]+6*sym_os[3]-6*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u+sym_os[5]-sym_os[4])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+((-2*sym_os[7])+sym_os[5]+sym_os[4]+2*sym_os[3]-2*sym_os[2])*u)/8;
		A(6,5)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[4]-sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-8*sym_os[6]+3*sym_os[5]+3*sym_os[4]+12*sym_os[3]-4*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+4*sym_os[6]-2*sym_os[5]-4*sym_os[3]+2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-2*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-8*sym_os[8])+12*sym_os[7]+4*sym_os[6]-6*sym_os[3]+2*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u+4*sym_os[8]-4*sym_os[7]-2*sym_os[6]-2*sym_os[4]+2*sym_os[3]+2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(3*sym_os[4]-3*sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-2*sym_os[6]+2*sym_os[5]-4*sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+(4*sym_os[8]-6*sym_os[7]+4*sym_os[6]+2*sym_os[4]-6*sym_os[3]+2*sym_os[2])*u-2*sym_os[8]+2*sym_os[7]-2*sym_os[6]+2*sym_os[3])/8;
		A(6,6)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[4]-sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+8*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]-4*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+4*sym_os[6]-2*sym_os[5]+4*sym_os[3]-2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-2*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(8*sym_os[8]-4*sym_os[7]-4*sym_os[6]-4*sym_os[5]-4*sym_os[4]+2*sym_os[3]+2*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u+4*sym_os[8]+4*sym_os[7]-2*sym_os[6]-2*sym_os[4]-2*sym_os[3]-2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[5]-sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-2*sym_os[6]-2*sym_os[5]+4*sym_os[3]-2*sym_os[2])*u*u+((-4*sym_os[8])+2*sym_os[7]-4*sym_os[6]+2*sym_os[4]+2*sym_os[3]+2*sym_os[2])*u-2*sym_os[8]-2*sym_os[7]-2*sym_os[6]-2*sym_os[3])/8;
		A(6,7)=-((sym_os[5]-sym_os[4])*v*v*v*v+(3*sym_os[5]-3*sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+8*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]-4*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+4*sym_os[6]+2*sym_os[5]-4*sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+6*sym_os[6]-3*sym_os[5]-3*sym_os[4]-12*sym_os[3]+6*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-8*sym_os[8])+4*sym_os[7]+12*sym_os[6]-6*sym_os[3]-6*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u-4*sym_os[8]-4*sym_os[7]+6*sym_os[6]-2*sym_os[4]+6*sym_os[3]-2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[5]-sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+2*sym_os[6]-2*sym_os[5]-4*sym_os[3]+2*sym_os[2])*u*u+((-4*sym_os[8])+2*sym_os[7]+4*sym_os[6]+2*sym_os[4]-2*sym_os[3]-2*sym_os[2])*u-2*sym_os[8]-2*sym_os[7]+2*sym_os[6]+2*sym_os[3])/8;
		A(6,8)=-((sym_os[5]-sym_os[4])*v*v*v*v+(3*sym_os[5]-3*sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-8*sym_os[6]+3*sym_os[5]+3*sym_os[4]+12*sym_os[3]-4*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+4*sym_os[6]+2*sym_os[5]-4*sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+6*sym_os[6]-3*sym_os[5]-3*sym_os[4]-12*sym_os[3]+6*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(8*sym_os[8]-12*sym_os[7]-12*sym_os[6]+4*sym_os[5]+4*sym_os[4]+18*sym_os[3]-6*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u-4*sym_os[8]+4*sym_os[7]+6*sym_os[6]-2*sym_os[4]-6*sym_os[3]+2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(3*sym_os[4]-3*sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+2*sym_os[6]+2*sym_os[5]-4*sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+(4*sym_os[8]-6*sym_os[7]-4*sym_os[6]+2*sym_os[4]+6*sym_os[3]-2*sym_os[2])*u-2*sym_os[8]+2*sym_os[7]+2*sym_os[6]-2*sym_os[3])/8;
		A(7,0)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[4]-sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-2*sym_os[7])+sym_os[5]+sym_os[4]+4*sym_os[3]-4*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u)*v*v+(((-2*sym_os[6])+sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-2*sym_os[3])+2*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u)*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u*u)/8;
		A(7,1)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[4]-sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-8*sym_os[6]+3*sym_os[5]+3*sym_os[4]+12*sym_os[3]-4*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+4*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+(((-2*sym_os[6])+sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-6*sym_os[3]+2*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u-2*sym_os[6]+sym_os[5]+sym_os[4]+2*sym_os[3]-2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(3*sym_os[4]-3*sym_os[5])*u*u*u+(3*sym_os[5]-3*sym_os[4])*u*u+(sym_os[4]-sym_os[5])*u)/8;
		A(7,2)=-((sym_os[5]-sym_os[4])*v*v*v*v+(3*sym_os[4]-3*sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-2*sym_os[7])+sym_os[5]+sym_os[4]+4*sym_os[3]-4*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u+3*sym_os[5]-3*sym_os[4])*v*v+((4*sym_os[8]-8*sym_os[7]-6*sym_os[6]+3*sym_os[5]+3*sym_os[4]+12*sym_os[3]-6*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[7]-2*sym_os[5]-2*sym_os[4]-6*sym_os[3]+6*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u-sym_os[5]+sym_os[4])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+((-2*sym_os[7])+sym_os[5]+sym_os[4]+2*sym_os[3]-2*sym_os[2])*u)/8;
		A(7,3)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[4]-sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+8*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]-4*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+4*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+(((-2*sym_os[6])+sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[6])+2*sym_os[5]+2*sym_os[4]+2*sym_os[3]+2*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u-2*sym_os[6]+sym_os[5]+sym_os[4]-2*sym_os[3]+2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[5]-sym_os[4])*u*u*u+(sym_os[4]-sym_os[5])*u*u+(sym_os[4]-sym_os[5])*u)/8;
		A(7,4)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[5]-sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-2*sym_os[7])+sym_os[5]+sym_os[4]+4*sym_os[3]-4*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u-sym_os[5]+sym_os[4])*v*v+(((-4*sym_os[8])+8*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[7])+2*sym_os[5]+2*sym_os[4]+2*sym_os[3]-2*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u-sym_os[5]+sym_os[4])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-2*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+((-2*sym_os[7])+sym_os[5]+sym_os[4]-2*sym_os[3]+2*sym_os[2])*u)/8;
		A(7,5)=-((sym_os[5]-sym_os[4])*v*v*v*v+(3*sym_os[4]-3*sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-8*sym_os[6]+3*sym_os[5]+3*sym_os[4]+12*sym_os[3]-4*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+4*sym_os[6]+2*sym_os[5]-4*sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-6*sym_os[6]+3*sym_os[5]+3*sym_os[4]+12*sym_os[3]-6*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-8*sym_os[8])+12*sym_os[7]+12*sym_os[6]-4*sym_os[5]-4*sym_os[4]-18*sym_os[3]+6*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u+4*sym_os[8]-4*sym_os[7]-6*sym_os[6]+2*sym_os[4]+6*sym_os[3]-2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(3*sym_os[4]-3*sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+2*sym_os[6]+2*sym_os[5]-4*sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+(4*sym_os[8]-6*sym_os[7]-4*sym_os[6]+2*sym_os[4]+6*sym_os[3]-2*sym_os[2])*u-2*sym_os[8]+2*sym_os[7]+2*sym_os[6]-2*sym_os[3])/8;
		A(7,6)=-((sym_os[5]-sym_os[4])*v*v*v*v+(3*sym_os[4]-3*sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+8*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]-4*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+4*sym_os[6]+2*sym_os[5]-4*sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-6*sym_os[6]+3*sym_os[5]+3*sym_os[4]+12*sym_os[3]-6*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(8*sym_os[8]-4*sym_os[7]-12*sym_os[6]+6*sym_os[3]+6*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u+4*sym_os[8]+4*sym_os[7]-6*sym_os[6]+2*sym_os[4]-6*sym_os[3]+2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[5]-sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+2*sym_os[6]-2*sym_os[5]-4*sym_os[3]+2*sym_os[2])*u*u+((-4*sym_os[8])+2*sym_os[7]+4*sym_os[6]+2*sym_os[4]-2*sym_os[3]-2*sym_os[2])*u-2*sym_os[8]-2*sym_os[7]+2*sym_os[6]+2*sym_os[3])/8;
		A(7,7)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[5]-sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+8*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]-4*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+4*sym_os[6]-2*sym_os[5]+4*sym_os[3]-2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-8*sym_os[8])+4*sym_os[7]+4*sym_os[6]+4*sym_os[5]+4*sym_os[4]-2*sym_os[3]-2*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u-4*sym_os[8]-4*sym_os[7]+2*sym_os[6]+2*sym_os[4]+2*sym_os[3]+2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[5]-sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-2*sym_os[6]-2*sym_os[5]+4*sym_os[3]-2*sym_os[2])*u*u+((-4*sym_os[8])+2*sym_os[7]-4*sym_os[6]+2*sym_os[4]+2*sym_os[3]+2*sym_os[2])*u-2*sym_os[8]-2*sym_os[7]-2*sym_os[6]-2*sym_os[3])/8;
		A(7,8)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[5]-sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-8*sym_os[6]+3*sym_os[5]+3*sym_os[4]+12*sym_os[3]-4*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+4*sym_os[6]-2*sym_os[5]-4*sym_os[3]+2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(8*sym_os[8]-12*sym_os[7]-4*sym_os[6]+6*sym_os[3]-2*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u-4*sym_os[8]+4*sym_os[7]+2*sym_os[6]+2*sym_os[4]-2*sym_os[3]-2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(3*sym_os[4]-3*sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-2*sym_os[6]+2*sym_os[5]-4*sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+(4*sym_os[8]-6*sym_os[7]+4*sym_os[6]+2*sym_os[4]-6*sym_os[3]+2*sym_os[2])*u-2*sym_os[8]+2*sym_os[7]-2*sym_os[6]+2*sym_os[3])/8;
		A(8,0)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[4]-sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(2*sym_os[7]-sym_os[5]-sym_os[4]-4*sym_os[3]+4*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u)*v*v+(((-2*sym_os[6])+sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(2*sym_os[3]-2*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u)*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[5]-sym_os[4])*u*u*u)/8;
		A(8,1)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[4]-sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-8*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[3]+4*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+4*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+(((-2*sym_os[6])+sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-2*sym_os[3]-2*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u-2*sym_os[6]+sym_os[5]+sym_os[4]-2*sym_os[3]+2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u*u+(sym_os[4]-sym_os[5])*u*u+(sym_os[5]-sym_os[4])*u)/8;
		A(8,2)=-((sym_os[5]-sym_os[4])*v*v*v*v+(3*sym_os[4]-3*sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(2*sym_os[7]-sym_os[5]-sym_os[4]-4*sym_os[3]+4*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u+3*sym_os[5]-3*sym_os[4])*v*v+((4*sym_os[8]-8*sym_os[7]-6*sym_os[6]+3*sym_os[5]+3*sym_os[4]+12*sym_os[3]-6*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[7])+2*sym_os[5]+2*sym_os[4]+6*sym_os[3]-6*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u-sym_os[5]+sym_os[4])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[5]-sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+(2*sym_os[7]-sym_os[5]-sym_os[4]-2*sym_os[3]+2*sym_os[2])*u)/8;
		A(8,3)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[4]-sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+8*sym_os[6]-3*sym_os[5]-3*sym_os[4]-12*sym_os[3]+4*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+4*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+(((-2*sym_os[6])+sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[6])+2*sym_os[5]+2*sym_os[4]+6*sym_os[3]-2*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u-2*sym_os[6]+sym_os[5]+sym_os[4]+2*sym_os[3]-2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(3*sym_os[5]-3*sym_os[4])*u*u*u+(3*sym_os[5]-3*sym_os[4])*u*u+(sym_os[5]-sym_os[4])*u)/8;
		A(8,4)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[5]-sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(2*sym_os[7]-sym_os[5]-sym_os[4]-4*sym_os[3]+4*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u-sym_os[5]+sym_os[4])*v*v+(((-4*sym_os[8])+8*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[7]-2*sym_os[5]-2*sym_os[4]-2*sym_os[3]+2*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u-sym_os[5]+sym_os[4])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[5]-sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-2*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+(2*sym_os[7]-sym_os[5]-sym_os[4]+2*sym_os[3]-2*sym_os[2])*u)/8;
		A(8,5)=-((sym_os[5]-sym_os[4])*v*v*v*v+(3*sym_os[4]-3*sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-8*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[3]+4*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+4*sym_os[6]+2*sym_os[5]-4*sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-6*sym_os[6]+3*sym_os[5]+3*sym_os[4]+12*sym_os[3]-6*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-8*sym_os[8])+4*sym_os[7]+12*sym_os[6]-6*sym_os[3]-6*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u+4*sym_os[8]+4*sym_os[7]-6*sym_os[6]+2*sym_os[4]-6*sym_os[3]+2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+2*sym_os[6]-2*sym_os[5]-4*sym_os[3]+2*sym_os[2])*u*u+(4*sym_os[8]-2*sym_os[7]-4*sym_os[6]-2*sym_os[4]+2*sym_os[3]+2*sym_os[2])*u-2*sym_os[8]-2*sym_os[7]+2*sym_os[6]+2*sym_os[3])/8;
		A(8,6)=-((sym_os[5]-sym_os[4])*v*v*v*v+(3*sym_os[4]-3*sym_os[5])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+8*sym_os[6]-3*sym_os[5]-3*sym_os[4]-12*sym_os[3]+4*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+4*sym_os[6]+2*sym_os[5]-4*sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-6*sym_os[6]+3*sym_os[5]+3*sym_os[4]+12*sym_os[3]-6*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(8*sym_os[8]-12*sym_os[7]-12*sym_os[6]+4*sym_os[5]+4*sym_os[4]+18*sym_os[3]-6*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u+4*sym_os[8]-4*sym_os[7]-6*sym_os[6]+2*sym_os[4]+6*sym_os[3]-2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(3*sym_os[5]-3*sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]+2*sym_os[6]+2*sym_os[5]-4*sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+((-4*sym_os[8])+6*sym_os[7]+4*sym_os[6]-2*sym_os[4]-6*sym_os[3]+2*sym_os[2])*u-2*sym_os[8]+2*sym_os[7]+2*sym_os[6]-2*sym_os[3])/8;
		A(8,7)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[5]-sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+8*sym_os[6]-3*sym_os[5]-3*sym_os[4]-12*sym_os[3]+4*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]+2*sym_os[7]+4*sym_os[6]-2*sym_os[5]-4*sym_os[3]+2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-8*sym_os[8])+12*sym_os[7]+4*sym_os[6]-6*sym_os[3]+2*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u-4*sym_os[8]+4*sym_os[7]+2*sym_os[6]+2*sym_os[4]-2*sym_os[3]-2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(3*sym_os[5]-3*sym_os[4])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-2*sym_os[6]+2*sym_os[5]-4*sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+((-4*sym_os[8])+6*sym_os[7]-4*sym_os[6]-2*sym_os[4]+6*sym_os[3]-2*sym_os[2])*u-2*sym_os[8]+2*sym_os[7]-2*sym_os[6]+2*sym_os[3])/8;
		A(8,8)=-((sym_os[5]-sym_os[4])*v*v*v*v+(sym_os[5]-sym_os[4])*v*v*v+(((-2*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]-2*sym_os[4]-8*sym_os[3]+4*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-8*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[3]+4*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u-2*sym_os[8]-2*sym_os[7]+4*sym_os[6]-2*sym_os[5]+4*sym_os[3]-2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(8*sym_os[8]-4*sym_os[7]-4*sym_os[6]-4*sym_os[5]-4*sym_os[4]+2*sym_os[3]+2*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u-4*sym_os[8]-4*sym_os[7]+2*sym_os[6]+2*sym_os[4]+2*sym_os[3]+2*sym_os[1])*v+(sym_os[5]-sym_os[4])*u*u*u*u+(sym_os[4]-sym_os[5])*u*u*u+((-2*sym_os[8])+4*sym_os[7]-2*sym_os[6]-2*sym_os[5]+4*sym_os[3]-2*sym_os[2])*u*u+(4*sym_os[8]-2*sym_os[7]+4*sym_os[6]-2*sym_os[4]-2*sym_os[3]-2*sym_os[2])*u-2*sym_os[8]-2*sym_os[7]-2*sym_os[6]-2*sym_os[3])/8;
		break;
	default:
		printf("Error: matrix A not implemented for Q%d and CLBM1\n",Q);
	}
}

void LBM::setA_CLBM2(matrix &A, int Q, ex u, ex v, ex w)
{
	switch (Q)
	{
	case 5:
		A(0,0) = (-sym_os[3]+2*sym_os[2]-sym_os[0])*v*v+(-sym_os[3]+2*sym_os[1]-sym_os[0])*u*u+sym_os[0];
		A(0,1) = (-sym_os[3]+2*sym_os[2]-sym_os[0])*v*v+(-sym_os[3]+2*sym_os[1]-sym_os[0])*u*u+(2*sym_os[3]-2*sym_os[1])*u-sym_os[3]+sym_os[0];
		A(0,2) = (-sym_os[3]+2*sym_os[2]-sym_os[0])*v*v+(2*sym_os[3]-2*sym_os[2])*v+(-sym_os[3]+2*sym_os[1]-sym_os[0])*u*u-sym_os[3]+sym_os[0];
		A(0,3) = (-sym_os[3]+2*sym_os[2]-sym_os[0])*v*v+(-sym_os[3]+2*sym_os[1]-sym_os[0])*u*u+(2*sym_os[1]-2*sym_os[3])*u-sym_os[3]+sym_os[0];
		A(0,4) = (-sym_os[3]+2*sym_os[2]-sym_os[0])*v*v+(2*sym_os[2]-2*sym_os[3])*v+(-sym_os[3]+2*sym_os[1]-sym_os[0])*u*u-sym_os[3]+sym_os[0];
		A(1,0) = -n1o4*(sym_os[4]-sym_os[3])*v*v-n1o4*(-sym_os[4]-sym_os[3]+4*sym_os[1]-2*sym_os[0])*u*u-n1o4*(2*sym_os[1]-2*sym_os[0])*u;
		A(1,1) = -n1o4*(sym_os[4]-sym_os[3])*v*v-n1o4*(-sym_os[4]-sym_os[3]+4*sym_os[1]-2*sym_os[0])*u*u-n1o4*(2*sym_os[4]+2*sym_os[3]-2*sym_os[1]-2*sym_os[0])*u+n1o4*sym_os[4]+n1o4*sym_os[3]+n1o2*sym_os[1];
		A(1,2) = -n1o4*(sym_os[4]-sym_os[3])*v*v-n1o4*(2*sym_os[3]-2*sym_os[4])*v-n1o4*(-sym_os[4]-sym_os[3]+4*sym_os[1]-2*sym_os[0])*u*u-n1o4*(2*sym_os[1]-2*sym_os[0])*u-n1o4*sym_os[4]+n1o4*sym_os[3];
		A(1,3) = -n1o4*(sym_os[4]-sym_os[3])*v*v-n1o4*(-sym_os[4]-sym_os[3]+4*sym_os[1]-2*sym_os[0])*u*u-n1o4*(-2*sym_os[4]-2*sym_os[3]+6*sym_os[1]-2*sym_os[0])*u+n1o4*sym_os[4]+n1o4*sym_os[3]-n1o2*sym_os[1];
		A(1,4) = -n1o4*(sym_os[4]-sym_os[3])*v*v-n1o4*(2*sym_os[4]-2*sym_os[3])*v-n1o4*(-sym_os[4]-sym_os[3]+4*sym_os[1]-2*sym_os[0])*u*u-n1o4*(2*sym_os[1]-2*sym_os[0])*u-n1o4*sym_os[4]+n1o4*sym_os[3];
		A(2,0) = n1o4*(sym_os[4]+sym_os[3]-4*sym_os[2]+2*sym_os[0])*v*v+n1o4*(2*sym_os[0]-2*sym_os[2])*v+n1o4*(sym_os[3]-sym_os[4])*u*u;
		A(2,1) = n1o4*(sym_os[4]+sym_os[3]-4*sym_os[2]+2*sym_os[0])*v*v+n1o4*(2*sym_os[0]-2*sym_os[2])*v+n1o4*(sym_os[3]-sym_os[4])*u*u+n1o4*(2*sym_os[4]-2*sym_os[3])*u-n1o4*sym_os[4]+n1o4*sym_os[3];
		A(2,2) = n1o4*(sym_os[4]+sym_os[3]-4*sym_os[2]+2*sym_os[0])*v*v+n1o4*(-2*sym_os[4]-2*sym_os[3]+2*sym_os[2]+2*sym_os[0])*v+n1o4*(sym_os[3]-sym_os[4])*u*u+n1o4*sym_os[4]+n1o4*sym_os[3]+n1o2*sym_os[2];
		A(2,3) = n1o4*(sym_os[4]+sym_os[3]-4*sym_os[2]+2*sym_os[0])*v*v+n1o4*(2*sym_os[0]-2*sym_os[2])*v+n1o4*(sym_os[3]-sym_os[4])*u*u+n1o4*(2*sym_os[3]-2*sym_os[4])*u-n1o4*sym_os[4]+n1o4*sym_os[3];
		A(2,4) = n1o4*(sym_os[4]+sym_os[3]-4*sym_os[2]+2*sym_os[0])*v*v+n1o4*(2*sym_os[4]+2*sym_os[3]-6*sym_os[2]+2*sym_os[0])*v+n1o4*(sym_os[3]-sym_os[4])*u*u+n1o4*sym_os[4]+n1o4*sym_os[3]-n1o2*sym_os[2];
		A(3,0) = -n1o4*(sym_os[4]-sym_os[3])*v*v-n1o4*(-sym_os[4]-sym_os[3]+4*sym_os[1]-2*sym_os[0])*u*u-n1o4*(2*sym_os[0]-2*sym_os[1])*u;
		A(3,1) = -n1o4*(sym_os[4]-sym_os[3])*v*v-n1o4*(-sym_os[4]-sym_os[3]+4*sym_os[1]-2*sym_os[0])*u*u-n1o4*(2*sym_os[4]+2*sym_os[3]-6*sym_os[1]+2*sym_os[0])*u+n1o4*sym_os[4]+n1o4*sym_os[3]-n1o2*sym_os[1];
		A(3,2) = -n1o4*(sym_os[4]-sym_os[3])*v*v-n1o4*(2*sym_os[3]-2*sym_os[4])*v-n1o4*(-sym_os[4]-sym_os[3]+4*sym_os[1]-2*sym_os[0])*u*u-n1o4*(2*sym_os[0]-2*sym_os[1])*u-n1o4*sym_os[4]+n1o4*sym_os[3];
		A(3,3) = -n1o4*(sym_os[4]-sym_os[3])*v*v-n1o4*(-sym_os[4]-sym_os[3]+4*sym_os[1]-2*sym_os[0])*u*u-n1o4*(-2*sym_os[4]-2*sym_os[3]+2*sym_os[1]+2*sym_os[0])*u+n1o4*sym_os[4]+n1o4*sym_os[3]+n1o2*sym_os[1];
		A(3,4) = -n1o4*(sym_os[4]-sym_os[3])*v*v-n1o4*(2*sym_os[4]-2*sym_os[3])*v-n1o4*(-sym_os[4]-sym_os[3]+4*sym_os[1]-2*sym_os[0])*u*u-n1o4*(2*sym_os[0]-2*sym_os[1])*u-n1o4*sym_os[4]+n1o4*sym_os[3];
		A(4,0) = n1o4*(sym_os[4]+sym_os[3]-4*sym_os[2]+2*sym_os[0])*v*v+n1o4*(2*sym_os[2]-2*sym_os[0])*v+n1o4*(sym_os[3]-sym_os[4])*u*u;
		A(4,1) = n1o4*(sym_os[4]+sym_os[3]-4*sym_os[2]+2*sym_os[0])*v*v+n1o4*(2*sym_os[2]-2*sym_os[0])*v+n1o4*(sym_os[3]-sym_os[4])*u*u+n1o4*(2*sym_os[4]-2*sym_os[3])*u-n1o4*sym_os[4]+n1o4*sym_os[3];
		A(4,2) = n1o4*(sym_os[4]+sym_os[3]-4*sym_os[2]+2*sym_os[0])*v*v+n1o4*(-2*sym_os[4]-2*sym_os[3]+6*sym_os[2]-2*sym_os[0])*v+n1o4*(sym_os[3]-sym_os[4])*u*u+n1o4*sym_os[4]+n1o4*sym_os[3]-n1o2*sym_os[2];
		A(4,3) = n1o4*(sym_os[4]+sym_os[3]-4*sym_os[2]+2*sym_os[0])*v*v+n1o4*(2*sym_os[2]-2*sym_os[0])*v+n1o4*(sym_os[3]-sym_os[4])*u*u+n1o4*(2*sym_os[3]-2*sym_os[4])*u-n1o4*sym_os[4]+n1o4*sym_os[3];
		A(4,4) = n1o4*(sym_os[4]+sym_os[3]-4*sym_os[2]+2*sym_os[0])*v*v+n1o4*(2*sym_os[4]+2*sym_os[3]-2*sym_os[2]-2*sym_os[0])*v+n1o4*(sym_os[3]-sym_os[4])*u*u+n1o4*sym_os[4]+n1o4*sym_os[3]+n1o2*sym_os[2];
		break;
	case 7:
		A(0,0) = (-sym_os[4]+2*sym_os[3]-sym_os[0])*w*w+(-sym_os[4]+2*sym_os[2]-sym_os[0])*v*v+(-sym_os[4]+2*sym_os[1]-sym_os[0])*u*u+sym_os[0];
		A(0,1) = (-sym_os[4]+2*sym_os[3]-sym_os[0])*w*w+(-sym_os[4]+2*sym_os[2]-sym_os[0])*v*v+(-sym_os[4]+2*sym_os[1]-sym_os[0])*u*u+(2*sym_os[4]-2*sym_os[1])*u-sym_os[4]+sym_os[0];
		A(0,2) = (-sym_os[4]+2*sym_os[3]-sym_os[0])*w*w+(-sym_os[4]+2*sym_os[2]-sym_os[0])*v*v+(2*sym_os[4]-2*sym_os[2])*v+(-sym_os[4]+2*sym_os[1]-sym_os[0])*u*u-sym_os[4]+sym_os[0];
		A(0,3) = (-sym_os[4]+2*sym_os[3]-sym_os[0])*w*w+(2*sym_os[4]-2*sym_os[3])*w+(-sym_os[4]+2*sym_os[2]-sym_os[0])*v*v+(-sym_os[4]+2*sym_os[1]-sym_os[0])*u*u-sym_os[4]+sym_os[0];
		A(0,4) = (-sym_os[4]+2*sym_os[3]-sym_os[0])*w*w+(-sym_os[4]+2*sym_os[2]-sym_os[0])*v*v+(-sym_os[4]+2*sym_os[1]-sym_os[0])*u*u+(2*sym_os[1]-2*sym_os[4])*u-sym_os[4]+sym_os[0];
		A(0,5) = (-sym_os[4]+2*sym_os[3]-sym_os[0])*w*w+(-sym_os[4]+2*sym_os[2]-sym_os[0])*v*v+(2*sym_os[2]-2*sym_os[4])*v+(-sym_os[4]+2*sym_os[1]-sym_os[0])*u*u-sym_os[4]+sym_os[0];
		A(0,6) = (-sym_os[4]+2*sym_os[3]-sym_os[0])*w*w+(2*sym_os[3]-2*sym_os[4])*w+(-sym_os[4]+2*sym_os[2]-sym_os[0])*v*v+(-sym_os[4]+2*sym_os[1]-sym_os[0])*u*u-sym_os[4]+sym_os[0];

		A(1,0) = -((sym_os[6]-sym_os[4])*w*w+(sym_os[5]-sym_os[4])*v*v+(-sym_os[6]-sym_os[5]-sym_os[4]+6*sym_os[1]-3*sym_os[0])*u*u+(3*sym_os[1]-3*sym_os[0])*u)/6;
		A(1,1) = -((sym_os[6]-sym_os[4])*w*w+(sym_os[5]-sym_os[4])*v*v+(-sym_os[6]-sym_os[5]-sym_os[4]+6*sym_os[1]-3*sym_os[0])*u*u+(2*sym_os[6]+2*sym_os[5]+2*sym_os[4]-3*sym_os[1]-3*sym_os[0])*u-sym_os[6]-sym_os[5]-sym_os[4]-3*sym_os[1])/6;
		A(1,2) = -((sym_os[6]-sym_os[4])*w*w+(sym_os[5]-sym_os[4])*v*v+(2*sym_os[4]-2*sym_os[5])*v+(-sym_os[6]-sym_os[5]-sym_os[4]+6*sym_os[1]-3*sym_os[0])*u*u+(3*sym_os[1]-3*sym_os[0])*u+sym_os[5]-sym_os[4])/6;
		A(1,3) = -((sym_os[6]-sym_os[4])*w*w+(2*sym_os[4]-2*sym_os[6])*w+(sym_os[5]-sym_os[4])*v*v+(-sym_os[6]-sym_os[5]-sym_os[4]+6*sym_os[1]-3*sym_os[0])*u*u+(3*sym_os[1]-3*sym_os[0])*u+sym_os[6]-sym_os[4])/6;
		A(1,4) = -((sym_os[6]-sym_os[4])*w*w+(sym_os[5]-sym_os[4])*v*v+(-sym_os[6]-sym_os[5]-sym_os[4]+6*sym_os[1]-3*sym_os[0])*u*u+(-2*sym_os[6]-2*sym_os[5]-2*sym_os[4]+9*sym_os[1]-3*sym_os[0])*u-sym_os[6]-sym_os[5]-sym_os[4]+3*sym_os[1])/6;
		A(1,5) = -((sym_os[6]-sym_os[4])*w*w+(sym_os[5]-sym_os[4])*v*v+(2*sym_os[5]-2*sym_os[4])*v+(-sym_os[6]-sym_os[5]-sym_os[4]+6*sym_os[1]-3*sym_os[0])*u*u+(3*sym_os[1]-3*sym_os[0])*u+sym_os[5]-sym_os[4])/6;
		A(1,6) = -((sym_os[6]-sym_os[4])*w*w+(2*sym_os[6]-2*sym_os[4])*w+(sym_os[5]-sym_os[4])*v*v+(-sym_os[6]-sym_os[5]-sym_os[4]+6*sym_os[1]-3*sym_os[0])*u*u+(3*sym_os[1]-3*sym_os[0])*u+sym_os[6]-sym_os[4])/6;

		A(2,0) = -((sym_os[6]-sym_os[4])*w*w+(-2*sym_os[5]-sym_os[4]+6*sym_os[2]-3*sym_os[0])*v*v+(3*sym_os[2]-3*sym_os[0])*v+(-sym_os[6]+2*sym_os[5]-sym_os[4])*u*u)/6;
		A(2,1) = -((sym_os[6]-sym_os[4])*w*w+(-2*sym_os[5]-sym_os[4]+6*sym_os[2]-3*sym_os[0])*v*v+(3*sym_os[2]-3*sym_os[0])*v+(-sym_os[6]+2*sym_os[5]-sym_os[4])*u*u+(2*sym_os[6]-4*sym_os[5]+2*sym_os[4])*u-sym_os[6]+2*sym_os[5]-sym_os[4])/6;
		A(2,2) = -((sym_os[6]-sym_os[4])*w*w+(-2*sym_os[5]-sym_os[4]+6*sym_os[2]-3*sym_os[0])*v*v+(4*sym_os[5]+2*sym_os[4]-3*sym_os[2]-3*sym_os[0])*v+(-sym_os[6]+2*sym_os[5]-sym_os[4])*u*u-2*sym_os[5]-sym_os[4]-3*sym_os[2])/6;
		A(2,3) = -((sym_os[6]-sym_os[4])*w*w+(2*sym_os[4]-2*sym_os[6])*w+(-2*sym_os[5]-sym_os[4]+6*sym_os[2]-3*sym_os[0])*v*v+(3*sym_os[2]-3*sym_os[0])*v+(-sym_os[6]+2*sym_os[5]-sym_os[4])*u*u+sym_os[6]-sym_os[4])/6;
		A(2,4) = -((sym_os[6]-sym_os[4])*w*w+(-2*sym_os[5]-sym_os[4]+6*sym_os[2]-3*sym_os[0])*v*v+(3*sym_os[2]-3*sym_os[0])*v+(-sym_os[6]+2*sym_os[5]-sym_os[4])*u*u+(-2*sym_os[6]+4*sym_os[5]-2*sym_os[4])*u-sym_os[6]+2*sym_os[5]-sym_os[4])/6;
		A(2,5) = -((sym_os[6]-sym_os[4])*w*w+(-2*sym_os[5]-sym_os[4]+6*sym_os[2]-3*sym_os[0])*v*v+(-4*sym_os[5]-2*sym_os[4]+9*sym_os[2]-3*sym_os[0])*v+(-sym_os[6]+2*sym_os[5]-sym_os[4])*u*u-2*sym_os[5]-sym_os[4]+3*sym_os[2])/6;
		A(2,6) = -((sym_os[6]-sym_os[4])*w*w+(2*sym_os[6]-2*sym_os[4])*w+(-2*sym_os[5]-sym_os[4]+6*sym_os[2]-3*sym_os[0])*v*v+(3*sym_os[2]-3*sym_os[0])*v+(-sym_os[6]+2*sym_os[5]-sym_os[4])*u*u+sym_os[6]-sym_os[4])/6;

		A(3,0) = ((2*sym_os[6]+sym_os[4]-6*sym_os[3]+3*sym_os[0])*w*w+(3*sym_os[0]-3*sym_os[3])*w+(sym_os[4]-sym_os[5])*v*v+(-2*sym_os[6]+sym_os[5]+sym_os[4])*u*u)/6;
		A(3,1) = ((2*sym_os[6]+sym_os[4]-6*sym_os[3]+3*sym_os[0])*w*w+(3*sym_os[0]-3*sym_os[3])*w+(sym_os[4]-sym_os[5])*v*v+(-2*sym_os[6]+sym_os[5]+sym_os[4])*u*u+(4*sym_os[6]-2*sym_os[5]-2*sym_os[4])*u-2*sym_os[6]+sym_os[5]+sym_os[4])/6;
		A(3,2) = ((2*sym_os[6]+sym_os[4]-6*sym_os[3]+3*sym_os[0])*w*w+(3*sym_os[0]-3*sym_os[3])*w+(sym_os[4]-sym_os[5])*v*v+(2*sym_os[5]-2*sym_os[4])*v+(-2*sym_os[6]+sym_os[5]+sym_os[4])*u*u-sym_os[5]+sym_os[4])/6;
		A(3,3) = ((2*sym_os[6]+sym_os[4]-6*sym_os[3]+3*sym_os[0])*w*w+(-4*sym_os[6]-2*sym_os[4]+3*sym_os[3]+3*sym_os[0])*w+(sym_os[4]-sym_os[5])*v*v+(-2*sym_os[6]+sym_os[5]+sym_os[4])*u*u+2*sym_os[6]+sym_os[4]+3*sym_os[3])/6;
		A(3,4) = ((2*sym_os[6]+sym_os[4]-6*sym_os[3]+3*sym_os[0])*w*w+(3*sym_os[0]-3*sym_os[3])*w+(sym_os[4]-sym_os[5])*v*v+(-2*sym_os[6]+sym_os[5]+sym_os[4])*u*u+(-4*sym_os[6]+2*sym_os[5]+2*sym_os[4])*u-2*sym_os[6]+sym_os[5]+sym_os[4])/6;
		A(3,5) = ((2*sym_os[6]+sym_os[4]-6*sym_os[3]+3*sym_os[0])*w*w+(3*sym_os[0]-3*sym_os[3])*w+(sym_os[4]-sym_os[5])*v*v+(2*sym_os[4]-2*sym_os[5])*v+(-2*sym_os[6]+sym_os[5]+sym_os[4])*u*u-sym_os[5]+sym_os[4])/6;
		A(3,6) = ((2*sym_os[6]+sym_os[4]-6*sym_os[3]+3*sym_os[0])*w*w+(4*sym_os[6]+2*sym_os[4]-9*sym_os[3]+3*sym_os[0])*w+(sym_os[4]-sym_os[5])*v*v+(-2*sym_os[6]+sym_os[5]+sym_os[4])*u*u+2*sym_os[6]+sym_os[4]-3*sym_os[3])/6;

		A(4,0) = -((sym_os[6]-sym_os[4])*w*w+(sym_os[5]-sym_os[4])*v*v+(-sym_os[6]-sym_os[5]-sym_os[4]+6*sym_os[1]-3*sym_os[0])*u*u+(3*sym_os[0]-3*sym_os[1])*u)/6;
		A(4,1) = -((sym_os[6]-sym_os[4])*w*w+(sym_os[5]-sym_os[4])*v*v+(-sym_os[6]-sym_os[5]-sym_os[4]+6*sym_os[1]-3*sym_os[0])*u*u+(2*sym_os[6]+2*sym_os[5]+2*sym_os[4]-9*sym_os[1]+3*sym_os[0])*u-sym_os[6]-sym_os[5]-sym_os[4]+3*sym_os[1])/6;
		A(4,2) = -((sym_os[6]-sym_os[4])*w*w+(sym_os[5]-sym_os[4])*v*v+(2*sym_os[4]-2*sym_os[5])*v+(-sym_os[6]-sym_os[5]-sym_os[4]+6*sym_os[1]-3*sym_os[0])*u*u+(3*sym_os[0]-3*sym_os[1])*u+sym_os[5]-sym_os[4])/6;
		A(4,3) = -((sym_os[6]-sym_os[4])*w*w+(2*sym_os[4]-2*sym_os[6])*w+(sym_os[5]-sym_os[4])*v*v+(-sym_os[6]-sym_os[5]-sym_os[4]+6*sym_os[1]-3*sym_os[0])*u*u+(3*sym_os[0]-3*sym_os[1])*u+sym_os[6]-sym_os[4])/6;
		A(4,4) = -((sym_os[6]-sym_os[4])*w*w+(sym_os[5]-sym_os[4])*v*v+(-sym_os[6]-sym_os[5]-sym_os[4]+6*sym_os[1]-3*sym_os[0])*u*u+(-2*sym_os[6]-2*sym_os[5]-2*sym_os[4]+3*sym_os[1]+3*sym_os[0])*u-sym_os[6]-sym_os[5]-sym_os[4]-3*sym_os[1])/6;
		A(4,5) = -((sym_os[6]-sym_os[4])*w*w+(sym_os[5]-sym_os[4])*v*v+(2*sym_os[5]-2*sym_os[4])*v+(-sym_os[6]-sym_os[5]-sym_os[4]+6*sym_os[1]-3*sym_os[0])*u*u+(3*sym_os[0]-3*sym_os[1])*u+sym_os[5]-sym_os[4])/6;
		A(4,6) = -((sym_os[6]-sym_os[4])*w*w+(2*sym_os[6]-2*sym_os[4])*w+(sym_os[5]-sym_os[4])*v*v+(-sym_os[6]-sym_os[5]-sym_os[4]+6*sym_os[1]-3*sym_os[0])*u*u+(3*sym_os[0]-3*sym_os[1])*u+sym_os[6]-sym_os[4])/6;

		A(5,0) = -((sym_os[6]-sym_os[4])*w*w+(-2*sym_os[5]-sym_os[4]+6*sym_os[2]-3*sym_os[0])*v*v+(3*sym_os[0]-3*sym_os[2])*v+(-sym_os[6]+2*sym_os[5]-sym_os[4])*u*u)/6;
		A(5,1) = -((sym_os[6]-sym_os[4])*w*w+(-2*sym_os[5]-sym_os[4]+6*sym_os[2]-3*sym_os[0])*v*v+(3*sym_os[0]-3*sym_os[2])*v+(-sym_os[6]+2*sym_os[5]-sym_os[4])*u*u+(2*sym_os[6]-4*sym_os[5]+2*sym_os[4])*u-sym_os[6]+2*sym_os[5]-sym_os[4])/6;
		A(5,2) = -((sym_os[6]-sym_os[4])*w*w+(-2*sym_os[5]-sym_os[4]+6*sym_os[2]-3*sym_os[0])*v*v+(4*sym_os[5]+2*sym_os[4]-9*sym_os[2]+3*sym_os[0])*v+(-sym_os[6]+2*sym_os[5]-sym_os[4])*u*u-2*sym_os[5]-sym_os[4]+3*sym_os[2])/6;
		A(5,3) = -((sym_os[6]-sym_os[4])*w*w+(2*sym_os[4]-2*sym_os[6])*w+(-2*sym_os[5]-sym_os[4]+6*sym_os[2]-3*sym_os[0])*v*v+(3*sym_os[0]-3*sym_os[2])*v+(-sym_os[6]+2*sym_os[5]-sym_os[4])*u*u+sym_os[6]-sym_os[4])/6;
		A(5,4) = -((sym_os[6]-sym_os[4])*w*w+(-2*sym_os[5]-sym_os[4]+6*sym_os[2]-3*sym_os[0])*v*v+(3*sym_os[0]-3*sym_os[2])*v+(-sym_os[6]+2*sym_os[5]-sym_os[4])*u*u+(-2*sym_os[6]+4*sym_os[5]-2*sym_os[4])*u-sym_os[6]+2*sym_os[5]-sym_os[4])/6;
		A(5,5) = -((sym_os[6]-sym_os[4])*w*w+(-2*sym_os[5]-sym_os[4]+6*sym_os[2]-3*sym_os[0])*v*v+(-4*sym_os[5]-2*sym_os[4]+3*sym_os[2]+3*sym_os[0])*v+(-sym_os[6]+2*sym_os[5]-sym_os[4])*u*u-2*sym_os[5]-sym_os[4]-3*sym_os[2])/6;
		A(5,6) = -((sym_os[6]-sym_os[4])*w*w+(2*sym_os[6]-2*sym_os[4])*w+(-2*sym_os[5]-sym_os[4]+6*sym_os[2]-3*sym_os[0])*v*v+(3*sym_os[0]-3*sym_os[2])*v+(-sym_os[6]+2*sym_os[5]-sym_os[4])*u*u+sym_os[6]-sym_os[4])/6;

		A(6,0) = ((2*sym_os[6]+sym_os[4]-6*sym_os[3]+3*sym_os[0])*w*w+(3*sym_os[3]-3*sym_os[0])*w+(sym_os[4]-sym_os[5])*v*v+(-2*sym_os[6]+sym_os[5]+sym_os[4])*u*u)/6;
		A(6,1) = ((2*sym_os[6]+sym_os[4]-6*sym_os[3]+3*sym_os[0])*w*w+(3*sym_os[3]-3*sym_os[0])*w+(sym_os[4]-sym_os[5])*v*v+(-2*sym_os[6]+sym_os[5]+sym_os[4])*u*u+(4*sym_os[6]-2*sym_os[5]-2*sym_os[4])*u-2*sym_os[6]+sym_os[5]+sym_os[4])/6;
		A(6,2) = ((2*sym_os[6]+sym_os[4]-6*sym_os[3]+3*sym_os[0])*w*w+(3*sym_os[3]-3*sym_os[0])*w+(sym_os[4]-sym_os[5])*v*v+(2*sym_os[5]-2*sym_os[4])*v+(-2*sym_os[6]+sym_os[5]+sym_os[4])*u*u-sym_os[5]+sym_os[4])/6;
		A(6,3) = ((2*sym_os[6]+sym_os[4]-6*sym_os[3]+3*sym_os[0])*w*w+(-4*sym_os[6]-2*sym_os[4]+9*sym_os[3]-3*sym_os[0])*w+(sym_os[4]-sym_os[5])*v*v+(-2*sym_os[6]+sym_os[5]+sym_os[4])*u*u+2*sym_os[6]+sym_os[4]-3*sym_os[3])/6;
		A(6,4) = ((2*sym_os[6]+sym_os[4]-6*sym_os[3]+3*sym_os[0])*w*w+(3*sym_os[3]-3*sym_os[0])*w+(sym_os[4]-sym_os[5])*v*v+(-2*sym_os[6]+sym_os[5]+sym_os[4])*u*u+(-4*sym_os[6]+2*sym_os[5]+2*sym_os[4])*u-2*sym_os[6]+sym_os[5]+sym_os[4])/6;
		A(6,5) = ((2*sym_os[6]+sym_os[4]-6*sym_os[3]+3*sym_os[0])*w*w+(3*sym_os[3]-3*sym_os[0])*w+(sym_os[4]-sym_os[5])*v*v+(2*sym_os[4]-2*sym_os[5])*v+(-2*sym_os[6]+sym_os[5]+sym_os[4])*u*u-sym_os[5]+sym_os[4])/6;
		A(6,6) = ((2*sym_os[6]+sym_os[4]-6*sym_os[3]+3*sym_os[0])*w*w+(4*sym_os[6]+2*sym_os[4]-3*sym_os[3]-3*sym_os[0])*w+(sym_os[4]-sym_os[5])*v*v+(-2*sym_os[6]+sym_os[5]+sym_os[4])*u*u+2*sym_os[6]+sym_os[4]+3*sym_os[3])/6;
		break;
	case 9:
		A(0,0)=(sym_os[5]*v*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+4*sym_os[2]-2*sym_os[0])*v*v+sym_os[4]*u*u*u*u+((-2*sym_os[4])+4*sym_os[1]-2*sym_os[0])*u*u+2*sym_os[0])/2;
		A(0,1)=(sym_os[5]*v*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[4]-8*sym_os[3]+4*sym_os[1])*u+2*sym_os[8]-4*sym_os[6]+sym_os[4]+4*sym_os[2]-2*sym_os[0])*v*v+sym_os[4]*u*u*u*u-2*sym_os[4]*u*u*u+((-sym_os[4])+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[4]-4*sym_os[1])*u-2*sym_os[4]+2*sym_os[0])/2;
		A(0,2)=(sym_os[5]*v*v*v*v-2*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+sym_os[5]+4*sym_os[2]-2*sym_os[0])*v*v+(((-4*sym_os[8])+8*sym_os[7]+4*sym_os[6]+2*sym_os[5]-8*sym_os[3]+4*sym_os[2])*u*u-4*sym_os[2])*v+sym_os[4]*u*u*u*u+(2*sym_os[8]-4*sym_os[7]-sym_os[5]-2*sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+2*sym_os[0])/2;
		A(0,3)=(sym_os[5]*v*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[4]+8*sym_os[3]-4*sym_os[1])*u+2*sym_os[8]-4*sym_os[6]+sym_os[4]+4*sym_os[2]-2*sym_os[0])*v*v+sym_os[4]*u*u*u*u+2*sym_os[4]*u*u*u+((-sym_os[4])+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[1]-4*sym_os[4])*u-2*sym_os[4]+2*sym_os[0])/2;
		A(0,4)=(sym_os[5]*v*v*v*v+2*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+sym_os[5]+4*sym_os[2]-2*sym_os[0])*v*v+((4*sym_os[8]-8*sym_os[7]-4*sym_os[6]-2*sym_os[5]+8*sym_os[3]-4*sym_os[2])*u*u+4*sym_os[2])*v+sym_os[4]*u*u*u*u+(2*sym_os[8]-4*sym_os[7]-sym_os[5]-2*sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+2*sym_os[0])/2;
		A(0,5)=(sym_os[5]*v*v*v*v-2*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[4]-8*sym_os[3]+4*sym_os[1])*u+2*sym_os[8]-4*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[2]-2*sym_os[0])*v*v+(((-4*sym_os[8])+8*sym_os[7]+4*sym_os[6]+2*sym_os[5]-8*sym_os[3]+4*sym_os[2])*u*u+(8*sym_os[8]-8*sym_os[7]-8*sym_os[6]+8*sym_os[3])*u-4*sym_os[8]+4*sym_os[6]-4*sym_os[2])*v+sym_os[4]*u*u*u*u-2*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-sym_os[5]-sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+4*sym_os[4]-4*sym_os[1])*u+2*sym_os[8]-2*sym_os[4]+2*sym_os[0])/2;
		A(0,6)=(sym_os[5]*v*v*v*v-2*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[4]+8*sym_os[3]-4*sym_os[1])*u+2*sym_os[8]-4*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[2]-2*sym_os[0])*v*v+(((-4*sym_os[8])+8*sym_os[7]+4*sym_os[6]+2*sym_os[5]-8*sym_os[3]+4*sym_os[2])*u*u+((-8*sym_os[8])+8*sym_os[7]+8*sym_os[6]-8*sym_os[3])*u-4*sym_os[8]+4*sym_os[6]-4*sym_os[2])*v+sym_os[4]*u*u*u*u+2*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-sym_os[5]-sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-4*sym_os[4]+4*sym_os[1])*u+2*sym_os[8]-2*sym_os[4]+2*sym_os[0])/2;
		A(0,7)=(sym_os[5]*v*v*v*v+2*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[4]+8*sym_os[3]-4*sym_os[1])*u+2*sym_os[8]-4*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[2]-2*sym_os[0])*v*v+((4*sym_os[8]-8*sym_os[7]-4*sym_os[6]-2*sym_os[5]+8*sym_os[3]-4*sym_os[2])*u*u+(8*sym_os[8]-8*sym_os[7]-8*sym_os[6]+8*sym_os[3])*u+4*sym_os[8]-4*sym_os[6]+4*sym_os[2])*v+sym_os[4]*u*u*u*u+2*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-sym_os[5]-sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-4*sym_os[4]+4*sym_os[1])*u+2*sym_os[8]-2*sym_os[4]+2*sym_os[0])/2;
		A(0,8)=(sym_os[5]*v*v*v*v+2*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[4]-8*sym_os[3]+4*sym_os[1])*u+2*sym_os[8]-4*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[2]-2*sym_os[0])*v*v+((4*sym_os[8]-8*sym_os[7]-4*sym_os[6]-2*sym_os[5]+8*sym_os[3]-4*sym_os[2])*u*u+((-8*sym_os[8])+8*sym_os[7]+8*sym_os[6]-8*sym_os[3])*u+4*sym_os[8]-4*sym_os[6]+4*sym_os[2])*v+sym_os[4]*u*u*u*u-2*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-sym_os[5]-sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+4*sym_os[4]-4*sym_os[1])*u+2*sym_os[8]-2*sym_os[4]+2*sym_os[0])/2;
		A(1,0)=-(sym_os[5]*v*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-2*sym_os[7])-sym_os[5]+4*sym_os[3]-4*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u-sym_os[5])*v*v+sym_os[4]*u*u*u*u+sym_os[4]*u*u*u+((-sym_os[4])+4*sym_os[1]-2*sym_os[0])*u*u+(2*sym_os[1]-2*sym_os[0])*u)/4;
		A(1,1)=-(sym_os[5]*v*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+8*sym_os[6]-sym_os[5]-2*sym_os[4]-4*sym_os[3]-4*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u*u+((-2*sym_os[4])+4*sym_os[1]-2*sym_os[0])*u*u+(3*sym_os[4]-2*sym_os[1]-2*sym_os[0])*u-sym_os[4]-2*sym_os[1])/4;
		A(1,2)=-(sym_os[5]*v*v*v*v-2*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-2*sym_os[7])-sym_os[5]+4*sym_os[3]-4*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u)*v*v+(((-4*sym_os[8])+8*sym_os[7]+4*sym_os[6]+2*sym_os[5]-8*sym_os[3]+4*sym_os[2])*u*u+(4*sym_os[7]+2*sym_os[5]-4*sym_os[3]+4*sym_os[2])*u+2*sym_os[5])*v+sym_os[4]*u*u*u*u+sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-sym_os[5]-sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+((-2*sym_os[7])-sym_os[5]+2*sym_os[1]-2*sym_os[0])*u-sym_os[5])/4;
		A(1,3)=-(sym_os[5]*v*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-8*sym_os[6]-sym_os[5]+2*sym_os[4]+12*sym_os[3]-4*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+sym_os[4]*u*u*u*u+3*sym_os[4]*u*u*u+(2*sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+((-sym_os[4])+6*sym_os[1]-2*sym_os[0])*u-sym_os[4]+2*sym_os[1])/4;
		A(1,4)=-(sym_os[5]*v*v*v*v+2*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-2*sym_os[7])-sym_os[5]+4*sym_os[3]-4*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u)*v*v+((4*sym_os[8]-8*sym_os[7]-4*sym_os[6]-2*sym_os[5]+8*sym_os[3]-4*sym_os[2])*u*u+((-4*sym_os[7])-2*sym_os[5]+4*sym_os[3]-4*sym_os[2])*u-2*sym_os[5])*v+sym_os[4]*u*u*u*u+sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-sym_os[5]-sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+((-2*sym_os[7])-sym_os[5]+2*sym_os[1]-2*sym_os[0])*u-sym_os[5])/4;
		A(1,5)=-(sym_os[5]*v*v*v*v-2*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+8*sym_os[6]-sym_os[5]-2*sym_os[4]-4*sym_os[3]-4*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-4*sym_os[6]+sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+4*sym_os[6]+2*sym_os[5]-8*sym_os[3]+4*sym_os[2])*u*u+(8*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[5]+4*sym_os[3]+4*sym_os[2])*u-4*sym_os[8]-4*sym_os[7]+4*sym_os[6]+2*sym_os[5]+4*sym_os[3])*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-sym_os[5]-2*sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]-sym_os[5]+3*sym_os[4]-2*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-sym_os[5]-sym_os[4]-2*sym_os[1])/4;
		A(1,6)=-(sym_os[5]*v*v*v*v-2*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-8*sym_os[6]-sym_os[5]+2*sym_os[4]+12*sym_os[3]-4*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-4*sym_os[6]+sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+4*sym_os[6]+2*sym_os[5]-8*sym_os[3]+4*sym_os[2])*u*u+((-8*sym_os[8])+12*sym_os[7]+8*sym_os[6]+2*sym_os[5]-12*sym_os[3]+4*sym_os[2])*u-4*sym_os[8]+4*sym_os[7]+4*sym_os[6]+2*sym_os[5]-4*sym_os[3])*v+sym_os[4]*u*u*u*u+3*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-sym_os[5]+2*sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-sym_os[5]-sym_os[4]+6*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-sym_os[5]-sym_os[4]+2*sym_os[1])/4;
		A(1,7)=-(sym_os[5]*v*v*v*v+2*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-8*sym_os[6]-sym_os[5]+2*sym_os[4]+12*sym_os[3]-4*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-4*sym_os[6]+sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-4*sym_os[6]-2*sym_os[5]+8*sym_os[3]-4*sym_os[2])*u*u+(8*sym_os[8]-12*sym_os[7]-8*sym_os[6]-2*sym_os[5]+12*sym_os[3]-4*sym_os[2])*u+4*sym_os[8]-4*sym_os[7]-4*sym_os[6]-2*sym_os[5]+4*sym_os[3])*v+sym_os[4]*u*u*u*u+3*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-sym_os[5]+2*sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-sym_os[5]-sym_os[4]+6*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-sym_os[5]-sym_os[4]+2*sym_os[1])/4;
		A(1,8)=-(sym_os[5]*v*v*v*v+2*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+8*sym_os[6]-sym_os[5]-2*sym_os[4]-4*sym_os[3]-4*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-4*sym_os[6]+sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-4*sym_os[6]-2*sym_os[5]+8*sym_os[3]-4*sym_os[2])*u*u+((-8*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[5]-4*sym_os[3]-4*sym_os[2])*u+4*sym_os[8]+4*sym_os[7]-4*sym_os[6]-2*sym_os[5]-4*sym_os[3])*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-sym_os[5]-2*sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]-sym_os[5]+3*sym_os[4]-2*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-sym_os[5]-sym_os[4]-2*sym_os[1])/4;
		A(2,0)=-(sym_os[5]*v*v*v*v+sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+sym_os[5]+4*sym_os[2]-2*sym_os[0])*v*v+(((-2*sym_os[6])+sym_os[4]+4*sym_os[3]-2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+2*sym_os[2]-2*sym_os[0])*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u)/4;
		A(2,1)=-(sym_os[5]*v*v*v*v+sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[4]-8*sym_os[3]+4*sym_os[1])*u+2*sym_os[8]-4*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[2]-2*sym_os[0])*v*v+(((-2*sym_os[6])+sym_os[4]+4*sym_os[3]-2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[6]-2*sym_os[4]-4*sym_os[3]+4*sym_os[1])*u-2*sym_os[6]+sym_os[4]+2*sym_os[2]-2*sym_os[0])*v+sym_os[4]*u*u*u*u-2*sym_os[4]*u*u*u+2*sym_os[4]*u-sym_os[4])/4;
		A(2,2)=-(sym_os[5]*v*v*v*v-sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+4*sym_os[2]-2*sym_os[0])*v*v+(((-4*sym_os[8])+8*sym_os[7]+2*sym_os[6]+2*sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u-sym_os[5]-2*sym_os[2]-2*sym_os[0])*v+sym_os[4]*u*u*u*u+(2*sym_os[8]-4*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+sym_os[5]-2*sym_os[2])/4;
		A(2,3)=-(sym_os[5]*v*v*v*v+sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[4]+8*sym_os[3]-4*sym_os[1])*u+2*sym_os[8]-4*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[2]-2*sym_os[0])*v*v+(((-2*sym_os[6])+sym_os[4]+4*sym_os[3]-2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[6])+2*sym_os[4]+4*sym_os[3]-4*sym_os[1])*u-2*sym_os[6]+sym_os[4]+2*sym_os[2]-2*sym_os[0])*v+sym_os[4]*u*u*u*u+2*sym_os[4]*u*u*u-2*sym_os[4]*u-sym_os[4])/4;
		A(2,4)=-(sym_os[5]*v*v*v*v+3*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+4*sym_os[5]+4*sym_os[2]-2*sym_os[0])*v*v+((4*sym_os[8]-8*sym_os[7]-6*sym_os[6]-2*sym_os[5]+sym_os[4]+12*sym_os[3]-6*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+3*sym_os[5]+6*sym_os[2]-2*sym_os[0])*v+sym_os[4]*u*u*u*u+(2*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+sym_os[5]+2*sym_os[2])/4;
		A(2,5)=-(sym_os[5]*v*v*v*v-sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[4]-8*sym_os[3]+4*sym_os[1])*u+2*sym_os[8]-4*sym_os[6]+sym_os[4]+4*sym_os[2]-2*sym_os[0])*v*v+(((-4*sym_os[8])+8*sym_os[7]+2*sym_os[6]+2*sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(8*sym_os[8]-8*sym_os[7]-4*sym_os[6]-2*sym_os[4]+4*sym_os[3]+4*sym_os[1])*u-4*sym_os[8]+2*sym_os[6]-sym_os[5]+sym_os[4]-2*sym_os[2]-2*sym_os[0])*v+sym_os[4]*u*u*u*u-2*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]+2*sym_os[6]-sym_os[5]-4*sym_os[3]+2*sym_os[2])*u*u+((-4*sym_os[8])+4*sym_os[7]-4*sym_os[6]+2*sym_os[4]+4*sym_os[3])*u+2*sym_os[8]+2*sym_os[6]+sym_os[5]-sym_os[4]-2*sym_os[2])/4;
		A(2,6)=-(sym_os[5]*v*v*v*v-sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[4]+8*sym_os[3]-4*sym_os[1])*u+2*sym_os[8]-4*sym_os[6]+sym_os[4]+4*sym_os[2]-2*sym_os[0])*v*v+(((-4*sym_os[8])+8*sym_os[7]+2*sym_os[6]+2*sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-8*sym_os[8])+8*sym_os[7]+4*sym_os[6]+2*sym_os[4]-4*sym_os[3]-4*sym_os[1])*u-4*sym_os[8]+2*sym_os[6]-sym_os[5]+sym_os[4]-2*sym_os[2]-2*sym_os[0])*v+sym_os[4]*u*u*u*u+2*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]+2*sym_os[6]-sym_os[5]-4*sym_os[3]+2*sym_os[2])*u*u+(4*sym_os[8]-4*sym_os[7]+4*sym_os[6]-2*sym_os[4]-4*sym_os[3])*u+2*sym_os[8]+2*sym_os[6]+sym_os[5]-sym_os[4]-2*sym_os[2])/4;
		A(2,7)=-(sym_os[5]*v*v*v*v+3*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[4]+8*sym_os[3]-4*sym_os[1])*u+2*sym_os[8]-4*sym_os[6]+4*sym_os[5]+sym_os[4]+4*sym_os[2]-2*sym_os[0])*v*v+((4*sym_os[8]-8*sym_os[7]-6*sym_os[6]-2*sym_os[5]+sym_os[4]+12*sym_os[3]-6*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(8*sym_os[8]-8*sym_os[7]-12*sym_os[6]+2*sym_os[4]+12*sym_os[3]-4*sym_os[1])*u+4*sym_os[8]-6*sym_os[6]+3*sym_os[5]+sym_os[4]+6*sym_os[2]-2*sym_os[0])*v+sym_os[4]*u*u*u*u+2*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]+4*sym_os[3]-2*sym_os[2])*u*u+(4*sym_os[8]-4*sym_os[7]-4*sym_os[6]-2*sym_os[4]+4*sym_os[3])*u+2*sym_os[8]-2*sym_os[6]+sym_os[5]-sym_os[4]+2*sym_os[2])/4;
		A(2,8)=-(sym_os[5]*v*v*v*v+3*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[4]-8*sym_os[3]+4*sym_os[1])*u+2*sym_os[8]-4*sym_os[6]+4*sym_os[5]+sym_os[4]+4*sym_os[2]-2*sym_os[0])*v*v+((4*sym_os[8]-8*sym_os[7]-6*sym_os[6]-2*sym_os[5]+sym_os[4]+12*sym_os[3]-6*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-8*sym_os[8])+8*sym_os[7]+12*sym_os[6]-2*sym_os[4]-12*sym_os[3]+4*sym_os[1])*u+4*sym_os[8]-6*sym_os[6]+3*sym_os[5]+sym_os[4]+6*sym_os[2]-2*sym_os[0])*v+sym_os[4]*u*u*u*u-2*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]+4*sym_os[3]-2*sym_os[2])*u*u+((-4*sym_os[8])+4*sym_os[7]+4*sym_os[6]+2*sym_os[4]-4*sym_os[3])*u+2*sym_os[8]-2*sym_os[6]+sym_os[5]-sym_os[4]+2*sym_os[2])/4;
		A(3,0)=-(sym_os[5]*v*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(2*sym_os[7]+sym_os[5]-4*sym_os[3]+4*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u-sym_os[5])*v*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u*u+((-sym_os[4])+4*sym_os[1]-2*sym_os[0])*u*u+(2*sym_os[0]-2*sym_os[1])*u)/4;
		A(3,1)=-(sym_os[5]*v*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+8*sym_os[6]+sym_os[5]-2*sym_os[4]-12*sym_os[3]+4*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+sym_os[4]*u*u*u*u-3*sym_os[4]*u*u*u+(2*sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+(sym_os[4]-6*sym_os[1]+2*sym_os[0])*u-sym_os[4]+2*sym_os[1])/4;
		A(3,2)=-(sym_os[5]*v*v*v*v-2*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(2*sym_os[7]+sym_os[5]-4*sym_os[3]+4*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u)*v*v+(((-4*sym_os[8])+8*sym_os[7]+4*sym_os[6]+2*sym_os[5]-8*sym_os[3]+4*sym_os[2])*u*u+((-4*sym_os[7])-2*sym_os[5]+4*sym_os[3]-4*sym_os[2])*u+2*sym_os[5])*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-sym_os[5]-sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+(2*sym_os[7]+sym_os[5]-2*sym_os[1]+2*sym_os[0])*u-sym_os[5])/4;
		A(3,3)=-(sym_os[5]*v*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-8*sym_os[6]+sym_os[5]+2*sym_os[4]+4*sym_os[3]+4*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+sym_os[4]*u*u*u*u+sym_os[4]*u*u*u+((-2*sym_os[4])+4*sym_os[1]-2*sym_os[0])*u*u+((-3*sym_os[4])+2*sym_os[1]+2*sym_os[0])*u-sym_os[4]-2*sym_os[1])/4;
		A(3,4)=-(sym_os[5]*v*v*v*v+2*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(2*sym_os[7]+sym_os[5]-4*sym_os[3]+4*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u)*v*v+((4*sym_os[8]-8*sym_os[7]-4*sym_os[6]-2*sym_os[5]+8*sym_os[3]-4*sym_os[2])*u*u+(4*sym_os[7]+2*sym_os[5]-4*sym_os[3]+4*sym_os[2])*u-2*sym_os[5])*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-sym_os[5]-sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+(2*sym_os[7]+sym_os[5]-2*sym_os[1]+2*sym_os[0])*u-sym_os[5])/4;
		A(3,5)=-(sym_os[5]*v*v*v*v-2*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+8*sym_os[6]+sym_os[5]-2*sym_os[4]-12*sym_os[3]+4*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-4*sym_os[6]+sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+4*sym_os[6]+2*sym_os[5]-8*sym_os[3]+4*sym_os[2])*u*u+(8*sym_os[8]-12*sym_os[7]-8*sym_os[6]-2*sym_os[5]+12*sym_os[3]-4*sym_os[2])*u-4*sym_os[8]+4*sym_os[7]+4*sym_os[6]+2*sym_os[5]-4*sym_os[3])*v+sym_os[4]*u*u*u*u-3*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-sym_os[5]+2*sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+sym_os[5]+sym_os[4]-6*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-sym_os[5]-sym_os[4]+2*sym_os[1])/4;
		A(3,6)=-(sym_os[5]*v*v*v*v-2*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-8*sym_os[6]+sym_os[5]+2*sym_os[4]+4*sym_os[3]+4*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-4*sym_os[6]+sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+4*sym_os[6]+2*sym_os[5]-8*sym_os[3]+4*sym_os[2])*u*u+((-8*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[5]-4*sym_os[3]-4*sym_os[2])*u-4*sym_os[8]-4*sym_os[7]+4*sym_os[6]+2*sym_os[5]+4*sym_os[3])*v+sym_os[4]*u*u*u*u+sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-sym_os[5]-2*sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]+sym_os[5]-3*sym_os[4]+2*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-sym_os[5]-sym_os[4]-2*sym_os[1])/4;
		A(3,7)=-(sym_os[5]*v*v*v*v+2*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-8*sym_os[6]+sym_os[5]+2*sym_os[4]+4*sym_os[3]+4*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-4*sym_os[6]+sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-4*sym_os[6]-2*sym_os[5]+8*sym_os[3]-4*sym_os[2])*u*u+(8*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[5]+4*sym_os[3]+4*sym_os[2])*u+4*sym_os[8]+4*sym_os[7]-4*sym_os[6]-2*sym_os[5]-4*sym_os[3])*v+sym_os[4]*u*u*u*u+sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-sym_os[5]-2*sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]+sym_os[5]-3*sym_os[4]+2*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-sym_os[5]-sym_os[4]-2*sym_os[1])/4;
		A(3,8)=-(sym_os[5]*v*v*v*v+2*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+8*sym_os[6]+sym_os[5]-2*sym_os[4]-12*sym_os[3]+4*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-4*sym_os[6]+sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-4*sym_os[6]-2*sym_os[5]+8*sym_os[3]-4*sym_os[2])*u*u+((-8*sym_os[8])+12*sym_os[7]+8*sym_os[6]+2*sym_os[5]-12*sym_os[3]+4*sym_os[2])*u+4*sym_os[8]-4*sym_os[7]-4*sym_os[6]-2*sym_os[5]+4*sym_os[3])*v+sym_os[4]*u*u*u*u-3*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-sym_os[5]+2*sym_os[4]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+sym_os[5]+sym_os[4]-6*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-sym_os[5]-sym_os[4]+2*sym_os[1])/4;
		A(4,0)=-(sym_os[5]*v*v*v*v-sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+sym_os[5]+4*sym_os[2]-2*sym_os[0])*v*v+((2*sym_os[6]-sym_os[4]-4*sym_os[3]+2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u-2*sym_os[2]+2*sym_os[0])*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u)/4;
		A(4,1)=-(sym_os[5]*v*v*v*v-sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[4]-8*sym_os[3]+4*sym_os[1])*u+2*sym_os[8]-4*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[2]-2*sym_os[0])*v*v+((2*sym_os[6]-sym_os[4]-4*sym_os[3]+2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[6])+2*sym_os[4]+4*sym_os[3]-4*sym_os[1])*u+2*sym_os[6]-sym_os[4]-2*sym_os[2]+2*sym_os[0])*v+sym_os[4]*u*u*u*u-2*sym_os[4]*u*u*u+2*sym_os[4]*u-sym_os[4])/4;
		A(4,2)=-(sym_os[5]*v*v*v*v-3*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+4*sym_os[5]+4*sym_os[2]-2*sym_os[0])*v*v+(((-4*sym_os[8])+8*sym_os[7]+6*sym_os[6]+2*sym_os[5]-sym_os[4]-12*sym_os[3]+6*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u-3*sym_os[5]-6*sym_os[2]+2*sym_os[0])*v+sym_os[4]*u*u*u*u+(2*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+sym_os[5]+2*sym_os[2])/4;
		A(4,3)=-(sym_os[5]*v*v*v*v-sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[4]+8*sym_os[3]-4*sym_os[1])*u+2*sym_os[8]-4*sym_os[6]+sym_os[5]+sym_os[4]+4*sym_os[2]-2*sym_os[0])*v*v+((2*sym_os[6]-sym_os[4]-4*sym_os[3]+2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[6]-2*sym_os[4]-4*sym_os[3]+4*sym_os[1])*u+2*sym_os[6]-sym_os[4]-2*sym_os[2]+2*sym_os[0])*v+sym_os[4]*u*u*u*u+2*sym_os[4]*u*u*u-2*sym_os[4]*u-sym_os[4])/4;
		A(4,4)=-(sym_os[5]*v*v*v*v+sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+4*sym_os[2]-2*sym_os[0])*v*v+((4*sym_os[8]-8*sym_os[7]-2*sym_os[6]-2*sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+sym_os[5]+2*sym_os[2]+2*sym_os[0])*v+sym_os[4]*u*u*u*u+(2*sym_os[8]-4*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+sym_os[5]-2*sym_os[2])/4;
		A(4,5)=-(sym_os[5]*v*v*v*v-3*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[4]-8*sym_os[3]+4*sym_os[1])*u+2*sym_os[8]-4*sym_os[6]+4*sym_os[5]+sym_os[4]+4*sym_os[2]-2*sym_os[0])*v*v+(((-4*sym_os[8])+8*sym_os[7]+6*sym_os[6]+2*sym_os[5]-sym_os[4]-12*sym_os[3]+6*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(8*sym_os[8]-8*sym_os[7]-12*sym_os[6]+2*sym_os[4]+12*sym_os[3]-4*sym_os[1])*u-4*sym_os[8]+6*sym_os[6]-3*sym_os[5]-sym_os[4]-6*sym_os[2]+2*sym_os[0])*v+sym_os[4]*u*u*u*u-2*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]+4*sym_os[3]-2*sym_os[2])*u*u+((-4*sym_os[8])+4*sym_os[7]+4*sym_os[6]+2*sym_os[4]-4*sym_os[3])*u+2*sym_os[8]-2*sym_os[6]+sym_os[5]-sym_os[4]+2*sym_os[2])/4;
		A(4,6)=-(sym_os[5]*v*v*v*v-3*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[4]+8*sym_os[3]-4*sym_os[1])*u+2*sym_os[8]-4*sym_os[6]+4*sym_os[5]+sym_os[4]+4*sym_os[2]-2*sym_os[0])*v*v+(((-4*sym_os[8])+8*sym_os[7]+6*sym_os[6]+2*sym_os[5]-sym_os[4]-12*sym_os[3]+6*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-8*sym_os[8])+8*sym_os[7]+12*sym_os[6]-2*sym_os[4]-12*sym_os[3]+4*sym_os[1])*u-4*sym_os[8]+6*sym_os[6]-3*sym_os[5]-sym_os[4]-6*sym_os[2]+2*sym_os[0])*v+sym_os[4]*u*u*u*u+2*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]+4*sym_os[3]-2*sym_os[2])*u*u+(4*sym_os[8]-4*sym_os[7]-4*sym_os[6]-2*sym_os[4]+4*sym_os[3])*u+2*sym_os[8]-2*sym_os[6]+sym_os[5]-sym_os[4]+2*sym_os[2])/4;
		A(4,7)=-(sym_os[5]*v*v*v*v+sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-4*sym_os[7]-8*sym_os[6]+2*sym_os[4]+8*sym_os[3]-4*sym_os[1])*u+2*sym_os[8]-4*sym_os[6]+sym_os[4]+4*sym_os[2]-2*sym_os[0])*v*v+((4*sym_os[8]-8*sym_os[7]-2*sym_os[6]-2*sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(8*sym_os[8]-8*sym_os[7]-4*sym_os[6]-2*sym_os[4]+4*sym_os[3]+4*sym_os[1])*u+4*sym_os[8]-2*sym_os[6]+sym_os[5]-sym_os[4]+2*sym_os[2]+2*sym_os[0])*v+sym_os[4]*u*u*u*u+2*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]+2*sym_os[6]-sym_os[5]-4*sym_os[3]+2*sym_os[2])*u*u+(4*sym_os[8]-4*sym_os[7]+4*sym_os[6]-2*sym_os[4]-4*sym_os[3])*u+2*sym_os[8]+2*sym_os[6]+sym_os[5]-sym_os[4]-2*sym_os[2])/4;
		A(4,8)=-(sym_os[5]*v*v*v*v+sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+4*sym_os[7]+8*sym_os[6]-2*sym_os[4]-8*sym_os[3]+4*sym_os[1])*u+2*sym_os[8]-4*sym_os[6]+sym_os[4]+4*sym_os[2]-2*sym_os[0])*v*v+((4*sym_os[8]-8*sym_os[7]-2*sym_os[6]-2*sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-8*sym_os[8])+8*sym_os[7]+4*sym_os[6]+2*sym_os[4]-4*sym_os[3]-4*sym_os[1])*u+4*sym_os[8]-2*sym_os[6]+sym_os[5]-sym_os[4]+2*sym_os[2]+2*sym_os[0])*v+sym_os[4]*u*u*u*u-2*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]+2*sym_os[6]-sym_os[5]-4*sym_os[3]+2*sym_os[2])*u*u+((-4*sym_os[8])+4*sym_os[7]-4*sym_os[6]+2*sym_os[4]+4*sym_os[3])*u+2*sym_os[8]+2*sym_os[6]+sym_os[5]-sym_os[4]-2*sym_os[2])/4;
		A(5,0)=(sym_os[5]*v*v*v*v+sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-2*sym_os[7])-sym_os[5]+4*sym_os[3]-4*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u)*v*v+(((-2*sym_os[6])+sym_os[4]+4*sym_os[3]-2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(2*sym_os[3]-2*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u)*v+sym_os[4]*u*u*u*u+sym_os[4]*u*u*u)/8;
		A(5,1)=(sym_os[5]*v*v*v*v+sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+8*sym_os[6]-sym_os[5]-2*sym_os[4]-4*sym_os[3]-4*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-4*sym_os[6]+sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+(((-2*sym_os[6])+sym_os[4]+4*sym_os[3]-2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[6]-2*sym_os[4]-2*sym_os[3]-2*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u-2*sym_os[6]+sym_os[4]-2*sym_os[3]+2*sym_os[1])*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u*u-sym_os[4]*u*u+sym_os[4]*u)/8;
		A(5,2)=(sym_os[5]*v*v*v*v-sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-2*sym_os[7])-sym_os[5]+4*sym_os[3]-4*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u-sym_os[5])*v*v+(((-4*sym_os[8])+8*sym_os[7]+2*sym_os[6]+2*sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[7]+2*sym_os[5]-2*sym_os[3]+2*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u+sym_os[5])*v+sym_os[4]*u*u*u*u+sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]+2*sym_os[6]-sym_os[5]-4*sym_os[3]+2*sym_os[2])*u*u+((-2*sym_os[7])-sym_os[5]-2*sym_os[3]+2*sym_os[2])*u)/8;
		A(5,3)=(sym_os[5]*v*v*v*v+sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-8*sym_os[6]-sym_os[5]+2*sym_os[4]+12*sym_os[3]-4*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-4*sym_os[6]+sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+(((-2*sym_os[6])+sym_os[4]+4*sym_os[3]-2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[6])+2*sym_os[4]+6*sym_os[3]-2*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u-2*sym_os[6]+sym_os[4]+2*sym_os[3]-2*sym_os[1])*v+sym_os[4]*u*u*u*u+3*sym_os[4]*u*u*u+3*sym_os[4]*u*u+sym_os[4]*u)/8;
		A(5,4)=(sym_os[5]*v*v*v*v+3*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-2*sym_os[7])-sym_os[5]+4*sym_os[3]-4*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u+3*sym_os[5])*v*v+((4*sym_os[8]-8*sym_os[7]-6*sym_os[6]-2*sym_os[5]+sym_os[4]+12*sym_os[3]-6*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[7])-2*sym_os[5]+6*sym_os[3]-6*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u+sym_os[5])*v+sym_os[4]*u*u*u*u+sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]+4*sym_os[3]-2*sym_os[2])*u*u+((-2*sym_os[7])-sym_os[5]+2*sym_os[3]-2*sym_os[2])*u)/8;
		A(5,5)=(sym_os[5]*v*v*v*v-sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+8*sym_os[6]-sym_os[5]-2*sym_os[4]-4*sym_os[3]-4*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+2*sym_os[6]+2*sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(8*sym_os[8]-4*sym_os[7]-4*sym_os[6]+2*sym_os[5]-2*sym_os[4]+2*sym_os[3]+2*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u-4*sym_os[8]-4*sym_os[7]+2*sym_os[6]+sym_os[5]+sym_os[4]+2*sym_os[3]+2*sym_os[1])*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+((-4*sym_os[8])+2*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+2*sym_os[3]+2*sym_os[2])*u+2*sym_os[8]+2*sym_os[7]+2*sym_os[6]+2*sym_os[3])/8;
		A(5,6)=(sym_os[5]*v*v*v*v-sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-8*sym_os[6]-sym_os[5]+2*sym_os[4]+12*sym_os[3]-4*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+2*sym_os[6]+2*sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-8*sym_os[8])+12*sym_os[7]+4*sym_os[6]+2*sym_os[5]+2*sym_os[4]-6*sym_os[3]+2*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u-4*sym_os[8]+4*sym_os[7]+2*sym_os[6]+sym_os[5]+sym_os[4]-2*sym_os[3]-2*sym_os[1])*v+sym_os[4]*u*u*u*u+3*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]+2*sym_os[6]-sym_os[5]+3*sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+(4*sym_os[8]-6*sym_os[7]+4*sym_os[6]-sym_os[5]+sym_os[4]-6*sym_os[3]+2*sym_os[2])*u+2*sym_os[8]-2*sym_os[7]+2*sym_os[6]-2*sym_os[3])/8;
		A(5,7)=(sym_os[5]*v*v*v*v+3*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-8*sym_os[6]-sym_os[5]+2*sym_os[4]+12*sym_os[3]-4*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-4*sym_os[6]+3*sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-6*sym_os[6]-2*sym_os[5]+sym_os[4]+12*sym_os[3]-6*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(8*sym_os[8]-12*sym_os[7]-12*sym_os[6]-2*sym_os[5]+2*sym_os[4]+18*sym_os[3]-6*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u+4*sym_os[8]-4*sym_os[7]-6*sym_os[6]+sym_os[5]+sym_os[4]+6*sym_os[3]-2*sym_os[1])*v+sym_os[4]*u*u*u*u+3*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]+3*sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+(4*sym_os[8]-6*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+6*sym_os[3]-2*sym_os[2])*u+2*sym_os[8]-2*sym_os[7]-2*sym_os[6]+2*sym_os[3])/8;
		A(5,8)=(sym_os[5]*v*v*v*v+3*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+8*sym_os[6]-sym_os[5]-2*sym_os[4]-4*sym_os[3]-4*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-4*sym_os[6]+3*sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-6*sym_os[6]-2*sym_os[5]+sym_os[4]+12*sym_os[3]-6*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-8*sym_os[8])+4*sym_os[7]+12*sym_os[6]-2*sym_os[5]-2*sym_os[4]-6*sym_os[3]-6*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u+4*sym_os[8]+4*sym_os[7]-6*sym_os[6]+sym_os[5]+sym_os[4]-6*sym_os[3]+2*sym_os[1])*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+((-4*sym_os[8])+2*sym_os[7]+4*sym_os[6]-sym_os[5]+sym_os[4]-2*sym_os[3]-2*sym_os[2])*u+2*sym_os[8]+2*sym_os[7]-2*sym_os[6]-2*sym_os[3])/8;
		A(6,0)=(sym_os[5]*v*v*v*v+sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(2*sym_os[7]+sym_os[5]-4*sym_os[3]+4*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u)*v*v+(((-2*sym_os[6])+sym_os[4]+4*sym_os[3]-2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-2*sym_os[3])+2*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u)*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u*u)/8;
		A(6,1)=(sym_os[5]*v*v*v*v+sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+8*sym_os[6]+sym_os[5]-2*sym_os[4]-12*sym_os[3]+4*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-4*sym_os[6]+sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+(((-2*sym_os[6])+sym_os[4]+4*sym_os[3]-2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[6]-2*sym_os[4]-6*sym_os[3]+2*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u-2*sym_os[6]+sym_os[4]+2*sym_os[3]-2*sym_os[1])*v+sym_os[4]*u*u*u*u-3*sym_os[4]*u*u*u+3*sym_os[4]*u*u-sym_os[4]*u)/8;
		A(6,2)=(sym_os[5]*v*v*v*v-sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(2*sym_os[7]+sym_os[5]-4*sym_os[3]+4*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u-sym_os[5])*v*v+(((-4*sym_os[8])+8*sym_os[7]+2*sym_os[6]+2*sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[7])-2*sym_os[5]+2*sym_os[3]-2*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u+sym_os[5])*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]+2*sym_os[6]-sym_os[5]-4*sym_os[3]+2*sym_os[2])*u*u+(2*sym_os[7]+sym_os[5]+2*sym_os[3]-2*sym_os[2])*u)/8;
		A(6,3)=(sym_os[5]*v*v*v*v+sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-8*sym_os[6]+sym_os[5]+2*sym_os[4]+4*sym_os[3]+4*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-4*sym_os[6]+sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+(((-2*sym_os[6])+sym_os[4]+4*sym_os[3]-2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[6])+2*sym_os[4]+2*sym_os[3]+2*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u-2*sym_os[6]+sym_os[4]-2*sym_os[3]+2*sym_os[1])*v+sym_os[4]*u*u*u*u+sym_os[4]*u*u*u-sym_os[4]*u*u-sym_os[4]*u)/8;
		A(6,4)=(sym_os[5]*v*v*v*v+3*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(2*sym_os[7]+sym_os[5]-4*sym_os[3]+4*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u+3*sym_os[5])*v*v+((4*sym_os[8]-8*sym_os[7]-6*sym_os[6]-2*sym_os[5]+sym_os[4]+12*sym_os[3]-6*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[7]+2*sym_os[5]-6*sym_os[3]+6*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u+sym_os[5])*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]+4*sym_os[3]-2*sym_os[2])*u*u+(2*sym_os[7]+sym_os[5]-2*sym_os[3]+2*sym_os[2])*u)/8;
		A(6,5)=(sym_os[5]*v*v*v*v-sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+8*sym_os[6]+sym_os[5]-2*sym_os[4]-12*sym_os[3]+4*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+2*sym_os[6]+2*sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(8*sym_os[8]-12*sym_os[7]-4*sym_os[6]-2*sym_os[5]-2*sym_os[4]+6*sym_os[3]-2*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u-4*sym_os[8]+4*sym_os[7]+2*sym_os[6]+sym_os[5]+sym_os[4]-2*sym_os[3]-2*sym_os[1])*v+sym_os[4]*u*u*u*u-3*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]+2*sym_os[6]-sym_os[5]+3*sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+((-4*sym_os[8])+6*sym_os[7]-4*sym_os[6]+sym_os[5]-sym_os[4]+6*sym_os[3]-2*sym_os[2])*u+2*sym_os[8]-2*sym_os[7]+2*sym_os[6]-2*sym_os[3])/8;
		A(6,6)=(sym_os[5]*v*v*v*v-sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-8*sym_os[6]+sym_os[5]+2*sym_os[4]+4*sym_os[3]+4*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+2*sym_os[6]+2*sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-8*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]+2*sym_os[4]-2*sym_os[3]-2*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u-4*sym_os[8]-4*sym_os[7]+2*sym_os[6]+sym_os[5]+sym_os[4]+2*sym_os[3]+2*sym_os[1])*v+sym_os[4]*u*u*u*u+sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+(4*sym_os[8]-2*sym_os[7]+4*sym_os[6]+sym_os[5]-sym_os[4]-2*sym_os[3]-2*sym_os[2])*u+2*sym_os[8]+2*sym_os[7]+2*sym_os[6]+2*sym_os[3])/8;
		A(6,7)=(sym_os[5]*v*v*v*v+3*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-8*sym_os[6]+sym_os[5]+2*sym_os[4]+4*sym_os[3]+4*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-4*sym_os[6]+3*sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-6*sym_os[6]-2*sym_os[5]+sym_os[4]+12*sym_os[3]-6*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(8*sym_os[8]-4*sym_os[7]-12*sym_os[6]+2*sym_os[5]+2*sym_os[4]+6*sym_os[3]+6*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u+4*sym_os[8]+4*sym_os[7]-6*sym_os[6]+sym_os[5]+sym_os[4]-6*sym_os[3]+2*sym_os[1])*v+sym_os[4]*u*u*u*u+sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+(4*sym_os[8]-2*sym_os[7]-4*sym_os[6]+sym_os[5]-sym_os[4]+2*sym_os[3]+2*sym_os[2])*u+2*sym_os[8]+2*sym_os[7]-2*sym_os[6]-2*sym_os[3])/8;
		A(6,8)=(sym_os[5]*v*v*v*v+3*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+8*sym_os[6]+sym_os[5]-2*sym_os[4]-12*sym_os[3]+4*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-4*sym_os[6]+3*sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-6*sym_os[6]-2*sym_os[5]+sym_os[4]+12*sym_os[3]-6*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-8*sym_os[8])+12*sym_os[7]+12*sym_os[6]+2*sym_os[5]-2*sym_os[4]-18*sym_os[3]+6*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u+4*sym_os[8]-4*sym_os[7]-6*sym_os[6]+sym_os[5]+sym_os[4]+6*sym_os[3]-2*sym_os[1])*v+sym_os[4]*u*u*u*u-3*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]+3*sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+((-4*sym_os[8])+6*sym_os[7]+4*sym_os[6]+sym_os[5]-sym_os[4]-6*sym_os[3]+2*sym_os[2])*u+2*sym_os[8]-2*sym_os[7]-2*sym_os[6]+2*sym_os[3])/8;
		A(7,0)=(sym_os[5]*v*v*v*v-sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(2*sym_os[7]+sym_os[5]-4*sym_os[3]+4*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u)*v*v+((2*sym_os[6]-sym_os[4]-4*sym_os[3]+2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(2*sym_os[3]-2*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u)*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u*u)/8;
		A(7,1)=(sym_os[5]*v*v*v*v-sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+8*sym_os[6]+sym_os[5]-2*sym_os[4]-12*sym_os[3]+4*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-4*sym_os[6]+sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+((2*sym_os[6]-sym_os[4]-4*sym_os[3]+2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[6])+2*sym_os[4]+6*sym_os[3]-2*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u+2*sym_os[6]-sym_os[4]-2*sym_os[3]+2*sym_os[1])*v+sym_os[4]*u*u*u*u-3*sym_os[4]*u*u*u+3*sym_os[4]*u*u-sym_os[4]*u)/8;
		A(7,2)=(sym_os[5]*v*v*v*v-3*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(2*sym_os[7]+sym_os[5]-4*sym_os[3]+4*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u+3*sym_os[5])*v*v+(((-4*sym_os[8])+8*sym_os[7]+6*sym_os[6]+2*sym_os[5]-sym_os[4]-12*sym_os[3]+6*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[7])-2*sym_os[5]+6*sym_os[3]-6*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u-sym_os[5])*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]+4*sym_os[3]-2*sym_os[2])*u*u+(2*sym_os[7]+sym_os[5]-2*sym_os[3]+2*sym_os[2])*u)/8;
		A(7,3)=(sym_os[5]*v*v*v*v-sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-8*sym_os[6]+sym_os[5]+2*sym_os[4]+4*sym_os[3]+4*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-4*sym_os[6]+sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+((2*sym_os[6]-sym_os[4]-4*sym_os[3]+2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[6]-2*sym_os[4]-2*sym_os[3]-2*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u+2*sym_os[6]-sym_os[4]+2*sym_os[3]-2*sym_os[1])*v+sym_os[4]*u*u*u*u+sym_os[4]*u*u*u-sym_os[4]*u*u-sym_os[4]*u)/8;
		A(7,4)=(sym_os[5]*v*v*v*v+sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(2*sym_os[7]+sym_os[5]-4*sym_os[3]+4*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u-sym_os[5])*v*v+((4*sym_os[8]-8*sym_os[7]-2*sym_os[6]-2*sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[7]+2*sym_os[5]-2*sym_os[3]+2*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u-sym_os[5])*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]+2*sym_os[6]-sym_os[5]-4*sym_os[3]+2*sym_os[2])*u*u+(2*sym_os[7]+sym_os[5]+2*sym_os[3]-2*sym_os[2])*u)/8;
		A(7,5)=(sym_os[5]*v*v*v*v-3*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+8*sym_os[6]+sym_os[5]-2*sym_os[4]-12*sym_os[3]+4*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-4*sym_os[6]+3*sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+6*sym_os[6]+2*sym_os[5]-sym_os[4]-12*sym_os[3]+6*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(8*sym_os[8]-12*sym_os[7]-12*sym_os[6]-2*sym_os[5]+2*sym_os[4]+18*sym_os[3]-6*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u-4*sym_os[8]+4*sym_os[7]+6*sym_os[6]-sym_os[5]-sym_os[4]-6*sym_os[3]+2*sym_os[1])*v+sym_os[4]*u*u*u*u-3*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]+3*sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+((-4*sym_os[8])+6*sym_os[7]+4*sym_os[6]+sym_os[5]-sym_os[4]-6*sym_os[3]+2*sym_os[2])*u+2*sym_os[8]-2*sym_os[7]-2*sym_os[6]+2*sym_os[3])/8;
		A(7,6)=(sym_os[5]*v*v*v*v-3*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-8*sym_os[6]+sym_os[5]+2*sym_os[4]+4*sym_os[3]+4*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-4*sym_os[6]+3*sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+6*sym_os[6]+2*sym_os[5]-sym_os[4]-12*sym_os[3]+6*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-8*sym_os[8])+4*sym_os[7]+12*sym_os[6]-2*sym_os[5]-2*sym_os[4]-6*sym_os[3]-6*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u-4*sym_os[8]-4*sym_os[7]+6*sym_os[6]-sym_os[5]-sym_os[4]+6*sym_os[3]-2*sym_os[1])*v+sym_os[4]*u*u*u*u+sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+(4*sym_os[8]-2*sym_os[7]-4*sym_os[6]+sym_os[5]-sym_os[4]+2*sym_os[3]+2*sym_os[2])*u+2*sym_os[8]+2*sym_os[7]-2*sym_os[6]-2*sym_os[3])/8;
		A(7,7)=(sym_os[5]*v*v*v*v+sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-2*sym_os[7]-8*sym_os[6]+sym_os[5]+2*sym_os[4]+4*sym_os[3]+4*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-2*sym_os[6]-2*sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(8*sym_os[8]-4*sym_os[7]-4*sym_os[6]+2*sym_os[5]-2*sym_os[4]+2*sym_os[3]+2*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u+4*sym_os[8]+4*sym_os[7]-2*sym_os[6]-sym_os[5]-sym_os[4]-2*sym_os[3]-2*sym_os[1])*v+sym_os[4]*u*u*u*u+sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+(4*sym_os[8]-2*sym_os[7]+4*sym_os[6]+sym_os[5]-sym_os[4]-2*sym_os[3]-2*sym_os[2])*u+2*sym_os[8]+2*sym_os[7]+2*sym_os[6]+2*sym_os[3])/8;
		A(7,8)=(sym_os[5]*v*v*v*v+sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+6*sym_os[7]+8*sym_os[6]+sym_os[5]-2*sym_os[4]-12*sym_os[3]+4*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-2*sym_os[6]-2*sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-8*sym_os[8])+12*sym_os[7]+4*sym_os[6]+2*sym_os[5]+2*sym_os[4]-6*sym_os[3]+2*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u+4*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]-sym_os[4]+2*sym_os[3]+2*sym_os[1])*v+sym_os[4]*u*u*u*u-3*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]+2*sym_os[6]-sym_os[5]+3*sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+((-4*sym_os[8])+6*sym_os[7]-4*sym_os[6]+sym_os[5]-sym_os[4]+6*sym_os[3]-2*sym_os[2])*u+2*sym_os[8]-2*sym_os[7]+2*sym_os[6]-2*sym_os[3])/8;
		A(8,0)=(sym_os[5]*v*v*v*v-sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-2*sym_os[7])-sym_os[5]+4*sym_os[3]-4*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u)*v*v+((2*sym_os[6]-sym_os[4]-4*sym_os[3]+2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-2*sym_os[3])+2*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u)*v+sym_os[4]*u*u*u*u+sym_os[4]*u*u*u)/8;
		A(8,1)=(sym_os[5]*v*v*v*v-sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+8*sym_os[6]-sym_os[5]-2*sym_os[4]-4*sym_os[3]-4*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-4*sym_os[6]+sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+((2*sym_os[6]-sym_os[4]-4*sym_os[3]+2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[6])+2*sym_os[4]+2*sym_os[3]+2*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u+2*sym_os[6]-sym_os[4]+2*sym_os[3]-2*sym_os[1])*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u*u-sym_os[4]*u*u+sym_os[4]*u)/8;
		A(8,2)=(sym_os[5]*v*v*v*v-3*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-2*sym_os[7])-sym_os[5]+4*sym_os[3]-4*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u+3*sym_os[5])*v*v+(((-4*sym_os[8])+8*sym_os[7]+6*sym_os[6]+2*sym_os[5]-sym_os[4]-12*sym_os[3]+6*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[7]+2*sym_os[5]-6*sym_os[3]+6*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u-sym_os[5])*v+sym_os[4]*u*u*u*u+sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]+4*sym_os[3]-2*sym_os[2])*u*u+((-2*sym_os[7])-sym_os[5]+2*sym_os[3]-2*sym_os[2])*u)/8;
		A(8,3)=(sym_os[5]*v*v*v*v-sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-8*sym_os[6]-sym_os[5]+2*sym_os[4]+12*sym_os[3]-4*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-4*sym_os[6]+sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+((2*sym_os[6]-sym_os[4]-4*sym_os[3]+2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(4*sym_os[6]-2*sym_os[4]-6*sym_os[3]+2*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u+2*sym_os[6]-sym_os[4]-2*sym_os[3]+2*sym_os[1])*v+sym_os[4]*u*u*u*u+3*sym_os[4]*u*u*u+3*sym_os[4]*u*u+sym_os[4]*u)/8;
		A(8,4)=(sym_os[5]*v*v*v*v+sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-2*sym_os[7])-sym_os[5]+4*sym_os[3]-4*sym_os[2]-2*sym_os[1]+2*sym_os[0])*u-sym_os[5])*v*v+((4*sym_os[8]-8*sym_os[7]-2*sym_os[6]-2*sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-4*sym_os[7])-2*sym_os[5]+2*sym_os[3]-2*sym_os[2]+2*sym_os[1]-2*sym_os[0])*u-sym_os[5])*v+sym_os[4]*u*u*u*u+sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]+2*sym_os[6]-sym_os[5]-4*sym_os[3]+2*sym_os[2])*u*u+((-2*sym_os[7])-sym_os[5]-2*sym_os[3]+2*sym_os[2])*u)/8;
		A(8,5)=(sym_os[5]*v*v*v*v-3*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+8*sym_os[6]-sym_os[5]-2*sym_os[4]-4*sym_os[3]-4*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-4*sym_os[6]+3*sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+6*sym_os[6]+2*sym_os[5]-sym_os[4]-12*sym_os[3]+6*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(8*sym_os[8]-4*sym_os[7]-12*sym_os[6]+2*sym_os[5]+2*sym_os[4]+6*sym_os[3]+6*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u-4*sym_os[8]-4*sym_os[7]+6*sym_os[6]-sym_os[5]-sym_os[4]+6*sym_os[3]-2*sym_os[1])*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+((-4*sym_os[8])+2*sym_os[7]+4*sym_os[6]-sym_os[5]+sym_os[4]-2*sym_os[3]-2*sym_os[2])*u+2*sym_os[8]+2*sym_os[7]-2*sym_os[6]-2*sym_os[3])/8;
		A(8,6)=(sym_os[5]*v*v*v*v-3*sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-8*sym_os[6]-sym_os[5]+2*sym_os[4]+12*sym_os[3]-4*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-4*sym_os[6]+3*sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+(((-4*sym_os[8])+8*sym_os[7]+6*sym_os[6]+2*sym_os[5]-sym_os[4]-12*sym_os[3]+6*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-8*sym_os[8])+12*sym_os[7]+12*sym_os[6]+2*sym_os[5]-2*sym_os[4]-18*sym_os[3]+6*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u-4*sym_os[8]+4*sym_os[7]+6*sym_os[6]-sym_os[5]-sym_os[4]-6*sym_os[3]+2*sym_os[1])*v+sym_os[4]*u*u*u*u+3*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]+3*sym_os[4]+4*sym_os[3]-2*sym_os[2])*u*u+(4*sym_os[8]-6*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+6*sym_os[3]-2*sym_os[2])*u+2*sym_os[8]-2*sym_os[7]-2*sym_os[6]+2*sym_os[3])/8;
		A(8,7)=(sym_os[5]*v*v*v*v+sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+(4*sym_os[8]-6*sym_os[7]-8*sym_os[6]-sym_os[5]+2*sym_os[4]+12*sym_os[3]-4*sym_os[2]-6*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]-2*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+4*sym_os[3]-2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-2*sym_os[6]-2*sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+(8*sym_os[8]-12*sym_os[7]-4*sym_os[6]-2*sym_os[5]-2*sym_os[4]+6*sym_os[3]-2*sym_os[2]+6*sym_os[1]-2*sym_os[0])*u+4*sym_os[8]-4*sym_os[7]-2*sym_os[6]-sym_os[5]-sym_os[4]+2*sym_os[3]+2*sym_os[1])*v+sym_os[4]*u*u*u*u+3*sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]+2*sym_os[6]-sym_os[5]+3*sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+(4*sym_os[8]-6*sym_os[7]+4*sym_os[6]-sym_os[5]+sym_os[4]-6*sym_os[3]+2*sym_os[2])*u+2*sym_os[8]-2*sym_os[7]+2*sym_os[6]-2*sym_os[3])/8;
		A(8,8)=(sym_os[5]*v*v*v*v+sym_os[5]*v*v*v+((2*sym_os[8]-4*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+8*sym_os[3]-4*sym_os[2]-4*sym_os[1]+2*sym_os[0])*u*u+((-4*sym_os[8])+2*sym_os[7]+8*sym_os[6]-sym_os[5]-2*sym_os[4]-4*sym_os[3]-4*sym_os[2]+2*sym_os[1]+2*sym_os[0])*u+2*sym_os[8]+2*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]-4*sym_os[3]+2*sym_os[1])*v*v+((4*sym_os[8]-8*sym_os[7]-2*sym_os[6]-2*sym_os[5]-sym_os[4]+4*sym_os[3]-2*sym_os[2]+4*sym_os[1]-2*sym_os[0])*u*u+((-8*sym_os[8])+4*sym_os[7]+4*sym_os[6]-2*sym_os[5]+2*sym_os[4]-2*sym_os[3]-2*sym_os[2]-2*sym_os[1]-2*sym_os[0])*u+4*sym_os[8]+4*sym_os[7]-2*sym_os[6]-sym_os[5]-sym_os[4]-2*sym_os[3]-2*sym_os[1])*v+sym_os[4]*u*u*u*u-sym_os[4]*u*u*u+(2*sym_os[8]-4*sym_os[7]+2*sym_os[6]-sym_os[5]-sym_os[4]-4*sym_os[3]+2*sym_os[2])*u*u+((-4*sym_os[8])+2*sym_os[7]-4*sym_os[6]-sym_os[5]+sym_os[4]+2*sym_os[3]+2*sym_os[2])*u+2*sym_os[8]+2*sym_os[7]+2*sym_os[6]+2*sym_os[3])/8;
		break;
	default:
		printf("Error: matrix A not implemented for Q%d and CLBM2\n",Q);
	}
}


void LBM::setA_MRT1_OG(matrix &A, int Q, ex u, ex v, ex w)
{
	ex one=1;
	switch (Q)
	{
	case 3:
		A(0,0) = ( (ex)2*sym_os[2] + sym_os[0] )/(ex)3;
		A(0,1) = ( sym_os[0] - sym_os[2] )/(ex)3;
		A(0,2) = ( sym_os[0] - sym_os[2] )/(ex)3;

		A(1,0) = ( sym_os[0] - sym_os[2] )/(ex)3;
		A(1,1) = ( (ex)2*sym_os[0] + (ex)3*sym_os[1] + sym_os[2] )/(ex)6;
		A(1,2) = ( (ex)2*sym_os[0] - (ex)3*sym_os[1] + sym_os[2] )/(ex)6;

		A(2,0) = ( sym_os[0] - sym_os[2] )/(ex)3;;
		A(2,1) = ( (ex)2*sym_os[0] - (ex)3*sym_os[1] + sym_os[2] )/(ex)6;
		A(2,2) = ( (ex)2*sym_os[0] + (ex)3*sym_os[1] + sym_os[2] )/(ex)6;
		break;
	case 5:
		A(0,0) = n2o3*sym_os[4]+n2o15*sym_os[3]+n1o5*sym_os[0];
		A(0,1) = n1o5*sym_os[0]-n1o5*sym_os[3];
		A(0,2) = -n1o3*sym_os[4]+n2o15*sym_os[3]+n1o5*sym_os[0];
		A(0,3) = n1o5*sym_os[0]-n1o5*sym_os[3];
		A(0,4) = -n1o3*sym_os[4]+n2o15*sym_os[3]+n1o5*sym_os[0];
		A(1,0) = n1o5*sym_os[0]-n1o5*sym_os[3];
		A(1,1) = n3o10*sym_os[3]+n1o2*sym_os[1]+n1o5*sym_os[0];
		A(1,2) = n1o5*sym_os[0]-n1o5*sym_os[3];
		A(1,3) = n3o10*sym_os[3]-n1o2*sym_os[1]+n1o5*sym_os[0];
		A(1,4) = n1o5*sym_os[0]-n1o5*sym_os[3];
		A(2,0) = -n1o3*sym_os[4]+n2o15*sym_os[3]+n1o5*sym_os[0];
		A(2,1) = n1o5*sym_os[0]-n1o5*sym_os[3];
		A(2,2) = n1o6*sym_os[4]+n2o15*sym_os[3]+n1o2*sym_os[2]+n1o5*sym_os[0];
		A(2,3) = n1o5*sym_os[0]-n1o5*sym_os[3];
		A(2,4) = n1o6*sym_os[4]+n2o15*sym_os[3]-n1o2*sym_os[2]+n1o5*sym_os[0];
		A(3,0) = n1o5*sym_os[0]-n1o5*sym_os[3];
		A(3,1) = n3o10*sym_os[3]-n1o2*sym_os[1]+n1o5*sym_os[0];
		A(3,2) = n1o5*sym_os[0]-n1o5*sym_os[3];
		A(3,3) = n3o10*sym_os[3]+n1o2*sym_os[1]+n1o5*sym_os[0];
		A(3,4) = n1o5*sym_os[0]-n1o5*sym_os[3];
		A(4,0) = -n1o3*sym_os[4]+n2o15*sym_os[3]+n1o5*sym_os[0];
		A(4,1) = n1o5*sym_os[0]-n1o5*sym_os[3];
		A(4,2) = n1o6*sym_os[4]+n2o15*sym_os[3]-n1o2*sym_os[2]+n1o5*sym_os[0];
		A(4,3) = n1o5*sym_os[0]-n1o5*sym_os[3];
		A(4,4) = n1o6*sym_os[4]+n2o15*sym_os[3]+n1o2*sym_os[2]+n1o5*sym_os[0];
		break;
	case 7:
		A(0,0) = (2*sym_os[6])/3+(2*sym_os[5])/15+(2*sym_os[4])/35+sym_os[0]/7;
		A(0,1) = sym_os[0]/7-sym_os[4]/7;
		A(0,2) = -sym_os[5]/5+(2*sym_os[4])/35+sym_os[0]/7;
		A(0,3) = -sym_os[6]/3+(2*sym_os[5])/15+(2*sym_os[4])/35+sym_os[0]/7;
		A(0,4) = sym_os[0]/7-sym_os[4]/7;
		A(0,5) = -sym_os[5]/5+(2*sym_os[4])/35+sym_os[0]/7;
		A(0,6) = -sym_os[6]/3+(2*sym_os[5])/15+(2*sym_os[4])/35+sym_os[0]/7;

		A(1,0) = sym_os[0]/7-sym_os[4]/7;
		A(1,1) = (5*sym_os[4])/14+sym_os[1]/2+sym_os[0]/7;
		A(1,2) = sym_os[0]/7-sym_os[4]/7;
		A(1,3) = sym_os[0]/7-sym_os[4]/7;
		A(1,4) = (5*sym_os[4])/14-sym_os[1]/2+sym_os[0]/7;
		A(1,5) = sym_os[0]/7-sym_os[4]/7;
		A(1,6) = sym_os[0]/7-sym_os[4]/7;

		A(2,0) = -sym_os[5]/5+(2*sym_os[4])/35+sym_os[0]/7;
		A(2,1) = sym_os[0]/7-sym_os[4]/7;
		A(2,2) = (3*sym_os[5])/10+(2*sym_os[4])/35+sym_os[2]/2+sym_os[0]/7;
		A(2,3) = -sym_os[5]/5+(2*sym_os[4])/35+sym_os[0]/7;
		A(2,4) = sym_os[0]/7-sym_os[4]/7;
		A(2,5) = (3*sym_os[5])/10+(2*sym_os[4])/35-sym_os[2]/2+sym_os[0]/7;
		A(2,6) = -sym_os[5]/5+(2*sym_os[4])/35+sym_os[0]/7;

		A(3,0) = -sym_os[6]/3+(2*sym_os[5])/15+(2*sym_os[4])/35+sym_os[0]/7;
		A(3,1) = sym_os[0]/7-sym_os[4]/7;
		A(3,2) = -sym_os[5]/5+(2*sym_os[4])/35+sym_os[0]/7;
		A(3,3) = sym_os[6]/6+(2*sym_os[5])/15+(2*sym_os[4])/35+sym_os[3]/2+sym_os[0]/7;
		A(3,4) = sym_os[0]/7-sym_os[4]/7;
		A(3,5) = -sym_os[5]/5+(2*sym_os[4])/35+sym_os[0]/7;
		A(3,6) = sym_os[6]/6+(2*sym_os[5])/15+(2*sym_os[4])/35-sym_os[3]/2+sym_os[0]/7;

		A(4,0) = sym_os[0]/7-sym_os[4]/7;
		A(4,1) = (5*sym_os[4])/14-sym_os[1]/2+sym_os[0]/7;
		A(4,2) = sym_os[0]/7-sym_os[4]/7;
		A(4,3) = sym_os[0]/7-sym_os[4]/7;
		A(4,4) = (5*sym_os[4])/14+sym_os[1]/2+sym_os[0]/7;
		A(4,5) = sym_os[0]/7-sym_os[4]/7;
		A(4,6) = sym_os[0]/7-sym_os[4]/7;

		A(5,0) = -sym_os[5]/5+(2*sym_os[4])/35+sym_os[0]/7;
		A(5,1) = sym_os[0]/7-sym_os[4]/7;
		A(5,2) = (3*sym_os[5])/10+(2*sym_os[4])/35-sym_os[2]/2+sym_os[0]/7;
		A(5,3) = -sym_os[5]/5+(2*sym_os[4])/35+sym_os[0]/7;
		A(5,4) = sym_os[0]/7-sym_os[4]/7;
		A(5,5) = (3*sym_os[5])/10+(2*sym_os[4])/35+sym_os[2]/2+sym_os[0]/7;
		A(5,6) = -sym_os[5]/5+(2*sym_os[4])/35+sym_os[0]/7;

		A(6,0) = -sym_os[6]/3+(2*sym_os[5])/15+(2*sym_os[4])/35+sym_os[0]/7;
		A(6,1) = sym_os[0]/7-sym_os[4]/7;
		A(6,2) = -sym_os[5]/5+(2*sym_os[4])/35+sym_os[0]/7;
		A(6,3) = sym_os[6]/6+(2*sym_os[5])/15+(2*sym_os[4])/35-sym_os[3]/2+sym_os[0]/7;
		A(6,4) = sym_os[0]/7-sym_os[4]/7;
		A(6,5) = -sym_os[5]/5+(2*sym_os[4])/35+sym_os[0]/7;
		A(6,6) = sym_os[6]/6+(2*sym_os[5])/15+(2*sym_os[4])/35+sym_os[3]/2+sym_os[0]/7;
		break;
	case 9:
		A(0,0)=(4*sym_os[8])/9+(2*sym_os[5])/9+(2*sym_os[4])/9+sym_os[0]/9;
		A(0,1)=(-(2*sym_os[8])/9)+(2*sym_os[5])/9-sym_os[4]/9+sym_os[0]/9;
		A(0,2)=(-(2*sym_os[8])/9)-sym_os[5]/9+(2*sym_os[4])/9+sym_os[0]/9;
		A(0,3)=(-(2*sym_os[8])/9)+(2*sym_os[5])/9-sym_os[4]/9+sym_os[0]/9;
		A(0,4)=(-(2*sym_os[8])/9)-sym_os[5]/9+(2*sym_os[4])/9+sym_os[0]/9;
		A(0,5)=sym_os[8]/9-sym_os[5]/9-sym_os[4]/9+sym_os[0]/9;
		A(0,6)=sym_os[8]/9-sym_os[5]/9-sym_os[4]/9+sym_os[0]/9;
		A(0,7)=sym_os[8]/9-sym_os[5]/9-sym_os[4]/9+sym_os[0]/9;
		A(0,8)=sym_os[8]/9-sym_os[5]/9-sym_os[4]/9+sym_os[0]/9;
		A(1,0)=(-(2*sym_os[8])/9)+(2*sym_os[5])/9-sym_os[4]/9+sym_os[0]/9;
		A(1,1)=sym_os[8]/9+sym_os[7]/3+(2*sym_os[5])/9+sym_os[4]/18+sym_os[1]/6+sym_os[0]/9;
		A(1,2)=sym_os[8]/9-sym_os[5]/9-sym_os[4]/9+sym_os[0]/9;
		A(1,3)=sym_os[8]/9-sym_os[7]/3+(2*sym_os[5])/9+sym_os[4]/18-sym_os[1]/6+sym_os[0]/9;
		A(1,4)=sym_os[8]/9-sym_os[5]/9-sym_os[4]/9+sym_os[0]/9;
		A(1,5)=(-sym_os[8]/18)-sym_os[7]/6-sym_os[5]/9+sym_os[4]/18+sym_os[1]/6+sym_os[0]/9;
		A(1,6)=(-sym_os[8]/18)+sym_os[7]/6-sym_os[5]/9+sym_os[4]/18-sym_os[1]/6+sym_os[0]/9;
		A(1,7)=(-sym_os[8]/18)+sym_os[7]/6-sym_os[5]/9+sym_os[4]/18-sym_os[1]/6+sym_os[0]/9;
		A(1,8)=(-sym_os[8]/18)-sym_os[7]/6-sym_os[5]/9+sym_os[4]/18+sym_os[1]/6+sym_os[0]/9;
		A(2,0)=(-(2*sym_os[8])/9)-sym_os[5]/9+(2*sym_os[4])/9+sym_os[0]/9;
		A(2,1)=sym_os[8]/9-sym_os[5]/9-sym_os[4]/9+sym_os[0]/9;
		A(2,2)=sym_os[8]/9+sym_os[6]/3+sym_os[5]/18+(2*sym_os[4])/9+sym_os[2]/6+sym_os[0]/9;
		A(2,3)=sym_os[8]/9-sym_os[5]/9-sym_os[4]/9+sym_os[0]/9;
		A(2,4)=sym_os[8]/9-sym_os[6]/3+sym_os[5]/18+(2*sym_os[4])/9-sym_os[2]/6+sym_os[0]/9;
		A(2,5)=(-sym_os[8]/18)-sym_os[6]/6+sym_os[5]/18-sym_os[4]/9+sym_os[2]/6+sym_os[0]/9;
		A(2,6)=(-sym_os[8]/18)-sym_os[6]/6+sym_os[5]/18-sym_os[4]/9+sym_os[2]/6+sym_os[0]/9;
		A(2,7)=(-sym_os[8]/18)+sym_os[6]/6+sym_os[5]/18-sym_os[4]/9-sym_os[2]/6+sym_os[0]/9;
		A(2,8)=(-sym_os[8]/18)+sym_os[6]/6+sym_os[5]/18-sym_os[4]/9-sym_os[2]/6+sym_os[0]/9;
		A(3,0)=(-(2*sym_os[8])/9)+(2*sym_os[5])/9-sym_os[4]/9+sym_os[0]/9;
		A(3,1)=sym_os[8]/9-sym_os[7]/3+(2*sym_os[5])/9+sym_os[4]/18-sym_os[1]/6+sym_os[0]/9;
		A(3,2)=sym_os[8]/9-sym_os[5]/9-sym_os[4]/9+sym_os[0]/9;
		A(3,3)=sym_os[8]/9+sym_os[7]/3+(2*sym_os[5])/9+sym_os[4]/18+sym_os[1]/6+sym_os[0]/9;
		A(3,4)=sym_os[8]/9-sym_os[5]/9-sym_os[4]/9+sym_os[0]/9;
		A(3,5)=(-sym_os[8]/18)+sym_os[7]/6-sym_os[5]/9+sym_os[4]/18-sym_os[1]/6+sym_os[0]/9;
		A(3,6)=(-sym_os[8]/18)-sym_os[7]/6-sym_os[5]/9+sym_os[4]/18+sym_os[1]/6+sym_os[0]/9;
		A(3,7)=(-sym_os[8]/18)-sym_os[7]/6-sym_os[5]/9+sym_os[4]/18+sym_os[1]/6+sym_os[0]/9;
		A(3,8)=(-sym_os[8]/18)+sym_os[7]/6-sym_os[5]/9+sym_os[4]/18-sym_os[1]/6+sym_os[0]/9;
		A(4,0)=(-(2*sym_os[8])/9)-sym_os[5]/9+(2*sym_os[4])/9+sym_os[0]/9;
		A(4,1)=sym_os[8]/9-sym_os[5]/9-sym_os[4]/9+sym_os[0]/9;
		A(4,2)=sym_os[8]/9-sym_os[6]/3+sym_os[5]/18+(2*sym_os[4])/9-sym_os[2]/6+sym_os[0]/9;
		A(4,3)=sym_os[8]/9-sym_os[5]/9-sym_os[4]/9+sym_os[0]/9;
		A(4,4)=sym_os[8]/9+sym_os[6]/3+sym_os[5]/18+(2*sym_os[4])/9+sym_os[2]/6+sym_os[0]/9;
		A(4,5)=(-sym_os[8]/18)+sym_os[6]/6+sym_os[5]/18-sym_os[4]/9-sym_os[2]/6+sym_os[0]/9;
		A(4,6)=(-sym_os[8]/18)+sym_os[6]/6+sym_os[5]/18-sym_os[4]/9-sym_os[2]/6+sym_os[0]/9;
		A(4,7)=(-sym_os[8]/18)-sym_os[6]/6+sym_os[5]/18-sym_os[4]/9+sym_os[2]/6+sym_os[0]/9;
		A(4,8)=(-sym_os[8]/18)-sym_os[6]/6+sym_os[5]/18-sym_os[4]/9+sym_os[2]/6+sym_os[0]/9;
		A(5,0)=(1*sym_os[8])/9-sym_os[5]/9-sym_os[4]/9+sym_os[0]/9;
		A(5,1)=(-sym_os[8]/18)-sym_os[7]/6-sym_os[5]/9+sym_os[4]/18+sym_os[1]/6+sym_os[0]/9;
		A(5,2)=(-sym_os[8]/18)-sym_os[6]/6+sym_os[5]/18-sym_os[4]/9+sym_os[2]/6+sym_os[0]/9;
		A(5,3)=(-sym_os[8]/18)+sym_os[7]/6-sym_os[5]/9+sym_os[4]/18-sym_os[1]/6+sym_os[0]/9;
		A(5,4)=(-sym_os[8]/18)+sym_os[6]/6+sym_os[5]/18-sym_os[4]/9-sym_os[2]/6+sym_os[0]/9;
		A(5,5)=sym_os[8]/36+sym_os[7]/12+sym_os[6]/12+sym_os[5]/18+sym_os[4]/18+sym_os[3]/4+sym_os[2]/6+sym_os[1]/6+sym_os[0]/9;
		A(5,6)=sym_os[8]/36-sym_os[7]/12+sym_os[6]/12+sym_os[5]/18+sym_os[4]/18-sym_os[3]/4+sym_os[2]/6-sym_os[1]/6+sym_os[0]/9;
		A(5,7)=sym_os[8]/36-sym_os[7]/12-sym_os[6]/12+sym_os[5]/18+sym_os[4]/18+sym_os[3]/4-sym_os[2]/6-sym_os[1]/6+sym_os[0]/9;
		A(5,8)=sym_os[8]/36+sym_os[7]/12-sym_os[6]/12+sym_os[5]/18+sym_os[4]/18-sym_os[3]/4-sym_os[2]/6+sym_os[1]/6+sym_os[0]/9;
		A(6,0)=(1*sym_os[8])/9-sym_os[5]/9-sym_os[4]/9+sym_os[0]/9;
		A(6,1)=(-sym_os[8]/18)+sym_os[7]/6-sym_os[5]/9+sym_os[4]/18-sym_os[1]/6+sym_os[0]/9;
		A(6,2)=(-sym_os[8]/18)-sym_os[6]/6+sym_os[5]/18-sym_os[4]/9+sym_os[2]/6+sym_os[0]/9;
		A(6,3)=(-sym_os[8]/18)-sym_os[7]/6-sym_os[5]/9+sym_os[4]/18+sym_os[1]/6+sym_os[0]/9;
		A(6,4)=(-sym_os[8]/18)+sym_os[6]/6+sym_os[5]/18-sym_os[4]/9-sym_os[2]/6+sym_os[0]/9;
		A(6,5)=sym_os[8]/36-sym_os[7]/12+sym_os[6]/12+sym_os[5]/18+sym_os[4]/18-sym_os[3]/4+sym_os[2]/6-sym_os[1]/6+sym_os[0]/9;
		A(6,6)=sym_os[8]/36+sym_os[7]/12+sym_os[6]/12+sym_os[5]/18+sym_os[4]/18+sym_os[3]/4+sym_os[2]/6+sym_os[1]/6+sym_os[0]/9;
		A(6,7)=sym_os[8]/36+sym_os[7]/12-sym_os[6]/12+sym_os[5]/18+sym_os[4]/18-sym_os[3]/4-sym_os[2]/6+sym_os[1]/6+sym_os[0]/9;
		A(6,8)=sym_os[8]/36-sym_os[7]/12-sym_os[6]/12+sym_os[5]/18+sym_os[4]/18+sym_os[3]/4-sym_os[2]/6-sym_os[1]/6+sym_os[0]/9;
		A(7,0)=(1*sym_os[8])/9-sym_os[5]/9-sym_os[4]/9+sym_os[0]/9;
		A(7,1)=(-sym_os[8]/18)+sym_os[7]/6-sym_os[5]/9+sym_os[4]/18-sym_os[1]/6+sym_os[0]/9;
		A(7,2)=(-sym_os[8]/18)+sym_os[6]/6+sym_os[5]/18-sym_os[4]/9-sym_os[2]/6+sym_os[0]/9;
		A(7,3)=(-sym_os[8]/18)-sym_os[7]/6-sym_os[5]/9+sym_os[4]/18+sym_os[1]/6+sym_os[0]/9;
		A(7,4)=(-sym_os[8]/18)-sym_os[6]/6+sym_os[5]/18-sym_os[4]/9+sym_os[2]/6+sym_os[0]/9;
		A(7,5)=sym_os[8]/36-sym_os[7]/12-sym_os[6]/12+sym_os[5]/18+sym_os[4]/18+sym_os[3]/4-sym_os[2]/6-sym_os[1]/6+sym_os[0]/9;
		A(7,6)=sym_os[8]/36+sym_os[7]/12-sym_os[6]/12+sym_os[5]/18+sym_os[4]/18-sym_os[3]/4-sym_os[2]/6+sym_os[1]/6+sym_os[0]/9;
		A(7,7)=sym_os[8]/36+sym_os[7]/12+sym_os[6]/12+sym_os[5]/18+sym_os[4]/18+sym_os[3]/4+sym_os[2]/6+sym_os[1]/6+sym_os[0]/9;
		A(7,8)=sym_os[8]/36-sym_os[7]/12+sym_os[6]/12+sym_os[5]/18+sym_os[4]/18-sym_os[3]/4+sym_os[2]/6-sym_os[1]/6+sym_os[0]/9;
		A(8,0)=(1*sym_os[8])/9-sym_os[5]/9-sym_os[4]/9+sym_os[0]/9;
		A(8,1)=(-sym_os[8]/18)-sym_os[7]/6-sym_os[5]/9+sym_os[4]/18+sym_os[1]/6+sym_os[0]/9;
		A(8,2)=(-sym_os[8]/18)+sym_os[6]/6+sym_os[5]/18-sym_os[4]/9-sym_os[2]/6+sym_os[0]/9;
		A(8,3)=(-sym_os[8]/18)+sym_os[7]/6-sym_os[5]/9+sym_os[4]/18-sym_os[1]/6+sym_os[0]/9;
		A(8,4)=(-sym_os[8]/18)-sym_os[6]/6+sym_os[5]/18-sym_os[4]/9+sym_os[2]/6+sym_os[0]/9;
		A(8,5)=sym_os[8]/36+sym_os[7]/12-sym_os[6]/12+sym_os[5]/18+sym_os[4]/18-sym_os[3]/4-sym_os[2]/6+sym_os[1]/6+sym_os[0]/9;
		A(8,6)=sym_os[8]/36-sym_os[7]/12-sym_os[6]/12+sym_os[5]/18+sym_os[4]/18+sym_os[3]/4-sym_os[2]/6-sym_os[1]/6+sym_os[0]/9;
		A(8,7)=sym_os[8]/36-sym_os[7]/12+sym_os[6]/12+sym_os[5]/18+sym_os[4]/18-sym_os[3]/4+sym_os[2]/6-sym_os[1]/6+sym_os[0]/9;
		A(8,8)=sym_os[8]/36+sym_os[7]/12+sym_os[6]/12+sym_os[5]/18+sym_os[4]/18+sym_os[3]/4+sym_os[2]/6+sym_os[1]/6+sym_os[0]/9;
		break;
	default:
		printf("Error: matrix A not implemented forQ%d and MRT1_OG\n",Q);
	}
}

void LBM::setA_MRT2_OG(matrix &A, int Q, ex u, ex v, ex w)
{
	ex one=1;
	switch (Q)
	{
	case 5:
		A(0,0) = n4o5*sym_os[3]+n1o5*sym_os[0];
		A(0,1) = -n1o5*sym_os[3]+n1o5*sym_os[0];
		A(0,2) = -n1o5*sym_os[3]+n1o5*sym_os[0];
		A(0,3) = -n1o5*sym_os[3]+n1o5*sym_os[0];
		A(0,4) = -n1o5*sym_os[3]+n1o5*sym_os[0];
		A(1,0) = -n1o5*sym_os[3]+n1o5*sym_os[0];
		A(1,1) = n1o4*sym_os[4]+n1o20*sym_os[3]+n1o2*sym_os[1]+n1o5*sym_os[0];
		A(1,2) = -n1o4*sym_os[4]+n1o20*sym_os[3]+n1o5*sym_os[0];
		A(1,3) = n1o4*sym_os[4]+n1o20*sym_os[3]-n1o2*sym_os[1]+n1o5*sym_os[0];
		A(1,4) = -n1o4*sym_os[4]+n1o20*sym_os[3]+n1o5*sym_os[0];
		A(2,0) = -n1o5*sym_os[3]+n1o5*sym_os[0];
		A(2,1) = -n1o4*sym_os[4]+n1o20*sym_os[3]+n1o5*sym_os[0];
		A(2,2) = n1o4*sym_os[4]+n1o20*sym_os[3]+n1o2*sym_os[2]+n1o5*sym_os[0];
		A(2,3) = -n1o4*sym_os[4]+n1o20*sym_os[3]+n1o5*sym_os[0];
		A(2,4) = n1o4*sym_os[4]+n1o20*sym_os[3]-n1o2*sym_os[2]+n1o5*sym_os[0];
		A(3,0) = -n1o5*sym_os[3]+n1o5*sym_os[0];
		A(3,1) = n1o4*sym_os[4]+n1o20*sym_os[3]-n1o2*sym_os[1]+n1o5*sym_os[0];
		A(3,2) = -n1o4*sym_os[4]+n1o20*sym_os[3]+n1o5*sym_os[0];
		A(3,3) = n1o4*sym_os[4]+n1o20*sym_os[3]+n1o2*sym_os[1]+n1o5*sym_os[0];
		A(3,4) = -n1o4*sym_os[4]+n1o20*sym_os[3]+n1o5*sym_os[0];
		A(4,0) = -n1o5*sym_os[3]+n1o5*sym_os[0];
		A(4,1) = -n1o4*sym_os[4]+n1o20*sym_os[3]+n1o5*sym_os[0];
		A(4,2) = n1o4*sym_os[4]+n1o20*sym_os[3]-n1o2*sym_os[2]+n1o5*sym_os[0];
		A(4,3) = -n1o4*sym_os[4]+n1o20*sym_os[3]+n1o5*sym_os[0];
		A(4,4) = n1o4*sym_os[4]+n1o20*sym_os[3]+n1o2*sym_os[2]+n1o5*sym_os[0];
		break;
	case 7:
		A(0,0) = (6*sym_os[4]+sym_os[0])/7;
		A(0,1) = -(sym_os[4]-sym_os[0])/7;
		A(0,2) =-(sym_os[4]-sym_os[0])/7;
		A(0,3) =-(sym_os[4]-sym_os[0])/7;
		A(0,4) =-(sym_os[4]-sym_os[0])/7;
		A(0,5) =-(sym_os[4]-sym_os[0])/7;
		A(0,6) =-(sym_os[4]-sym_os[0])/7;

		A(1,0) =-(sym_os[4]-sym_os[0])/7;
		A(1,1) =(7*sym_os[6]+21*sym_os[5]+2*sym_os[4]+42*sym_os[1]+12*sym_os[0])/84;
		A(1,2) =(7*sym_os[6]-21*sym_os[5]+2*sym_os[4]+12*sym_os[0])/84;
		A(1,3) =-(7*sym_os[6]-sym_os[4]-6*sym_os[0])/42;
		A(1,4) =(7*sym_os[6]+21*sym_os[5]+2*sym_os[4]-42*sym_os[1]+12*sym_os[0])/84;
		A(1,5) =(7*sym_os[6]-21*sym_os[5]+2*sym_os[4]+12*sym_os[0])/84;
		A(1,6) =-(7*sym_os[6]-sym_os[4]-6*sym_os[0])/42;

		A(2,0) =-(sym_os[4]-sym_os[0])/7;
		A(2,1) =(7*sym_os[6]-21*sym_os[5]+2*sym_os[4]+12*sym_os[0])/84;
		A(2,2) =(7*sym_os[6]+21*sym_os[5]+2*sym_os[4]+42*sym_os[2]+12*sym_os[0])/84;
		A(2,3) =-(7*sym_os[6]-sym_os[4]-6*sym_os[0])/42;
		A(2,4) =(7*sym_os[6]-21*sym_os[5]+2*sym_os[4]+12*sym_os[0])/84;
		A(2,5) =(7*sym_os[6]+21*sym_os[5]+2*sym_os[4]-42*sym_os[2]+12*sym_os[0])/84;
		A(2,6) =-(7*sym_os[6]-sym_os[4]-6*sym_os[0])/42;

		A(3,0) =-(sym_os[4]-sym_os[0])/7;
		A(3,1) =-(7*sym_os[6]-sym_os[4]-6*sym_os[0])/42;
		A(3,2) =-(7*sym_os[6]-sym_os[4]-6*sym_os[0])/42;
		A(3,3) =(14*sym_os[6]+sym_os[4]+21*sym_os[3]+6*sym_os[0])/42;
		A(3,4) =-(7*sym_os[6]-sym_os[4]-6*sym_os[0])/42;
		A(3,5) =-(7*sym_os[6]-sym_os[4]-6*sym_os[0])/42;
		A(3,6) =(14*sym_os[6]+sym_os[4]-21*sym_os[3]+6*sym_os[0])/42;

		A(4,0) =-(sym_os[4]-sym_os[0])/7;
		A(4,1) =(7*sym_os[6]+21*sym_os[5]+2*sym_os[4]-42*sym_os[1]+12*sym_os[0])/84;
		A(4,2) =(7*sym_os[6]-21*sym_os[5]+2*sym_os[4]+12*sym_os[0])/84;
		A(4,3) =-(7*sym_os[6]-sym_os[4]-6*sym_os[0])/42;
		A(4,4) =(7*sym_os[6]+21*sym_os[5]+2*sym_os[4]+42*sym_os[1]+12*sym_os[0])/84;
		A(4,5) =(7*sym_os[6]-21*sym_os[5]+2*sym_os[4]+12*sym_os[0])/84;
		A(4,6) =-(7*sym_os[6]-sym_os[4]-6*sym_os[0])/42;

		A(5,0) =-(sym_os[4]-sym_os[0])/7;
		A(5,1) =(7*sym_os[6]-21*sym_os[5]+2*sym_os[4]+12*sym_os[0])/84;
		A(5,2) =(7*sym_os[6]+21*sym_os[5]+2*sym_os[4]-42*sym_os[2]+12*sym_os[0])/84;
		A(5,3) =-(7*sym_os[6]-sym_os[4]-6*sym_os[0])/42;
		A(5,4) =(7*sym_os[6]-21*sym_os[5]+2*sym_os[4]+12*sym_os[0])/84;
		A(5,5) =(7*sym_os[6]+21*sym_os[5]+2*sym_os[4]+42*sym_os[2]+12*sym_os[0])/84;
		A(5,6) =-(7*sym_os[6]-sym_os[4]-6*sym_os[0])/42;

		A(6,0) =-(sym_os[4]-sym_os[0])/7;
		A(6,1) =-(7*sym_os[6]-sym_os[4]-6*sym_os[0])/42;
		A(6,2) =-(7*sym_os[6]-sym_os[4]-6*sym_os[0])/42;
		A(6,3) =(14*sym_os[6]+sym_os[4]-21*sym_os[3]+6*sym_os[0])/42;
		A(6,4) =-(7*sym_os[6]-sym_os[4]-6*sym_os[0])/42;
		A(6,5) =-(7*sym_os[6]-sym_os[4]-6*sym_os[0])/42;
		A(6,6) =(14*sym_os[6]+sym_os[4]+21*sym_os[3]+6*sym_os[0])/42;
		break;
	default:
		printf("Error: matrix A not implemented for Q%d and MRT2_OG\n",Q);
	}
}

void LBM::setA_CLBM1_OG(matrix &A, int Q, ex u, ex v, ex w)
{
	ex one=1;
	switch (Q)
	{
	default:
		printf("Error: matrix A not implemented for Q%d and CLBM1_OG\n",Q);
	}
}

void LBM::setA_CLBM2_OG(matrix &A, int Q, ex u, ex v, ex w)
{
	ex one=1;
	switch (Q)
	{
	default:
		printf("Error: matrix A not implemented for Q%d and CLBM2_OG\n",Q);
	}
}



void LBM::setM_15(matrix &A)
{
	int maxima_matrix[15][15] {
		{1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1},
		{-2,	-1,	-1,	-1,	-1,	-1,	-1,	1,	1,	1,	1,	1,	1,	1,	1},
		{16,	-4,	-4,	-4,	-4,	-4,	-4,	1,	1,	1,	1,	1,	1,	1,	1},
		{0,	1,	0,	0,	-1,	0,	0,	1,	1,	1,	1,	-1,	-1,	-1,	-1},
		{0,	0,	1,	0,	0,	-1,	0,	1,	1,	-1,	-1,	1,	1,	-1,	-1},
		{0,	0,	0,	1,	0,	0,	-1,	1,	-1,	1,	-1,	1,	-1,	1,	-1},
		{0,	-4,	0,	0,	4,	0,	0,	1,	1,	1,	1,	-1,	-1,	-1,	-1},
		{0,	0,	-4,	0,	0,	4,	0,	1,	1,	-1,	-1,	1,	1,	-1,	-1},
		{0,	0,	0,	-4,	0,	0,	4,	1,	-1,	1,	-1,	1,	-1,	1,	-1},
		{0,	2,	-1,	-1,	2,	-1,	-1,	0,	0,	0,	0,	0,	0,	0,	0},
		{0,	0,	1,	-1,	0,	1,	-1,	0,	0,	0,	0,	0,	0,	0,	0},
		{0,	0,	0,	0,	0,	0,	0,	1,	1,	-1,	-1,	-1,	-1,	1,	1},
		{0,	0,	0,	0,	0,	0,	0,	1,	-1,	-1,	1,	1,	-1,	-1,	1},
		{0,	0,	0,	0,	0,	0,	0,	1,	-1,	1,	-1,	-1,	1,	-1,	1},
		{0,	0,	0,	0,	0,	0,	0,	1,	-1,	-1,	1,	-1,	1,	1,	-1}
		};
	for (int i=0;i<15;i++)
	for (int j=0;j<15;j++)
		A(i,j) = (ex)(maxima_matrix[i][j]);
}

void LBM::setM_19(matrix &A)
{
	int maxima_matrix[19][19] {
		{1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1},
		{-30,	-11,	-11,	-11,	-11,	-11,	-11,	8,	8,	8,	8,	8,	8,	8,	8,	8,	8,	8,	8},
		{12,	-4,	-4,	-4,	-4,	-4,	-4,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1},
		{0,	1,	0,	0,	-1,	0,	0,	0,	0,	0,	0,	1,	1,	-1,	-1,	1,	1,	-1,	-1},
		{0,	0,	1,	0,	0,	-1,	0,	1,	1,	-1,	-1,	1,	-1,	1,	-1,	0,	0,	0,	0},
		{0,	0,	0,	1,	0,	0,	-1,	1,	-1,	1,	-1,	0,	0,	0,	0,	1,	-1,	1,	-1},
		{0,	-4,	0,	0,	4,	0,	0,	0,	0,	0,	0,	1,	1,	-1,	-1,	1,	1,	-1,	-1},
		{0,	0,	-4,	0,	0,	4,	0,	1,	1,	-1,	-1,	1,	-1,	1,	-1,	0,	0,	0,	0},
		{0,	0,	0,	-4,	0,	0,	4,	1,	-1,	1,	-1,	0,	0,	0,	0,	1,	-1,	1,	-1},
		{0,	2,	-1,	-1,	2,	-1,	-1,	-2,	-2,	-2,	-2,	1,	1,	1,	1,	1,	1,	1,	1},
		{0,	0,	1,	-1,	0,	1,	-1,	0,	0,	0,	0,	1,	1,	1,	1,	-1,	-1,	-1,	-1},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	-1,	-1,	1,	0,	0,	0,	0},
		{0,	0,	0,	0,	0,	0,	0,	1,	-1,	-1,	1,	0,	0,	0,	0,	0,	0,	0,	0},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	-1,	-1,	1},
		{0,	-4,	2,	2,	-4,	2,	2,	-2,	-2,	-2,	-2,	1,	1,	1,	1,	1,	1,	1,	1},
		{0,	0,	-2,	2,	0,	-2,	2,	0,	0,	0,	0,	1,	1,	1,	1,	-1,	-1,	-1,	-1},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	1,	-1,	-1,	-1,	-1,	1,	1},
		{0,	0,	0,	0,	0,	0,	0,	1,	1,	-1,	-1,	-1,	1,	-1,	1,	0,	0,	0,	0},
		{0,	0,	0,	0,	0,	0,	0,	-1,	1,	-1,	1,	0,	0,	0,	0,	1,	-1,	1,	-1}
		};
	for (int i=0;i<19;i++)
	for (int j=0;j<19;j++)
		A(i,j) = (ex)(maxima_matrix[i][j]);
}


void LBM::setM_27(matrix &A)
{
	int maxima_matrix[27][27] {
		{1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1},
		{0,	1,	0,	0,	-1,	0,	0,	0,	0,	0,	0,	1,	1,	-1,	-1,	1,	1,	-1,	-1,	1,	1,	1,	1,	-1,	-1,	-1,	-1},
		{0,	0,	1,	0,	0,	-1,	0,	1,	1,	-1,	-1,	1,	-1,	1,	-1,	0,	0,	0,	0,	1,	1,	-1,	-1,	1,	1,	-1,	-1},
		{0,	0,	0,	1,	0,	0,	-1,	1,	-1,	1,	-1,	0,	0,	0,	0,	1,	-1,	1,	-1,	1,	-1,	1,	-1,	1,	-1,	1,	-1},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	-1,	-1,	1,	0,	0,	0,	0,	1,	1,	-1,	-1,	-1,	-1,	1,	1},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	-1,	-1,	1,	1,	-1,	1,	-1,	-1,	1,	-1,	1},
		{0,	0,	0,	0,	0,	0,	0,	1,	-1,	-1,	1,	0,	0,	0,	0,	0,	0,	0,	0,	1,	-1,	-1,	1,	1,	-1,	-1,	1},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	-1,	-1,	1,	-1,	1,	1,	-1},
		{0,	1,	0,	0,	1,	0,	0,	0,	0,	0,	0,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1},
		{0,	0,	1,	0,	0,	1,	0,	1,	1,	1,	1,	1,	1,	1,	1,	0,	0,	0,	0,	1,	1,	1,	1,	1,	1,	1,	1},
		{0,	0,	0,	1,	0,	0,	1,	1,	1,	1,	1,	0,	0,	0,	0,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	-1,	1,	-1,	0,	0,	0,	0,	1,	1,	-1,	-1,	1,	1,	-1,	-1},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	-1,	1,	-1,	1,	-1,	1,	-1,	1,	-1,	1,	-1},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	-1,	-1,	1,	1,	-1,	-1,	1},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	1,	-1,	-1,	0,	0,	0,	0,	1,	1,	1,	1,	-1,	-1,	-1,	-1},
		{0,	0,	0,	0,	0,	0,	0,	1,	-1,	1,	-1,	0,	0,	0,	0,	0,	0,	0,	0,	1,	-1,	1,	-1,	1,	-1,	1,	-1},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	-1,	1,	-1,	-1,	1,	-1,	1},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	1,	-1,	-1,	1,	1,	1,	1,	-1,	-1,	-1,	-1},
		{0,	0,	0,	0,	0,	0,	0,	1,	1,	-1,	-1,	0,	0,	0,	0,	0,	0,	0,	0,	1,	1,	-1,	-1,	1,	1,	-1,	-1},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	1,	-1,	-1,	-1,	-1,	1,	1},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	1,	1,	1,	0,	0,	0,	0,	1,	1,	1,	1,	1,	1,	1,	1},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1},
		{0,	0,	0,	0,	0,	0,	0,	1,	1,	1,	1,	0,	0,	0,	0,	0,	0,	0,	0,	1,	1,	1,	1,	1,	1,	1,	1},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	-1,	1,	-1,	1,	-1,	1,	-1},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	1,	-1,	-1,	1,	1,	-1,	-1},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	1,	1,	1,	-1,	-1,	-1,	-1},
		{0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	1,	1,	1,	1,	1,	1,	1,	1}
		};
	for (int i=0;i<27;i++)
	for (int j=0;j<27;j++)
		A(i,j) = (ex)(maxima_matrix[i][j]);
}


void LBM::setM(matrix &A, int Q)
{
	switch (Q)
	{
	case 3:
		A(0,0) = no1;	A(0,1) = no1;	A(0,2) = no1;
		A(1,0) = 0;	A(1,1) = no1;	A(1,2) = -no1;
		A(2,0) = 0;	A(2,1) = no1;	A(2,2) = no1;
		break;
	case 5:
		A(0,0) = no1;	A(0,1) = no1;	A(0,2) = no1;	A(0,3) = no1;	A(0,4) = no1;
		A(1,0) = 0;	A(1,1) = no1;	A(1,2) = 0;	A(1,3) =-no1;	A(1,4) = 0;
		A(2,0) = 0;	A(2,1) = 0;	A(2,2) = no1;	A(2,3) = 0;	A(2,4) =-no1;
		A(3,0) = 0;	A(3,1) = no1;	A(3,2) = 0;	A(3,3) = no1;	A(3,4) = 0;
		A(4,0) = 0;	A(4,1) = 0;	A(4,2) = no1;	A(4,3) = 0;	A(4,4) = no1;
		break;
	case 7:
		A(0,0) = no1;	A(0,1) = no1;	A(0,2) = no1;	A(0,3) = no1;	A(0,4) = no1;	A(0,5) = no1;	A(0,6) = no1;
		A(1,0) = 0;	A(1,1) = no1;	A(1,2) = 0;	A(1,3) = 0;	A(1,4) =-no1;	A(1,5) = 0;	A(1,6) = 0;
		A(2,0) = 0;	A(2,1) = 0;	A(2,2) = no1;	A(2,3) = 0;	A(2,4) = 0;	A(2,5) =-no1;	A(2,6) = 0;
		A(3,0) = 0;	A(3,1) = 0;	A(3,2) = 0;	A(3,3) = no1;	A(3,4) = 0;	A(3,5) = 0;	A(3,6) = -no1;
		A(4,0) = 0;	A(4,1) = no1;	A(4,2) = 0;	A(4,3) = 0;	A(4,4) = no1;	A(4,5) = 0;	A(4,6) = 0;
		A(5,0) = 0;	A(5,1) = 0;	A(5,2) = no1;	A(5,3) = 0;	A(5,4) = 0;	A(5,5) = no1;	A(5,6) = 0;
		A(6,0) = 0;	A(6,1) = 0;	A(6,2) = 0;	A(6,3) = no1;	A(6,4) = 0;	A(6,5) = 0;	A(6,6) = no1;
		break;
	case 9:
		A(0,0) = no1;	A(0,1) = no1;	A(0,2) = no1;	A(0,3) = no1;	A(0,4) = no1;	A(0,5) = no1;	A(0,6) = no1;	A(0,7) = no1;	A(0,8) = no1;
		A(1,0) = 0;	A(1,1) = no1;	A(1,2) = 0;	A(1,3) =-no1;	A(1,4) = 0;	A(1,5) = no1;	A(1,6) =-no1;	A(1,7) =-no1;	A(1,8) = no1;
		A(2,0) = 0;	A(2,1) = 0;	A(2,2) = no1;	A(2,3) = 0;	A(2,4) =-no1;	A(2,5) = no1;	A(2,6) = no1;	A(2,7) =-no1;	A(2,8) =-no1;
		A(3,0) = 0;	A(3,1) = 0;	A(3,2) = 0;	A(3,3) = 0;	A(3,4) = 0;	A(3,5) = no1;	A(3,6) =-no1;	A(3,7) = no1;	A(3,8) =-no1;
		A(4,0) = 0;	A(4,1) = no1;	A(4,2) = 0;	A(4,3) = no1;	A(4,4) = 0;	A(4,5) = no1;	A(4,6) = no1;	A(4,7) = no1;	A(4,8) = no1;
		A(5,0) = 0;	A(5,1) = 0;	A(5,2) = no1;	A(5,3) = 0;	A(5,4) = no1;	A(5,5) = no1;	A(5,6) = no1;	A(5,7) = no1;	A(5,8) = no1;
		A(6,0) = 0;	A(6,1) = 0;	A(6,2) = 0;	A(6,3) = 0;	A(6,4) = 0;	A(6,5) = no1;	A(6,6) = no1;	A(6,7) =-no1;	A(6,8) =-no1;
		A(7,0) = 0;	A(7,1) = 0;	A(7,2) = 0;	A(7,3) = 0;	A(7,4) = 0;	A(7,5) = no1;	A(7,6) =-no1;	A(7,7) =-no1;	A(7,8) = no1;
		A(8,0) = 0;	A(8,1) = 0;	A(8,2) = 0;	A(8,3) = 0;	A(8,4) = 0;	A(8,5) = no1;	A(8,6) = no1;	A(8,7) = no1;	A(8,8) = no1;
		break;
	case 15:setM_15(A);
		break;
	case 19:setM_19(A);
		break;
	case 27:setM_27(A);
		break;
	default:
		printf("Error: matrix M not implemented for Q%d\n",Q);
	}
}

// CLBM Matrix
ex LBM::getK( int Q, int i, int j, ex u=0, ex v=0, ex w=0)
{
	// only first DIM+1 rows
	if (i==0) return (ex)1;
	switch (Q)
	{
	case 3:
		if (i==1 && j==0) return -u; if (i==1 && j==0) return (ex)1-u; if (i==0 && j==0) return -(ex)1-u;
		break;
	case 5:
		if (i==1 && j==0) return -u;
		if (i==1 && j==1) return (ex)1-u;
		if (i==1 && j==2) return -u;
		if (i==1 && j==3) return -u-1;
		if (i==1 && j==4) return -u;

		if (i==2 && j==0) return -v;
		if (i==2 && j==1) return -v;
		if (i==2 && j==2) return (ex)1-u;
		if (i==2 && j==3) return -v;
		if (i==2 && j==4) return -v-1;
		break;
	case 7:
		if (i==1 && j==0) return -u;
		if (i==1 && j==1) return (ex)1-u;
		if (i==1 && j==2) return -u;
		if (i==1 && j==3) return -u;
		if (i==1 && j==4) return -u-1;
		if (i==1 && j==5) return -u;
		if (i==1 && j==6) return -u;

		if (i==2 && j==0) return -v;
		if (i==2 && j==1) return -v;
		if (i==2 && j==2) return -u+1;
		if (i==2 && j==3) return -v;
		if (i==2 && j==4) return -v;
		if (i==2 && j==5) return -v-1;
		if (i==2 && j==6) return -v;

		if (i==3 && j==0) return -w;
		if (i==3 && j==1) return -w;
		if (i==3 && j==2) return -w;
		if (i==3 && j==3) return -w+1;
		if (i==3 && j==4) return -w;
		if (i==3 && j==5) return -w;
		if (i==3 && j==6) return -w-1;
		break;
	default:
		cout << "ERROR: setK not implemented for Q="<<Q<<endl;
	}
	cout << "ERROR: getK i="<<i<<","<<j<<" not implemented for Q="<<Q<<endl;
	return 0;
}

void LBM::setC_3(matrix &C)
{
	int maxima_matrix[1][3] = {
		{0,	1,	-1}
	};
	for (int i=0;i<1;i++)
	for (int j=0;j<3;j++)
		C(i,j) = (ex)(maxima_matrix[i][j]);
}

void LBM::setC_5(matrix &C)
{
	int maxima_matrix[2][5] = {
		{0,	1,	0,	-1,	0},
		{0,	0,	1,	0,	-1}
	};
	for (int i=0;i<2;i++)
	for (int j=0;j<5;j++)
		C(i,j) = (ex)(maxima_matrix[i][j]);
}

void LBM::setC_9(matrix &C)
{
	int maxima_matrix[2][9] = {
		{0,	1,	0,	-1,	0,	1,	-1,	-1,	1},
		{0,	0,	1,	0,	-1,	1,	1,	-1,	-1}
	};
	for (int i=0;i<2;i++)
	for (int j=0;j<9;j++)
		C(i,j) = (ex)(maxima_matrix[i][j]);
}

void LBM::setC_7(matrix &C)
{
	int maxima_matrix[3][7] = {
		{0,	1,	0,	0,	-1,	0,	0},
		{0,	0,	1,	0,	0,	-1,	0},
		{0,	0,	0,	1,	0,	0,	-1}
	};
	for (int i=0;i<3;i++)
	for (int j=0;j<7;j++)
		C(i,j) = (ex)(maxima_matrix[i][j]);
}

void LBM::setC_15(matrix &C)
{
	int maxima_matrix[3][15] = {
		{0,	1,	0,	0,	-1,	0,	0,	1,	1,	1,	1,	-1,	-1,	-1,	-1},
		{0,	0,	1,	0,	0,	-1,	0,	1,	1,	-1,	-1,	1,	1,	-1,	-1},
		{0,	0,	0,	1,	0,	0,	-1,	1,	-1,	1,	-1,	1,	-1,	1,	-1}
	};
	for (int i=0;i<3;i++)
	for (int j=0;j<15;j++)
		C(i,j) = (ex)(maxima_matrix[i][j]);
}

void LBM::setC_19(matrix &C)
{
	int maxima_matrix[3][19] = {
		{0,	1,	0,	0,	-1,	0,	0,	0,	0,	0,	0,	1,	1,	-1,	-1,	1,	1,	-1,	-1},
		{0,	0,	1,	0,	0,	-1,	0,	1,	1,	-1,	-1,	1,	-1,	1,	-1,	0,	0,	0,	0},
		{0,	0,	0,	1,	0,	0,	-1,	1,	-1,	1,	-1,	0,	0,	0,	0,	1,	-1,	1,	-1}
	};
	for (int i=0;i<3;i++)
	for (int j=0;j<19;j++)
		C(i,j) = (ex)(maxima_matrix[i][j]);
}

void LBM::setC_27(matrix &C)
{
	int maxima_matrix[3][27] = {
	{0,	1,	0,	0,	-1,	0,	0,	0,	0,	0,	0,	1,	1,	-1,	-1,	1,	1,	-1,	-1,	1,	1,	1,	1,	-1,	-1,	-1,	-1},
	{0,	0,	1,	0,	0,	-1,	0,	1,	1,	-1,	-1,	1,	-1,	1,	-1,	0,	0,	0,	0,	1,	1,	-1,	-1,	1,	1,	-1,	-1},
	{0,	0,	0,	1,	0,	0,	-1,	1,	-1,	1,	-1,	0,	0,	0,	0,	1,	-1,	1,	-1,	1,	-1,	1,	-1,	1,	-1,	1,	-1}
	};
	for (int i=0;i<3;i++)
	for (int j=0;j<27;j++)
		C(i,j) = (ex)(maxima_matrix[i][j]);
}

void LBM::setC(matrix &C, int Q)
{
	switch (Q)
	{
	case 3:	setC_3(C);
		break;
	case 5:	setC_5(C);
		break;
	case 7:	setC_7(C);
		break;
	case 9:	setC_9(C);
		break;
	case 15:setC_15(C);
		break;
	case 19:setC_19(C);
		break;
	case 27:setC_27(C);
		break;
	default:
		printf("Error: matrix C not implemented for Q%d\n",Q);
	}
}

void LBM::set_model_desc(char *desc, int m)
{
	switch (m)
	{
		case 0: sprintf(desc,"srt"); break;
		case 1: sprintf(desc,"mrt1"); break;
		case 2: sprintf(desc,"mrt2"); break;
		case 3: sprintf(desc,"mrt1og"); break;
		case 4: sprintf(desc,"mrt2og"); break;
		case 5: sprintf(desc,"clbm1"); break;
		case 6: sprintf(desc,"clbm2"); break;
		default:
			sprintf(desc,"unknown_model_number_%d",m);
	}
}

bool LBM::is_const_matrix(int m)
{
	if (m>=5) return false;
	return true;
}